UNIT OPCi386;

INTERFACE

USES
  OPT, OPL, OPLi386, OPOi386, OPM, _SYSTEM;   

CONST
  EQ = OPLi386.rcje;  NE = OPLi386.rcjne; LT = OPLi386.rcjl;
  LE = OPLi386.rcjle; GT = OPLi386.rcjg;  GE = OPLi386.rcjge;
  AB = OPLi386.rcja;  AE = OPLi386.rcjae; BL = OPLi386.rcjb;
  BE = OPLi386.rcjbe; CS = OPLi386.rcjc;  CC = OPLi386.rcjnc;

  always = OPLi386.rcjmp;
  never = -1;

  aNil = -1;	  (* label nil *)
  BUG  = 40;

  nStat = OPLi386.newStat;
  ESP = OPLi386.RiscESP;
  FP = OPLi386.RiscFP;
  none = OPLi386.none;
	noScale = OPOi386.noScale;

  BaseTypeOffs = -8;
  MethodOffs = -72;
  ArrBlkFirstElem = 8;
  ArrBlkLastDim = 12;
	StaticLinkAdr = 8;

  Ofalse = 0; Otrue = 1; Onil = 0;

  finlabel = OPLi386.finlabel;

TYPE
  Item = RECORD
    mode, mnolev, scale: SHORTINT;	(* mnolev < 0 -> mno (module number) = -mnolev *)
    typ: OPT.Structs;
    node: OPT.Nodes;
    adr, offs, inx, descReg, descOffs: LONGINT
  END;
  BLabel = LONGINT;

(* Items:
   mode  | adr    offs     inx
---------------------------------
 1 Var   | vadr	(adr will be stored in obj.linkadr too. 
         |	Local variables)
 2 VarPar| vadr   	
 3 Con   | val    	(adr: Boolean, Byte..LInt)
         | adr    len(of string)
         | adr   	(LReal, Real)
 6 LProc | adr            
 7 XProc | pno    LinkAdr 	(LinkAdr will be stored in obj.linkadr too)
 9 CProc |     	
10 IProc | adr    LinkAdr	(not implemented yet)
13 TProc | TagAdr Off
         |
15 Reg   | reg          
16 Abs   | vadr            reg	(inx = none: no index register)
17 RegRel| reg    off      reg	(inx = none: no index register)
18 Coc   | CC     Tjmp     Fjmp
*)

	VAR
	level: SHORTINT; 	(* nesting level *)
	pc: INTEGER; 	(* pc of pseudo RISC machine *)
	vTrue, vFalse, Stack: Item; 	(* constant items True, False and Stack *)
	ebp: Item; 	(* base pointer item *)
	formTab: ARRAY [0..16] OF SHORTINT;
	CCtab: ARRAY [0..2, 0..(vgeq-veql+1)] OF INTEGER;	(* condition code table, CCtab[0] = integer; CCtab[1] = char *)
	InvCCtab: ARRAY [0..(CC-EQ) DIV 32 + 1] OF INTEGER; (* inverted conditon code table*)
	lastTD: Item;	(*optimizing the method call: TD of the self parameter. Set by Parameter ParRecord, used by call. only set whe a AVR Record is pushed*)

  PROCEDURE Init;
  PROCEDURE NewStat(textPos: LONGINT);
  PROCEDURE DefFinallyLabel(n: OPT.Nodes);
  PROCEDURE FNEnter (proc: OPT.Objects; dataSize: LONGINT; node: OPT.Nodes);
  PROCEDURE FNExit(proc: OPT.Objects);
  PROCEDURE AllocSDynArr(VAR proc: OPT.Objects);
  PROCEDURE FNTrap(n: LONGINT; node: OPT.Nodes); (* HALT (n) *)
  PROCEDURE PushRegs;
  PROCEDURE SetRange(VAR z, x, y: Item);
  PROCEDURE FNNot(VAR z, x: Item);
  PROCEDURE FNNeg(VAR z, x: Item);
  PROCEDURE Field (VAR x: Item; offset: LONGINT);
  PROCEDURE DeRef (VAR x: Item);
  PROCEDURE FNIndex(VAR z, index: Item);
  PROCEDURE TypeTest (VAR x: Item; testtyp: OPT.Structs; guard, equal, varRec: BOOLEAN);
  PROCEDURE FNProcedure(VAR proc: Item; n: OPT.Nodes);
  PROCEDURE Convert (VAR x: Item; form: SHORTINT);
  PROCEDURE SetElem(VAR z, x: Item);
  PROCEDURE AbsVal(VAR z, x: Item);
  PROCEDURE FNCap(VAR z, x: Item);
  PROCEDURE FNOdd(VAR z, x: Item); (* z := ODD (x) *)
  PROCEDURE FNAsh(VAR z, x, y: Item);
  PROCEDURE SYSmop (VAR z, x: Item; subcl: SHORTINT; typ: OPT.Structs);
  PROCEDURE GetTdAdr (VAR typ: OPT.Structs; VAR tag: Item);
  PROCEDURE CondAnd (VAR x: Item);
  PROCEDURE CondOr (VAR x: Item);
  PROCEDURE Relation (VAR x: Item);
  PROCEDURE FNMul(VAR z, x, y: Item; f: INTEGER);
  PROCEDURE FNDiv(VAR z, x, y: Item; f: INTEGER);
  PROCEDURE FNMod(VAR z, x, y: Item);
  PROCEDURE FNAnd(VAR z, x, y: Item);
  PROCEDURE FNAdd(VAR z, x, y: Item; f: INTEGER);
  PROCEDURE FNSub(VAR z, x, y: Item; f: INTEGER);
  PROCEDURE FNOr(VAR z, x, y: Item);
  PROCEDURE FNCmp(VAR z, x, y: Item; rel: INTEGER);
  PROCEDURE FNIn (VAR z, x, y: Item);
  PROCEDURE Msk (VAR z, x, y: Item);
  PROCEDURE FNLen (VAR len, x, y: Item);
  PROCEDURE SYSdop (VAR z, x, y: Item; subcl: SHORTINT);
  PROCEDURE AllocateStatic(VAR tos: Item; size: LONGINT);
  PROCEDURE CorrectStackAdr(VAR x: Item; size, dynblocks: LONGINT);
  PROCEDURE PushRetDesc(VAR x, tos: Item);
  PROCEDURE Parameter (VAR ap: Item; fp: OPT.Objects; apIsDeref, alwaysRef: BOOLEAN);
  PROCEDURE FNCall(VAR x: Item; proc: OPT.Objects; node: OPT.Nodes);
  PROCEDURE RemoveStatic(size: LONGINT);
  PROCEDURE ResetStack(dynblocks: LONGINT);
  PROCEDURE PopResult(n: OPT.Nodes; VAR z: Item);
  PROCEDURE FNAssign(VAR z, x: Item);
  PROCEDURE SYSinterrupt(clear: BOOLEAN);
  PROCEDURE SYSmove (VAR z, x, nofBytes: Item);
  PROCEDURE IncDec (VAR z, x: Item; increment: BOOLEAN);
  PROCEDURE Include (VAR z, x: Item; incl: BOOLEAN);
  PROCEDURE SYSgetput(VAR z, x: Item; getfn: BOOLEAN);
  PROCEDURE SYSgetputReg (VAR z, x: Item; getrfn: BOOLEAN);
  PROCEDURE SYSportio(VAR z, x: Item; pin: BOOLEAN);
  PROCEDURE Await(z, self, y: Item);
  PROCEDURE Lock(self, excl: Item);		(* active only *)
  PROCEDURE Unlock(self, excl: Item);		(* active only *)
  PROCEDURE FNCopy (VAR z, x: Item);
  PROCEDURE NewSys (VAR z, x: Item);
  PROCEDURE NewRec (VAR z: Item; typ: OPT.Structs);
  PROCEDURE PushLen(VAR z: Item);
  PROCEDURE PopLen(VAR block: Item);
  PROCEDURE GenDimTrap (VAR len: Item);
  PROCEDURE MulDim(VAR nofelem, len: Item);
  PROCEDURE NewArray(VAR z, nofelem: Item; nofdim: LONGINT; typ: OPT.Structs; dimUsed: BOOLEAN);
  PROCEDURE CallRecBody(VAR self, body: Item; typ: OPT.Structs);
  PROCEDURE Jncc (VAR x: Item; VAR loc: BLabel; node: OPT.Nodes);
  PROCEDURE FNJmp (VAR loc: BLabel; node: OPT.Nodes);
  PROCEDURE FixLinkWith (L, val: bLabel);
  PROCEDURE FixLink (L: BLabel);
  PROCEDURE FNCase(VAR x: Item; low, high: LONGINT; VAR tab: LONGINT; VAR L: BLabel; node: OPT.Nodes);
  PROCEDURE DefLabel (VAR L: bLabel);
  PROCEDURE CaseFixup (tab: LONGINT; elseLabel: BLabel; len: LONGINT); (* case *)
  PROCEDURE FNJcc (VAR x: Item; VAR loc: BLabel; node: OPT.Nodes);
  PROCEDURE FNReturn(VAR res: Item; proc: OPT.Objects);
  PROCEDURE GenAsm(n: OPT.Nodes);
  PROCEDURE OPCBegin;
  PROCEDURE NewItem(VAR x: Item);

IMPLEMENTATION

PROCEDURE Init;
BEGIN
  pc := 0;
  level := 0
END;

PROCEDURE NewItem(VAR x: Item);
BEGIN
  x.mode:=0;
  x.mnolev:=0;
  x.scale:=0;
  x.typ:=OPT.linttyp;
  x.adr:=none;
  x.offs:=0;
  x.inx:=0;
  x.descReg:=0;
  x.descOffs:=0;
END;

PROCEDURE IncReg (pReg: LONGINT);
BEGIN
  IF pReg >= 0 THEN INC(OPL.InstrTab[pReg].used);
END;
	
PROCEDURE InitInstr (op: INTEGER; mode, form: SHORTINT; node: OPT.Nodes);
VAR
  i: LONGINT;
  instr: OPL.InstructionTable;
BEGIN
  IF pc >= LENGTH(OPL.InstrTab) THEN BEGIN
    SETLENGTH(instr, LENGTH(OPL.InstrTab)*2);
    FOR i := 0 TO LENGTH(OPL.InstrTab)-1 DO instr[i] := OPL.InstrTab[i];
    OPL.InstrTab := instr
  END;
  ASSERT ((mode >= 0) AND (mode <= 3));
  ASSERT ((op MOD 32 = 0) AND (op > 0));
  OPL.InstrTab [pc].op   := op + formTab[form] + mode * 8;
  OPL.InstrTab [pc].node := node;
  OPL.InstrTab [pc].src1 := none;
  OPL.InstrTab [pc].src2 := none;
  OPL.InstrTab [pc].inx  := none;
  OPL.InstrTab [pc].dest := pc;
  OPL.InstrTab [pc].used := 0;
  OPL.InstrTab [pc].hint := OPL.noHint;
  OPL.InstrTab [pc].pc   := OPL.aNil;
  OPL.InstrTab [pc].reg  := OPL.aNil;
  OPL.InstrTab [pc].abs  := FALSE;
END;

PROCEDURE NewStat(textPos: LONGINT);
BEGIN
  InitInstr (nStat, 0, 0, NIL);
  OPL.InstrTab[pc].src1 := textPos;
  OPM.errpos.CodePos := textPos;
  INC (pc)
END;

PROCEDURE DefFinallyLabel(n: OPT.Nodes);
BEGIN
  InitInstr(finlabel, 0, 0, n);
  INC (pc)
END;

PROCEDURE MoveCoc (VAR z, y: Item); FORWARD;

PROCEDURE GenLoad (op: INTEGER; VAR z, x: Item);
BEGIN
  CASE x.mode OF
      omAbs:
            BEGIN
              InitInstr (op, OPOi386.MemReg, x.typ.form, x.node);
              OPL.InstrTab[pc].src2 := x.adr;
              OPL.InstrTab[pc].inx := x.inx;
              OPL.InstrTab[pc].scale := x.scale;
              OPL.InstrTab[pc].src1 := none;
              IncReg (x.inx);
            END;
      omVar:
            BEGIN
              ASSERT (x.mnolev > 0);
              InitInstr (op, OPOi386.MemReg, x.typ.form, x.node);
              OPL.InstrTab[pc].src2 := x.adr;
              OPL.InstrTab[pc].inx := none;
              OPL.InstrTab[pc].scale := x.scale;
              OPL.InstrTab[pc].src1 := FP;
            END;
      omCon:
            BEGIN
              ASSERT ( (op <> rclea) AND (op <> fload) AND (op <> fild) AND (op <> rcpop) );
              ASSERT ( (op < rcldbdw) OR (op > rcldbwu) );
              InitInstr (op, OPOi386.ImmReg, x.typ.form, x.node);
              OPL.InstrTab[pc].src2 := x.adr;
            END;
      omReg:
            BEGIN
              ASSERT ((op <> rclea) AND (op <> fild));
              InitInstr (op, OPOi386.RegReg, x.typ.form, x.node);
              OPL.InstrTab[pc].src1 := x.adr;
              IF op <> rcpop THEN IncReg (x.adr);
            END;
      omRegRel:
            BEGIN
              InitInstr (op, OPOi386.MemReg, x.typ.form, x.node);
              OPL.InstrTab[pc].src1 := x.adr;
              OPL.InstrTab[pc].src2 := x.offs;
              OPL.InstrTab[pc].inx := x.inx;
              OPL.InstrTab[pc].scale := x.scale;
              IncReg (x.adr);
              IncReg (x.inx);
            END;
      omCoc:
            BEGIN
              MoveCoc(z, x);
              EXIT;
            END;
  ELSE
    HALT (BUG)
  END;
  z.adr := pc;
  z.mode := omReg;
  INC (pc)
END;
					
PROCEDURE GenPush (VAR x: Item);
VAR
  dummy: Item;
BEGIN
  ASSERT ((x.mode = omReg) OR (x.mode = omCon){, 100});
  dummy.node := NIL;
  GenLoad (rcpush, dummy, x)
END;

PROCEDURE GenPop (VAR x: Item);
VAR
  dummy: Item;
BEGIN
  dummy.node := NIL;
  x.mode := omReg;
  x.adr := pc;
  x.node := NIL;
  GenLoad (rcpop, dummy, x);
END;

PROCEDURE GenLoadProc (op: INTEGER; VAR z: Item; val: LONGINT; node: OPT.Nodes);
(* ldProc, ldXProc *)
BEGIN
  InitInstr (op, OPOi386.ImmReg, sfLInt, node);
  z.mode := omReg;
  z.typ := OPT.ptrtyp;
  z.adr := pc;
  OPL.InstrTab [pc].src1 := val;
  INC (pc)
END;
	
PROCEDURE GenStore (op: INTEGER; VAR z, x: Item);
BEGIN
  ASSERT (x.mode = omReg);
  CASE z.mode OF
      omAbs:
            BEGIN
              InitInstr (op, OPOi386.RegMem, z.typ.form, z.node);
              OPL.InstrTab[pc].dest := x.adr;
              OPL.InstrTab[pc].src2 := z.adr;
              OPL.InstrTab[pc].inx := z.inx;
              OPL.InstrTab[pc].scale := z.scale;
              OPL.InstrTab[pc].src1 := none;
              IncReg (z.inx);
            END;
      omVar:
            BEGIN
              ASSERT (z.mnolev > 0);
              InitInstr (op, OPOi386.RegMem, z.typ.form, z.node); OPL.InstrTab[pc].dest := x.adr;
              OPL.InstrTab[pc].src2 := z.adr; OPL.InstrTab[pc].inx := none;
              OPL.InstrTab[pc].src1 := FP;
            END;
      omRegRel:
            BEGIN
              InitInstr (op, OPOi386.RegMem, z.typ.form, z.node);
              OPL.InstrTab[pc].dest := x.adr;
              OPL.InstrTab[pc].src1 := z.adr;
              OPL.InstrTab[pc].src2 := z.offs;
              OPL.InstrTab[pc].inx := z.inx;
              OPL.InstrTab[pc].scale := z.scale;
              IncReg (z.adr); IncReg (z.inx);
            END;
  ELSE
    HALT (BUG)
  END;
  IF op <> rcgetReg THEN IncReg (x.adr);
  INC (pc)
END;

PROCEDURE GenPopReg (hint, form: SHORTINT);
BEGIN
  IF hint = OPL.noHint THEN
    InitInstr (rcpopReg, 0, 0, NIL)
  ELSE BEGIN
    InitInstr (rcpopReg, OPOi386.RegReg, form, NIL);
    OPL.InstrTab [pc].hint := hint;
  END;
  INC (pc)
END;

PROCEDURE Gen0 (op: INTEGER);
(* used for: cld, pushReg, label *)
BEGIN
  InitInstr (op, 0, 0, NIL);
  INC (pc)
END;

PROCEDURE Gen1 (op: INTEGER; adr: LONGINT; node: OPT.Nodes);
BEGIN
  IF op = rcjmpReg THEN BEGIN
    InitInstr (op, OPOi386.RegReg, sfLInt, node);
    IncReg (adr)
  END ELSE
    InitInstr (op, 0, 0, node);
  OPL.InstrTab[pc].src2 := adr;
  INC (pc)
END;

PROCEDURE Gen2 (op: INTEGER; VAR z, x: Item);
BEGIN
  ASSERT (x.mode = omReg);
  InitInstr (op, OPOi386.RegReg, x.typ.form, x.node);
  OPL.InstrTab[pc].src1 := x.adr;
  IncReg (x.adr);
  z.adr := pc;
  z.mode := omReg;
  INC (pc)
END;
	
PROCEDURE Gen3 (op: INTEGER; VAR z, x, y: Item);
BEGIN
  IF y.mode = omCon THEN
    InitInstr (op, OPOi386.ImmReg, x.typ.form, x.node)
  ELSE BEGIN
    InitInstr (op, OPOi386.RegReg, x.typ.form, x.node);
    IncReg (y.adr);
  END;
  OPL.InstrTab[pc].src1 := x.adr;
  OPL.InstrTab[pc].src2 := y.adr;
  IncReg (x.adr);
  IF (op = rcdiv) OR (op = rcmod) THEN
    OPL.InstrTab[x.adr].hint := OPL.tryEAX
  ELSE IF (op >= rcsal) AND (op <= rcrol) AND (y.mode <> omCon) THEN
    OPL.InstrTab[y.adr].hint := OPL.useECX
  ELSE IF (op = rcmul) AND (x.typ.form = sfSInt) THEN
    OPL.InstrTab [x.adr].hint := OPL.tryEAX;
  z.adr := pc;
  z.mode := omReg;
  INC (pc)
END;

PROCEDURE PushRegs;
BEGIN
  Gen0 (rcpushReg);
END;

PROCEDURE GenCall (op: INTEGER; adr, parSize: LONGINT; node: OPT.Nodes);
BEGIN
  IF op = rccallReg THEN BEGIN
    InitInstr (op, OPOi386.RegReg, sfLInt, node);
    IncReg (adr)
  END ELSE
    InitInstr (op, 0, 0, node);
  OPL.InstrTab [pc].src1 := parSize;
  OPL.InstrTab[pc].src2 := adr;
  INC (pc)
END;
	
PROCEDURE GenFlags (op: INTEGER; VAR x, y: Item);
BEGIN
  ASSERT((x.mode=omReg) AND (y.mode IN [omReg, omCon]){, 100});
  IF y.mode = omCon THEN
    InitInstr (op, OPOi386.ImmReg, x.typ.form, x.node)
  ELSE BEGIN
    InitInstr (op, OPOi386.RegReg, x.typ.form, x.node);
    IncReg (y.adr)
  END;
  OPL.InstrTab[pc].src1 := x.adr;
  OPL.InstrTab[pc].src2 := y.adr;
  IncReg (x.adr);
  INC (pc)
END;
	
PROCEDURE GenSetC (op: INTEGER; VAR z: Item); (* set condition *)
BEGIN
  InitInstr (op, OPOi386.RegReg, sfSInt, z.node);
  z.adr := pc;
  z.mode := omReg;
  INC (pc)
END;

PROCEDURE GenRepMovs (VAR z, x, nofElem: Item; moveSize: SHORTINT;  reverse: BOOLEAN); (* z := x; block copy *)
BEGIN
  InitInstr (rcrepMovs, 0, 0, NIL);
  OPL.InstrTab [pc].dest := z.adr; (* destination address *)
  OPL.InstrTab [pc].src1 := x.adr; (* source address *)
  OPL.InstrTab [pc].src2 := nofElem.adr; (* nofelements *)
  OPL.InstrTab [pc].inx := moveSize; (* move size: 8, 16 or 32 bit *)
  IF reverse THEN  OPL.InstrTab[pc].hint := 1;
  INC (pc)
END;
	
PROCEDURE GenPhi (reg0, reg1: LONGINT);
BEGIN
  InitInstr (rcphi, 0, 0, NIL);
  OPL.InstrTab [pc].src1 := reg0;
  OPL.InstrTab [pc].src2 := reg1;
  OPL.InstrTab [reg1].hint := -1 - reg0;
  INC (pc)
END;
	
PROCEDURE  MakeCon (VAR cons: Item; typ: OPT.Structs; val: LONGINT);
BEGIN
  cons.mode := omCon;
  cons.adr := val;
  cons.offs := 0;
  cons.typ := typ;
  cons.node := NIL
END;

PROCEDURE SetCC (VAR x: Item; cc: LONGINT);
BEGIN
  x.typ := OPT.booltyp;
  x.mode := omCoc;
  x.adr := cc;
  x.offs := aNil;
  x.inx := aNil;
  x.node := NIL
END;

PROCEDURE DefLabel (VAR L: bLabel);
BEGIN
  L := pc;
  Gen0 (rclabel)
END;
	
PROCEDURE FixLinkWith (L, val: bLabel);
VAR
  L1: bLabel;
BEGIN
  ASSERT ((L < 0) AND (val >= 0));
  L := -2 - L;
  WHILE L <> aNil DO BEGIN
    L1 := -2 - OPL.InstrTab[L].src2;
    OPL.InstrTab[L].src2 := val;
    L := L1
  END
END;
	
PROCEDURE FixLink (L: BLabel);
VAR
  L1: BLabel;
BEGIN
  ASSERT(L < 0);
  L := -2 - L;
  IF L <> aNil THEN BEGIN
    REPEAT
      L1 := -2 - OPL.InstrTab[L].src2;
      OPL.InstrTab[L].src2 := pc;
      L := L1
    UNTIL L = aNil;
    Gen0 (rclabel)
  END
END;
	
FUNCTION MergedLinks (L0, L1: BLabel): BLabel;
VAR
  L2, L3: BLabel;
(* -2 - Label = position in the pseudo code *)
BEGIN
  L2 := -2 - L0;
  IF L2 <> aNil THEN BEGIN
    REPEAT
      L3 := -2 - OPL.InstrTab[L2].src2;
      IF L3 = aNil THEN BREAK;
      L2 := L3
    UNTIL FALSE;
    OPL.InstrTab[L2].src2 := L1;
    RESULT:= L0;
    EXIT;
  END ELSE
    RESULT:= L1
END;

PROCEDURE FNJmp (VAR loc: BLabel; node: OPT.Nodes);
VAR
  pos: BLabel;
BEGIN
  IF loc < 0 THEN BEGIN(* forward jump *)
    pos := loc;
    loc := -2-pc;
    Gen1 (rcjmp, pos, node)
  END ELSE (* backward jump *)
    Gen1 (rcjmp, loc, node)
END;

PROCEDURE load (VAR x: Item); FORWARD
	
PROCEDURE FNJcc (VAR x: Item; VAR loc: BLabel; node: OPT.Nodes);
BEGIN
  IF x.mode <> omCoc THEN BEGIN
    ASSERT (x.typ.form = sfBool);
    load (x);
    GenFlags (rccmp, x, VTrue);
    SetCC (x, EQ)
  END;
  IF x.adr <> never THEN BEGIN
    IF loc < 0 THEN BEGIN(* forward jump *)
      Gen1 (x.adr, x.offs, node); (* Tjmp *)
      loc := -2-(pc-1)
    END ELSE BEGIN(* backward jump *)
      Gen1 (x.adr, loc, node);
      FixLinkWith (x.offs, loc);
    END
  END ELSE IF loc>=0 THEN
    FixLinkWith (x.offs, loc); (* Fjmp *)
  FixLink (x.inx) (* Fjmp *)
END;
	
PROCEDURE Jncc (VAR x: Item; VAR loc: BLabel; node: OPT.Nodes);
VAR
  j: INTEGER;
BEGIN
  IF x.mode <> omCoc THEN BEGIN
    ASSERT (x.typ.form = sfBool);
    load (x);
    GenFlags (rccmp, x, VTrue);
    SetCC (x, EQ)
  END;
  IF x.adr <> always THEN BEGIN
    IF x.adr = never THEN
      j := always
    ELSE
      j := InvCCtab [(x.adr - EQ) DIV 32];		(* jncc bug *)
    IF loc < 0 THEN BEGIN(* forward jump *)
      Gen1 (j, x.inx, node); (* Fjmp *)
      loc := -2-(pc-1)
    END ELSE BEGIN(* backward jump *)
      Gen1 (j, loc, node);
      FixLinkWith (x.inx, loc) (* Fjmp *)
    END;
  END ELSE IF loc>=0 THEN
    FixLinkWith (x.inx, loc); (* Fjmp *)
  FixLink (x.offs) (* Tjmp *)
END;
				
PROCEDURE CompleteSL (VAR x: Item); 
(* complete static link *)
VAR
  n: INTEGER;
  sl: Item;
BEGIN
  IF x.mnolev > 0 THEN BEGIN
    n := level - x.mnolev;
    IF (x.mode IN [omVar, omVarPar]) AND (n > 0) THEN BEGIN
      sl.mode := omRegRel;
      sl.typ := OPT.ptrtyp;
      sl.mnolev := level;
      sl.inx := none;
      sl.adr := FP;
      sl.offs := StaticLinkAdr;
      sl.node := NIL;
      IF n > 1 THEN BEGIN
        load (sl);
        sl.mode := omRegRel;
        sl.offs := StaticLinkAdr;
        WHILE n > 2 DO BEGIN
          load (sl);
          sl.mode := omRegRel;
          sl.offs := StaticLinkAdr;
          DEC (n)
        END
      END;
      load (sl);
      x.mode := omRegRel;
      x.offs := x.adr;
      x.adr := sl.adr;
      x.inx := none
    END
  END
END;

PROCEDURE loadAdr (VAR x: Item);
VAR
  typ: OPT.Structs;
BEGIN
  ASSERT (x.mode IN [omAbs, omVar, omVarPar, omRegRel]);
  typ := x.typ;
  IF (x.mode = omVarPar) OR ((x.mode = omVar)AND(typ.comp IN [csOpenArr, csSDynArr])) THEN	BEGIN(*indirect*)
    CompleteSL (x);
    x.typ := OPT.ptrtyp;
    IF x.mode = omVarPar THEN BEGIN(* x is from the own procedure *)
      x.mode := omRegRel;
      x.offs := x.adr;
      x.adr := FP;
      x.inx := none
    END;
    GenLoad (rcld, x, x)
  END ELSE BEGIN(* Abs, Var, RegRel *)
    CompleteSL (x);
    x.typ := OPT.ptrtyp;
    GenLoad (rclea, x, x)
  END;
  x.offs := 0;
  x.inx := none;
  x.scale := noScale;
  x.typ := typ
END;

PROCEDURE loadf (VAR x: Item);
BEGIN
  IF x.mode <> omReg THEN BEGIN
    IF x.mode = omVarPar THEN BEGIN
      loadAdr (x);
      x.mode := omRegRel
    END ELSE
      CompleteSL (x);
    GenLoad (fload, x, x)
  END
END;
	
PROCEDURE load (VAR x: Item);
BEGIN
  IF x.mode <> omReg THEN BEGIN
    IF x.mode = omVarPar THEN BEGIN
      loadAdr (x);
      x.mode := omRegRel
    END ELSE
      CompleteSL (x);
    GenLoad (rcld, x, x)
  END
END;
	
PROCEDURE NextWord(VAR z: Item; y: Item);	(* z := 4[y] *)
BEGIN
  z := y;
  CASE z.mode OF
    omReg        :  z.adr := y.offs;
    omVar, omAbs : INC(z.adr, 4);
    omRegRel     : INC(z.offs, 4);
  END;
END;
	
PROCEDURE DeRef (VAR x: Item);
VAR
  btyp: OPT.Structs;
BEGIN
  btyp := x.typ.BaseTyp;
  load (x);
  x.mode := omRegRel;
  x.offs := 0;
  x.inx := none;
  x.scale := noScale;
  IF btyp<>NIL THEN BEGIN
    x.typ := btyp;
    IF btyp.comp=csOpenArr THEN BEGIN	(*compatibility, point to the first elem in array*)
      x.offs := ArrBlkFirstElem + 8 + 8*_SYSTEM._DIV((1+btyp.n),2);
      x.descReg := x.adr;
      x.descOffs := ArrBlkLastDim + 4*btyp.n
    END ELSE IF btyp.comp=csStaticArr THEN 	(*compatibility, point to the first elem in array*)
      x.offs := ArrBlkFirstElem + 8
    ELSE IF btyp.comp = csDynArr THEN	;(* do nothing! Deref on demand*)
  END
END;

PROCEDURE GetTdAdr (VAR typ: OPT.Structs; VAR tag: Item);
BEGIN
  ASSERT (typ <> NIL);
  tag.mode := omAbs;
  tag.mnolev := typ.mno;
  tag.typ := OPT.ptrtyp;
  tag.inx := none;
  tag.adr := 0;
  tag.node := NIL;
  tag.adr := typ.tdadr
END;

PROCEDURE TypeTest (VAR x: Item; testtyp: OPT.Structs; guard, equal, varRec: BOOLEAN);
VAR
  testTag, t0, t1, z: Item;
BEGIN
  ASSERT (testtyp <> NIL);
  IF NOT guard OR (OPM.typchk IN OPM.options) THEN BEGIN
    z := x;
    IF varRec THEN BEGIN(* type descriptor is on the stack *)
      CompleteSL (z);
      IF z.mode IN [omVar, omVarPar] THEN BEGIN
        INC (z.adr, 4);
        z.mode := omVar  (* type descriptor is on the stack (hidden parameter) *)
      END ELSE BEGIN
        ASSERT (z.mode = omRegRel);
        INC (z.offs, 4)
      END
    END ELSE BEGIN(* pointer *)
      IF testtyp.form = sfPointer THEN BEGIN
        testtyp := testtyp.BaseTyp;
        DeRef (z)
      END;
      z.offs := -4;
      IF OPM.TDMask <> 0 THEN BEGIN
        t1 := z; load(t1);
        MakeCon(t0, OPT.linttyp, OPM.TDMask);
        Gen3(rcand, z, t1, t0);
      END;
    END;
    IF NOT equal THEN BEGIN
      DeRef (z);
      z.offs := z.offs + BaseTypeOffs - 4 * testtyp.extlev
    END;
    load (z);
    GetTdAdr (testtyp, testTag);
    load (testTag);
    GenFlags (rccmp, z, testTag);
    IF equal THEN
      Gen1 (rctne, OPL.EqualGuardTrap, NIL)
    ELSE IF guard THEN
      Gen1 (rctne, OPL.GuardTrap, NIL)
    ELSE
      SetCC (x, EQ);
  END
END;

PROCEDURE FNProcedure(VAR proc: Item; n: OPT.Nodes);
VAR
  obj: OPT.Objects;
  index: LONGINT;
BEGIN
  obj := n.obj;
  proc.mode := obj.
  mode; proc.node := n;
  proc.mnolev := obj.mnolev;
  proc.inx := none;
  IF  NOT  (proc.mode IN [omTProc, omTyp]) THEN BEGIN(* external, local procedure *)		(* was active only, seems safe *)
    IF (proc.mode IN [omXProc]) AND (proc.mnolev < 0) AND (_SYSTEM._DIV(obj.adr,$10000) = 0) THEN BEGIN
      OPLi386.NewLink (-obj.mnolev, obj.adr, index);
      obj.adr := index * $10000 + obj.adr (* obj.adr = link index * 10000H + entry number *)
    END;
    proc.adr := obj.adr 
  END ELSE (* proc.mode = TProc *)
    proc.offs := MethodOffs - 4 * _SYSTEM._DIV(obj.adr,$10000);
END;
	
PROCEDURE CheckOverflow (VAR x: Item);
BEGIN
  IF (OPM.ovflchk IN OPM.options) THEN Gen1 (rcto, OPL.OverflowTrap, NIL);
END;
			
PROCEDURE CheckIntRange (VAR x: Item; form: SHORTINT);
BEGIN
END;

PROCEDURE CheckRealRange (VAR x: Item; form: SHORTINT);
BEGIN
END;

PROCEDURE CheckIndex (VAR index, len: Item);
BEGIN
  IF (index.mode=omCon)AND(len.mode<>omCon) THEN BEGIN	(*reversed check*)
    IF (len.mode<>omReg) THEN load(len);
    GenFlags (rccmp, len, index);
    Gen1 (rctle, OPL.RangeTrap, NIL)
  END ELSE BEGIN
    IF (index.mode<>omReg) THEN load(index);
    IF (len.mode<>omCon) THEN load(len);
    GenFlags (rccmp, index, len);
    Gen1 (rctae, OPL.RangeTrap, NIL);
  END;
END;
	
PROCEDURE ChangeType (VAR item: Item; form: SHORTINT);
BEGIN
  CASE form OF
      sfByte: item.typ:= OPT.bytetyp;
      sfBool: item.typ:= OPT.booltyp;
      sfChar: item.typ:= OPT.chartyp;
      sfSInt: item.typ:= OPT.sinttyp;
      sfInt: item.typ:= OPT.inttyp;
      sfLInt: item.typ:= OPT.linttyp;
      sfReal: item.typ:= OPT.realtyp;
      sfLReal: item.typ:= OPT.lrltyp;
      sfSet: item.typ:= OPT.settyp;
      sfPointer: item.typ:= OPT.ptrtyp;
      sfProcTyp: item.typ:= OPT.notyp;
  ELSE
    HALT (BUG)
  END
END;

PROCEDURE SetElem(VAR z, x: Item); (* z := {x} *)
BEGIN
  MakeCon (z, OPT.settyp, 0);
  load (z);
  IF x.typ.form <> sfLInt THEN
    Convert (x, sfLInt)
  ELSE
    load (x);
  Gen3 (rcbts, z, z, x)
END;

PROCEDURE Convert (VAR x: Item; form: SHORTINT);
VAR
  t: Item;
  xform: SHORTINT;

  PROCEDURE IntToInt (VAR z, x: Item; xform, zform: SHORTINT);
  BEGIN
    ASSERT ((xform IN intSet) AND (zform IN intSet) AND (x.typ.form IN intSet) AND (z.typ.form IN intSet));
    IF xform < zform THEN BEGIN(* smaller to bigger *)
      IF x.mode = omVarPar THEN BEGIN
        loadAdr (x);
        x.mode := omRegRel;
      END ELSE
        CompleteSL (x);
      ChangeType (x, zform);
      IF x.mode <> omCon THEN	BEGIN														(* const conversion *)
        IF zform = sfInt THEN
          GenLoad (rcldbw, z, x)
        ELSE IF xform = sfInt THEN
          GenLoad (rcldwdw, z, x)
        ELSE
          GenLoad (rcldbdw, z, x);
      END ELSE
        z := x;
    END ELSE BEGIN(* bigger to smaller *)
      IF x.mode = omReg THEN BEGIN
        ChangeType (x, zform);
        Gen2 (rcshort, z, x)
      END;
      CheckIntRange (x, zform)
    END;
    (* z.typ will be changed by Convert *)
  END;
		
  PROCEDURE IntToReal (VAR z, x: Item; xform, zform: SHORTINT);
  (* integer to real convertion *)
    VAR  temp, cons: Item;
  BEGIN
    IF x.mode = omVarPar THEN BEGIN
      loadAdr (x);
      x.mode := omRegRel;
    END ELSE
      CompleteSL (x);
    ChangeType (x, sfLInt);
    IF xform = sfSInt THEN
      GenLoad (rcldbdw, x, x)
    ELSE IF xform = sfInt THEN
      GenLoad (rcldwdw, x, x)
    ELSE
      load (x);
    GenPush (x); (* store x on the stack as a temporary variable *)
    temp.mode := omRegRel;
    temp.typ := OPT.linttyp;
    temp.adr := ESP;
    temp.offs := 0;
    temp.inx := none;
    temp.node := NIL;
    ChangeType (z, zform);
    ChangeType (temp, zform);
    GenLoad (fild, z, temp);
    MakeCon (cons, OPT.linttyp, 4);
    temp := Stack;
    Gen3 (rcadd, temp, Stack, cons) (* free temporary variable space *)
  END;
		
  PROCEDURE IntToCharByte (VAR z, x: Item; form: SHORTINT);
  BEGIN
    IF (x.typ.form <> sfSInt) AND (x.mode = omReg) THEN BEGIN
      ChangeType (x, form);
      Gen2 (rcshort, z, x)
    END;
    CheckIntRange (x, form)
  END;
		
  PROCEDURE RealToReal (VAR z, x: Item; xform, zform: SHORTINT);
  (* real to real convertion *)
  BEGIN
    IF xform < zform THEN BEGIN(* smaller to bigger *)
      z := x;
      loadf (z);
    END ELSE BEGIN(* bigger to smaller *)
      z := x;
      loadf (z);
      CheckRealRange (x, sfReal)
    END
  END;
		
  PROCEDURE RealToInt(VAR z, x: Item; xform, zform: SHORTINT);
  (* real to integer convertion *)
  BEGIN
    loadf (x);
    ChangeType (x, sfLInt);
    Gen2 (rcentier, z, x)
    (* z.typ will be changed by Convert *) 
  END;
		
  PROCEDURE CharToInt (VAR z, x: Item; zform: SHORTINT);
  (* character to integer convertion *)
  BEGIN
    IF x.mode = omVarPar THEN BEGIN
      loadAdr (x);
      x.mode := omRegRel
    END ELSE
      CompleteSL (x);
    ChangeType (x, zform);
    IF zform = sfInt THEN
      GenLoad (rcldbwu, z, x)
    ELSE IF zform = sfLInt THEN
      GenLoad (rcldbdwu, z, x);
    (* z.typ will be changed by Convert *)
  END;
	
  PROCEDURE ByteToInt (VAR z, x: Item; zform: SHORTINT);
  (* byte to integer convertion *)
  BEGIN
    IF x.mode = omVarPar THEN BEGIN
      loadAdr (x);
      x.mode := omRegRel
    END ELSE
      CompleteSL (x);
    ChangeType (x, zform);
    IF zform = sfInt THEN
      GenLoad (rcldbw, z, x)
    ELSE IF zform = sfLInt THEN
      GenLoad (rcldbdw, z, x);
    (* z.typ will be changed by convert *)
  END;
		
BEGIN (* Convert *)
  IF x.mode = omCoc THEN BEGIN
    t := x;
    MoveCoc (x, t)
  END;
  t := x;
  xform := x.typ.form;
  IF xform IN intSet THEN BEGIN
    IF form IN intSet THEN
      IntToInt (x, t, xform, form)
    ELSE IF form IN realSet THEN
      IntToReal (x, t, xform, form)
    ELSE IF form IN [sfChar, sfByte] THEN
      IntToCharByte (x, t, form)
    ELSE
      HALT (BUG);
  END ELSE IF xform IN realSet THEN BEGIN
    IF form IN realSet THEN
      RealToReal (x, t, xform, form)
    ELSE IF form = sfLInt THEN
      RealToInt (x, t, xform, form)
    ELSE IF form IN [sfInt, sfSInt] THEN BEGIN
      RealToInt (x, t, xform, form);
      ChangeType (x, sfLInt);
      t := x;
      IntToInt (x, t, sfLInt, form)
    END ELSE IF form IN [sfChar, sfByte] THEN BEGIN
      RealToInt (x, t, xform, sfLInt);
      ChangeType (x, sfLInt);
      t := x;
      IntToCharByte (x, t, form)
    END ELSE
      HALT (BUG);
  END ELSE IF (xform IN [sfByte, sfChar]) AND (form IN [sfByte,sfChar]) THEN (* ok, type is changed by OPV *)
  ELSE IF xform = sfChar THEN BEGIN
    ASSERT (form IN intSet + realSet);											(* shift error *)
    IF form IN intSet THEN
      CharToInt (x, t, form)
    ELSE BEGIN	(* form IN realSet *)
      CharToInt(x, t, sfLInt);
      t:=x;
      IntToReal(x, t, sfLInt, form)
    END
  END ELSE IF xform = sfByte THEN BEGIN
    IF form IN intSet THEN
      ByteToInt (x, t, form)
    ELSE IF form IN realSet THEN
      RealToInt (x, t, xform, form)
    ELSE
      HALT (BUG);
  END ELSE IF (form = sfProcTyp) AND (xform <> sfNoTyp) THEN
    HALT (BUG);
  ChangeType (x, form)
END;

PROCEDURE Include (VAR z, x: Item; incl: BOOLEAN); (* INCL / EXCL (z, x) *)
VAR
  y,  v: Item;

  FUNCTION exp2 (x: LONGINT): LONGINT;
  VAR
    i: LONGINT;
  BEGIN
    ASSERT (x >= 0);
    i := 1;
    WHILE x > 0 DO BEGIN
      i := i * 2;
      DEC  (x)
    END;
    RESULT:= i
  END;

BEGIN
  v := z;
  load (v);
  y := v;
  IF incl THEN BEGIN
    IF x.mode = omCon THEN BEGIN
      IF x.adr = MaxSet THEN
        x.adr := MinLInt
      ELSE
        x.adr := exp2 (x.adr);
      Gen3 (rcor, v, y, x)
    END ELSE BEGIN
      IF x.typ.form <> sfLInt THEN Convert (x, sfLInt);
      load (x);
      Gen3 (rcbts, v, y, x)
    END
  END ELSE BEGIN
    IF x.mode = omCon THEN BEGIN
      IF x.adr = MaxSet THEN
        x.adr := MaxLInt
      ELSE
        x.adr := -1 - exp2 (x.adr);
      Gen3 (rcand, v, y, x)
    END ELSE BEGIN
      IF x.typ.form <> sfLInt THEN Convert (x, sfLInt);
      load (x);
      Gen3 (rcbtr, v, y, x)
    END
  END;
  IF z.mode = omVarPar THEN BEGIN
    loadAdr (z);
    z.mode := omRegRel
  END ELSE
    CompleteSL (z);
  GenStore (rcstore, z, v)
END;

PROCEDURE MoveCoc (VAR z, y: Item); (* z := Coc (y) *)
VAR
  L: BLabel;
  phiItem: Item;
BEGIN
  ASSERT (y.mode = omCoc);
  IF (y.offs = aNil) AND (y.inx = aNil) THEN (* no jump links *)
    GenSetC (rcsete + y.adr - EQ, z)
  ELSE BEGIN
    Jncc (y, y.inx, NIL);
    GenLoad (rcld, z, VTrue);
    phiItem := z;
    L := aNil;
    FNJmp (L, NIL);
    FixLink (y.inx); (* Fjmp fixup *)
    GenLoad (rcld, z, VFalse);
    GenPhi (phiItem.adr, z.adr); (* registers in z and phiItem must be the same when generating Intel code *)
    FixLink (L)
  END;
END;

PROCEDURE Relation (VAR x: Item);
VAR
  y: Item;
BEGIN 
  IF x.mode = omCoc THEN BEGIN
    y := x;
    MoveCoc (x, y)
  END
END;

PROCEDURE FNCmp(VAR z, x, y: Item; rel: INTEGER); (* z := cmp (x, y) *)
  VAR xform: INTEGER;
BEGIN
  xform := x.typ.form;
  IF xform IN [sfByte, sfChar..sfLInt, sfSet, sfNilTyp, sfPointer, sfProcTyp] THEN BEGIN
    load (x);
    IF y.mode <> omCon THEN load (y);
    GenFlags (rccmp, x, y)
  END ELSE IF xform IN realSet THEN BEGIN
    loadf (y);
    loadf (x);
    GenFlags (fcmp, x, y)
  END ELSE IF xform = sfBool THEN BEGIN(* only eq or neq *)
    load (x);
    IF y.mode = omCoc THEN BEGIN
      z := y;
      MoveCoc (y, z)
    END ELSE IF y.mode <> omCon THEN
      load (y);
    GenFlags (rccmp, x, y)
  END ELSE BEGIN(* strings *)
    ASSERT ((xform = sfString) OR (x.typ.comp IN ArraySet));
    loadAdr (x);
    loadAdr (y);
    GenFlags (rccmpString, x, y);
    xform := sfString;
  END;
  ASSERT (rel IN [veql, vneq, vlss, vleq, vgtr, vgeq]);
  IF xform IN [sfChar, sfString, sfReal, sfLReal] THEN
    SetCC (z, CCtab [1, rel-veql])
  ELSE
    SetCC (z, CCtab [0, rel-veql]);
END;

PROCEDURE FNIn (VAR z, x, y: Item); (* z := x IN y *)
VAR
  t: Item;
BEGIN
  IF (y.mode <> omCon) AND (y.typ.form <> sfSet) THEN BEGIN
    t := y;
    SetElem (y, t)
  END ELSE
    load (y);
  IF x.mode <> omCon THEN BEGIN
    IF x.typ.form <> sfLInt THEN
      Convert (x, sfLInt)
    ELSE
      load (x);
    GenFlags (rcbt, y, x);
    SetCC (z, CS)
  END ELSE BEGIN
    x.adr := _SYSTEM.LSH(1,x.adr);
    x.typ := OPT.linttyp;
    GenFlags (rctest, y, x);
    SetCC (z, NE)	(*jump on ZF flag*)
  END
END;

PROCEDURE SetRange(VAR z, x, y: Item); (* z := {x..y} *)
VAR
  t, xShifted, yShifted, VmaxSet, fullSet: Item;
  val: LONGINT;
BEGIN
  z.typ := OPT.settyp;
  MakeCon (fullSet, OPT.linttyp, -1);
  MakeCon (VmaxSet, OPT.linttyp, MaxSet);
  IF x.mode = omCon THEN BEGIN
    ASSERT (y.mode <> omCon);
    load (y);
    VmaxSet.typ := y.typ;
    load (VmaxSet);
    t := y;
    Gen3 (rcsub, y, VmaxSet, t);
    load (fullSet);
    Gen3 (rcshr, z, fullSet,  y);
    IF x.adr <> 0 THEN BEGIN
      IF x.adr = MaxSet THEN
        x.adr := MIN (_LONGINTS)
      ELSE BEGIN
        val := 2;
        WHILE x.adr > 1 DO BEGIN
          val := val*2;
          DEC (x.adr)
        END;
        x.adr := -val
      END;
      t := z; Gen3 (rcand, z, t, x)
    END
  END ELSE IF y.mode = omCon THEN BEGIN
    load (x);
    load (fullSet);
    Gen3 (rcsal, z, fullSet, x);
    IF y.adr <> MaxSet THEN BEGIN
      IF y.adr = MaxSet-1 THEN
        y.adr := MAX (_LONGINTS)
      ELSE BEGIN
        val := 2;
        WHILE y.adr > 0 DO BEGIN
          val := val*2;
          DEC (y.adr)
        END;
        y.adr := val-1
      END;
      t := z; Gen3 (rcand, z, t, y)
    END
  END ELSE BEGIN(* x, y # constant *)
    load (x);
    t := fullSet;
    load (t);
    xShifted.typ := OPT.linttyp;
    xShifted.node := NIL;
    Gen3 (rcsal, xShifted, t, x);
    load (y);
    VmaxSet.typ := y.typ;
    load (VmaxSet);
    t := y;
    Gen3 (rcsub, y, VmaxSet, t);
    load (fullSet);
    yShifted.typ := OPT.linttyp;
    yShifted.node := NIL;
    Gen3 (rcshr, yShifted, fullSet, y);
    Gen3 (rcand, z, xShifted, yShifted)
  END;
END;

PROCEDURE FNNot(VAR z, x: Item); (* z := NOT (x) *)
VAR
  temp: LONGINT;
BEGIN
  IF x.mode = omCoc THEN BEGIN
    z := x;
    z.adr := InvCCtab [(z.adr - EQ) DIV 32];
    temp := z.offs;
    z.offs := z.inx;
    z.inx := temp; (* permute Fjmp and Tjmp *)
  END ELSE BEGIN
    load (x);
    GenFlags (rccmp, x, VFalse);
    SetCC (z, EQ)
  END;
END;
	
PROCEDURE FNNeg(VAR z, x: Item); (* z := -x *)
BEGIN
  ASSERT ( (x.typ.form = sfSet) OR (x.typ.form IN intSet) OR (x.typ.form IN realSet) );
  IF x.typ.form = sfSet THEN BEGIN
    load (x);
    Gen2 (rcnot, z, x)
  END ELSE IF x.typ.form IN intSet THEN BEGIN
    load (x);
    Gen2 (rcneg, z, x)
  END ELSE BEGIN(* x.typ.form = realSet *)
    loadf (x);
    Gen2 (fchs, z, x)
  END
END;
	
PROCEDURE AbsVal(VAR z, x: Item); (* z := ABS (x) *)
BEGIN
  ASSERT ( (x.typ.form IN intSet) OR (x.typ.form IN realSet) );
  IF x.typ.form IN intSet THEN BEGIN
    load (x);
    Gen2 (rcabs, z, x);
    IF x.typ.form IN [sfInt, sfLInt] THEN OPL.InstrTab [pc-1].hint := OPL.useEAX;
  END ELSE BEGIN(* x.typ.form IN realSet *)
    loadf (x);
    Gen2 (fabs, z, x)
  END
END;
	
PROCEDURE FNCap(VAR z, x: Item); (* z := CAP (x) *)
  VAR c: Item;
BEGIN
  MakeCon (c, OPT.chartyp, $5F);
  load (x);
  Gen3 (rcand, z, x, c)
END;
	
PROCEDURE FNOdd(VAR z, x: Item); (* z := ODD (x) *)
VAR
  c: Item;
BEGIN
  MakeCon (c, OPT.sinttyp, 1);
  load (x);
  Gen3 (rcand, z, x, c)
END;

PROCEDURE FNAsh(VAR z, x, y: Item); (* z := ASH (x, y) *)
VAR
  L, Lc: BLabel;
  t, phiItem: Item;
BEGIN
  IF y.mode = omCon THEN BEGIN
    IF x.mode = omCon THEN
      MakeCon (z, OPT.linttyp, ASH (x.adr, y.adr))
    ELSE IF y.adr > 0 THEN BEGIN
      load (x);
      Gen3 (rcsal, z, x, y)
    END ELSE IF y.adr < 0 THEN BEGIN
      y.adr := _SYSTEM.ABS(y.adr);
      load (x);
      Gen3 (rcsar, z, x, y);
    END;
  END ELSE BEGIN
    load (x);
    load (y);
    z.typ := OPT.linttyp;
    MakeCon (t, OPT.linttyp, 0);
    GenFlags (rccmp, y, t);
    Lc := -2-pc;
    Gen1 (rcjl, aNil, NIL);
    Gen3 (rcsal, z, x, y);
    L := -2-pc;
    Gen1 (rcjmp, aNil, NIL);
    phiItem := z;
    FixLink (Lc);
    Gen2 (rcneg, t, y);
    Gen3 (rcsar, z, x, t);
    GenPhi (phiItem.adr, z.adr); (* z and phiItem must be in the same register when generating target code *)
    FixLink (L)
  END
END;

PROCEDURE FNAdd(VAR z, x, y: Item; f: INTEGER); (* z := x + y *)
BEGIN
  ASSERT ( (f IN realSet) OR (f = sfSet) OR (f IN intSet) );
  IF f IN realSet THEN BEGIN
    loadf (x);
    loadf (y);
    Gen3 (fadd, z, x, y)
  END ELSE IF f = sfSet THEN BEGIN
    IF x.mode = omCon THEN BEGIN
      load (y);
      Gen3 (rcor, z, y, x)
    END ELSE BEGIN
      load (x);
      IF y.mode <> omCon THEN load (y);
      Gen3 (rcor, z, x, y)
    END;
  END ELSE BEGIN(* f IN intSet *)
    IF x.mode = omCon THEN BEGIN
      load (y);
      Gen3 (rcadd, z, y, x)
    END ELSE BEGIN
      load (x);
      IF y.mode <> omCon THEN load (y);
      Gen3 (rcadd, z, x, y)
    END;
    CheckOverflow (z);
  END
END;

PROCEDURE FNSub(VAR z, x, y: Item; f: INTEGER); (* z := x - y *)
VAR
  t: Item;
BEGIN
  ASSERT ( (f IN realSet) OR (f = sfSet) OR (f IN intSet) );
  IF f IN realSet THEN BEGIN
    loadf (x);
    loadf (y);
    Gen3 (fsub, z, x, y)
  END ELSE IF f = sfSet THEN BEGIN
    load (x);
    IF y.mode <> omCon THEN BEGIN
      load (y);
      Gen2 (rcnot, t, y)
    END ELSE BEGIN
      y.adr := -1- y.adr;
      t := y
    END;
    z := x;
    Gen3 (rcand, z, x, t)
  END ELSE BEGIN(* f IN intSet *)
    load (x);
    IF y.mode <> omCon THEN load (y);
    Gen3 (rcsub, z, x, y);
    CheckOverflow (z)
  END
END;

PROCEDURE FNMul(VAR z, x, y: Item; f: INTEGER); (* z := x * y *)
VAR
  val, exp: LONGINT;
  shift: Item;
BEGIN
  ASSERT ( (f = sfSet) OR (f IN intSet) OR (f IN realSet) );
  IF f IN realSet THEN BEGIN
    loadf (x);
    loadf (y);
    Gen3 (fmul, z, x, y)
  END ELSE IF f = sfSet THEN BEGIN
    IF x.mode = omCon THEN BEGIN
      load (y);
      Gen3 (rcand, z, y, x)
    END ELSE BEGIN
      load (x);
      IF (y.mode <> omReg) AND (y.mode <> omCon) THEN load (y);
      Gen3 (rcand, z, x, y)
    END
  END ELSE BEGIN(* f IN intSet *)
  (*  The shift-left check is necessary because several procedures in OPC call Mul directly. *)
    MakeCon (shift, OPT.linttyp, 0);
    IF x.mode = omCon THEN BEGIN
      val := x.adr;
      exp := 0;
      ASSERT (val <> 0);
      WHILE  NOT ODD (val) DO BEGIN
        val := _SYSTEM._DIV(val,2);
        INC (exp)
      END;
      load (y);
      IF (val = 1) AND  NOT (OPM.ovflchk IN OPM.options) THEN BEGIN(* shift *)
        IF exp <> 0 THEN BEGIN
          shift.adr := exp;
          Gen3 (rcsal, z, y, shift)
        END ELSE
          z := y;
      END ELSE
        Gen3 (rcmul, z, y, x);
    END ELSE IF y.mode = omCon THEN BEGIN
      val := y.adr;
      exp := 0;
      ASSERT (val <> 0);
      WHILE  NOT ODD (val) DO BEGIN
        val := _SYSTEM._DIV(val,2);
        INC (exp)
      END;
      load (x);
      IF (val = 1) AND  NOT (OPM.ovflchk IN OPM.options) THEN BEGIN(* shift *)
        IF exp <> 0 THEN BEGIN
          shift.adr := exp;
          Gen3 (rcsal, z, x, shift)
        END ELSE
          z := x;
      END ELSE
        Gen3 (rcmul, z, x, y);
    END ELSE BEGIN
      load (x);
      load (y);
      Gen3  (rcmul, z, x, y)
    END;
    CheckOverflow (z)
  END
END;

PROCEDURE FNDiv(VAR z, x, y: Item; f: INTEGER); (* z := x DIV y; z := x / y *)
BEGIN
  ASSERT ( (f IN realSet) OR (f = sfSet) OR (f IN intSet) );
  IF f IN realSet THEN BEGIN
    loadf (x);
    loadf (y);
    Gen3 (fdiv, z, x, y)
  END ELSE IF f = sfSet THEN BEGIN
    IF x.mode = omCon THEN BEGIN
      load (y);
      Gen3 (rcxor, z, y, x)
    END ELSE BEGIN
      load (x);
      IF (y.mode <> omReg) AND (y.mode <> omCon) THEN load (y);
      Gen3 (rcxor, z, x, y)
    END
  END ELSE BEGIN(* f IN intSet *) (* front end checks for shift right possibilities *)
    load (x);
    IF y.mode <> omCon THEN load (y);
    Gen3 (rcdiv, z, x, y)
  END
END;

PROCEDURE FNMod(VAR z, x, y: Item); (* z := x MOD y *)
BEGIN 
  load (x);
  IF y.mode <> omCon THEN load (y); (* front end checks for shift left possibilities *)
  Gen3 (rcmod, z, x, y)
END;

PROCEDURE CondAnd (VAR x: Item); (* prepares conditional and *)
VAR
  jcc: INTEGER;
BEGIN
  IF x.mode = omCoc THEN
    jcc := InvCCtab [(x.adr - EQ) DIV 32]
  ELSE BEGIN(* form = Bool, mode # Con *)
    ASSERT ( (x.typ.form = sfBool) AND (x.mode <> omCon) );
    load (x);
    GenFlags (rccmp, x, VTrue);
    SetCC (x, NE);
    jcc := rcjne
  END;
  Gen1 (jcc, x.inx, NIL); (* ... Fjmp *) 
  x.inx := -2-(pc-1);
  FixLink (x.offs) (* Tjmp here *)
END;

PROCEDURE FNAnd(VAR z, x, y: Item); (* conditional and *)
BEGIN
  z := x;
  IF y.mode <> omCoc THEN BEGIN
    IF y.mode = omCon THEN BEGIN
      IF y.adr = Otrue THEN
        SetCC (y, always)
      ELSE
        SetCC (y, never);
    END ELSE BEGIN
      load (y);
      GenFlags (rccmp, y, VTrue);
      SetCC (y, EQ)
    END
  END;
  IF y.inx <> aNil THEN z.inx := MergedLinks (z.inx, y.inx); (* merge Fjmp *)
  z.adr := y.adr;
  z.offs := y.offs; (* Tjmp *)
END;

PROCEDURE CondOr (VAR x: Item); (* prepares conditional or *)
VAR
  jcc: INTEGER;
BEGIN
  IF x.mode = omCoc THEN
    jcc := x.adr
  ELSE BEGIN(* form = Bool, mode # Con *)
    ASSERT ( (x.typ.form = sfBool) AND (x.mode <> omCon) );
    load (x);
    GenFlags (rccmp, x, VTrue);
    SetCC (x, EQ);
    jcc := rcje
  END;
  Gen1 (jcc, x.offs, NIL); (* ... Tjmp *)
  x.offs := -2-(pc-1);
  FixLink (x.inx) (* Fjmp here *)
END;

PROCEDURE FNOr(VAR z, x, y: Item); (* conditional or *)
BEGIN
  z := x;
  IF y.mode <> omCoc THEN BEGIN
    IF y.mode = omCon THEN BEGIN
      IF y.adr = Otrue THEN
        SetCC (y, always)
      ELSE
        SetCC (y, never);
    END ELSE BEGIN
      load (y);
      GenFlags (rccmp, y, VTrue);
      SetCC (y, EQ)
    END;
  END;
  IF y.offs <> aNil THEN z.offs := MergedLinks (x.offs, y.offs); (* merge Tjmp *)
  z.adr := y.adr;
  z.inx := y.inx (* Fjmp *)
END;

PROCEDURE SYSmop (VAR z, x: Item; subcl: SHORTINT; typ: OPT.Structs);
VAR
  xform, typform: SHORTINT;
  t, Vconst: Item;
BEGIN
  CASE subcl OF
     yadr:  loadAdr (z);
     ycc:
            BEGIN
              z := x;
              OPM.err (200) (* !!later: not implemented *)
            END;
     yval:
            BEGIN
              xform := x.typ.form;
              typform := typ.form;
              MakeCon (Vconst, OPT.linttyp, 4);
              IF (xform IN  [sfByte, sfChar, sfSInt, sfInt, sfLInt, sfSet, sfPointer, sfProcTyp])
              AND (typform IN realSet) AND (x.mode = omReg) THEN BEGIN
                (* load (x) *);
                GenPush (x);
                t.mode := omRegRel;
                t.adr := ESP;
                t.offs := 0;
                t.inx := none;
                t.typ := OPT.realtyp;
                t.node := NIL;
                GenLoad (fload, z, t);
                t := Stack;
                Gen3 (rcadd, t, t, Vconst)
              END ELSE IF (xform IN realSet) AND (typform IN [sfByte, sfChar, sfSInt, sfInt, sfLInt, sfSet, sfPointer, sfProcTyp])
              AND (x.mode = omReg) THEN BEGIN
                t := Stack;
                Gen3 (rcsub, t, t, Vconst);
                t.mode := omRegRel;
                t.adr := ESP;
                t.offs := 0;
                t.inx := none;
                t.typ := OPT.realtyp;
                t.node := NIL;
                loadf (x);
                GenStore (fstore, t, x);
                t.typ := typ;
                GenLoad (rcld, z, t);
                t := Stack;
                Gen3 (rcadd, t, t, Vconst)
              END;
              IF (z.typ.size <> typ.size) AND ((x.mode <> omCon) OR (typ.size > 4)) THEN OPM.err(-304);
              z.typ := typ;
            END;
      yget8, yget16, yget32:
              BEGIN
                load(z);
                z.mode := omRegRel;
                z.offs := 0;
                z.inx := none;
                z.typ := typ;
                load(z);	(*load value in register, otherwise SYSTEM.VAL might change the size of the memory access *)
              END;
    END;
END;
	
PROCEDURE Mem (VAR x: Item); (* x := MEM [x] *)
BEGIN
  IF x.mode = omCon THEN BEGIN
    x.mode := omAbs;
    x.inx := none
  END ELSE BEGIN
    load (x);
    x.mode := omRegRel;
    x.offs := 0;
    x.inx := none
  END
END;

PROCEDURE SYSdop (VAR z, x, y: Item; subcl: SHORTINT);	(* implementation of SYSTEM.BIT, LSH, ROT *)
VAR
  t, phiItem: Item;
  L, Lc: BLabel;
BEGIN 
  CASE subcl OF
    ybit:
            BEGIN
              Mem (x);
              load (x);
              IF y.mode <> omCon THEN load (y);
              GenFlags (rcbt, x, y);
              GenSetC (rcsetc, z)
            END;
    ylsh:
            BEGIN
              IF y.mode = omCon THEN BEGIN
                IF x.mode = omCon THEN
                  MakeCon (z, OPT.linttyp, _SYSTEM.LSH(x.adr, y.adr))										(* shift error *)
                ELSE IF y.adr > 0 THEN BEGIN
                  load (x);
                  Gen3 (rcsal, z, x, y)
                END ELSE IF y.adr < 0 THEN BEGIN
                  y.adr := _SYSTEM.ABS(y.adr);
                  load (x);
                  Gen3 (rcshr, z, x, y)
                END
              END ELSE BEGIN
                load (x);
                load (y);
                MakeCon (t, OPT.linttyp, 0);
                GenFlags (rccmp, y, t);
                Lc := -2-pc;
                Gen1 (rcjl, aNil, NIL);
                Gen3 (rcsal, z, x, y);
                phiItem := z;
                L := -2-pc;
                Gen1 (rcjmp, aNil, NIL);
                FixLink (Lc);
                Gen2 (rcneg, t, y);
                Gen3 (rcshr, z, x, t);
                GenPhi (phiItem.adr, z.adr); (* phiItem, z must have the same register when generating targed code *)
                FixLink (L)
              END
            END;
    yrot:
            BEGIN
              IF y.mode = omCon  THEN BEGIN
                IF x.mode = omCon THEN BEGIN
                  MakeCon (z, x.typ, 0);
                  CASE x.typ.form OF
                    sfByte, sfChar, sfSInt:
                      z.adr := _SYSTEM.ROT(SHORTINT(x.adr),y.adr);
                    sfInt:
                      z.adr := _SYSTEM.ROT(INTEGER(x.adr),y.adr);
                    sfLInt:
                      z.adr := _SYSTEM.ROT(x.adr,y.adr);
                  END
                END ELSE BEGIN
                  load (x);
                  Gen3 (rcrol,  z, x, y)
                END
              END ELSE BEGIN
                load (x);
                load (y);
                Gen3 (rcrol, z, x, y)
              END
            END;
  END
END;
	
PROCEDURE SYSgetput(VAR z, x: Item; getfn: BOOLEAN);	(* SYSTEM.GET, SYSTEM.PUT *)
VAR
  form: SHORTINT;
  constAddr: BOOLEAN;					(* GetPut *)
BEGIN
  IF getfn THEN BEGIN
    constAddr := x.mode = omCon;					(* GetPut *)
    Mem (x);
    x.typ := z.typ;
    form := z.typ.form
  END ELSE BEGIN
    constAddr := z.mode = omCon;					(* GetPut *)
    Mem (z);
    z.typ := x.typ;
    form := x.typ.form
  END;
  IF form IN realSet THEN
    loadf (x)
  ELSE
    load (x);
  OPL.InstrTab [pc-1].abs := constAddr AND getfn;			(* if x is the address and a constant, no fixup *)	(* GetPut *)
  IF z.mode = omVarPar THEN BEGIN
    loadAdr (z);
    z.mode := omRegRel
  END ELSE
    CompleteSL (z);
  IF form IN realSet THEN
    GenStore (fstore, z, x)
  ELSE
    GenStore (rcstore, z, x);
  OPL.InstrTab [pc-1].abs := constAddr AND  NOT getfn			(* if z is the address and a constant, no fixup *)	(* GetPut *)
END;

PROCEDURE SYSportio(VAR z, x: Item; pin: BOOLEAN);
VAR
  Vreg: Item;
BEGIN
  Vreg.mode := omReg;
  Vreg.node := NIL;
  IF pin THEN BEGIN
    IF x.mode <> omReg THEN load(x);		(* load x into reg *)
    OPL.InstrTab [x.adr].hint := OPL.useEDX;	(* let the reg be DX *)
    ASSERT((x.mode = 15) AND (x.typ = OPT.inttyp));
    x.typ := z.typ;		(* generate correct size *)
    GenLoad(rcin, Vreg, x);
    OPL.InstrTab [pc-1].hint := OPL.useEAX;
    IF z.mode = omVarPar THEN BEGIN
      loadAdr(z);
      z.mode := omRegRel		(* generate 0[Reg] where Reg = Address(par) *)
    END ELSE
      CompleteSL(z);
    GenStore(rcstore, z, Vreg)
  END ELSE BEGIN
    IF z.mode <> omReg THEN  load(z);		(* load z into reg *)
    OPL.InstrTab [z.adr].hint := OPL.useEDX;
    ASSERT( z.mode = omReg );
    IF x.mode = omVarPar THEN BEGIN
      loadAdr(x);
      x.mode := omRegRel		(* generate 0[Reg] where Reg = Address(par) *)
    END ELSE
      CompleteSL(x);
    IF x.mode <> omReg THEN GenLoad(rcld, x, x);
    OPL.InstrTab [x.adr].hint := OPL.useEAX;
    ASSERT( x.mode = omReg );
    GenLoad(rcout, Vreg, x);
    OPL.InstrTab [pc-1].dest := z.adr
  END;
END;

PROCEDURE SYSgetputReg (VAR z, x: Item; getrfn: BOOLEAN);
VAR
  Vreg: Item;
BEGIN
  Vreg.mode := omReg;
  Vreg.node := NIL;
  ASSERT(x.typ <> OPT.lrltyp);
  IF getrfn THEN BEGIN
    IF z.typ = OPT.realtyp THEN z.typ := OPT.linttyp;		(*real, move 32-bit*)
    Vreg.adr := x.adr;
    Vreg.typ := z.typ;
    IF z.mode = omVarPar THEN BEGIN
      loadAdr (z);
      z.mode := omRegRel
    END ELSE
      CompleteSL (z);
    GenStore (rcgetReg, z, Vreg);
    OPL.InstrTab [pc-1].hint := OPL.useEAX + x.adr
  END ELSE BEGIN(* putrrn *)
    IF x.typ = OPT.realtyp THEN x.typ := OPT.linttyp;		(*real, move 32-bit*)
    IF x.mode = omCon THEN
      x.typ := OPT.linttyp
    ELSE IF x.mode = omVarPar THEN BEGIN
      loadAdr (x);
      x.mode := omRegRel
    END ELSE
      CompleteSL (x);
    GenLoad (rcputReg, Vreg, x);
    OPL.InstrTab [pc-1].hint := OPL.useEAX + z.adr
  END;
END;

PROCEDURE SYSinterrupt(clear: BOOLEAN);
BEGIN
  IF clear THEN Gen0(rccli) ELSE Gen0(rcsti);
END;

PROCEDURE Msk (VAR z, x, y: Item); (* z := x AND y *)
BEGIN
  y.adr := -1 - y.adr; (* invert y.adr *)
  load (x);
  Gen3 (rcand, z, x, y)
END;

PROCEDURE Field (VAR x: Item; offset: LONGINT); (* Record.field *)
BEGIN
  ASSERT (x.mode IN [omVar, omVarPar, omAbs, omRegRel]);
  IF x.mode IN [omVar, omAbs] THEN
    INC (x.adr, offset)
  ELSE IF x.mode = omVarPar THEN BEGIN
    loadAdr (x);
    x.mode := omRegRel;
    x.offs := offset
  END ELSE (* RegRel *)
    INC (x.offs, offset)
END;

PROCEDURE MakeVar(VAR x: Item; Vvar: OPT.Objects);
BEGIN
  ASSERT(Vvar.mode = omVar{, 100});
  x.mode := omVar;
  x.adr := Vvar.linkadr;
  x.offs := 0;
  x.typ := Vvar.typ;
  x.node := NIL;
  x.mnolev := Vvar.mnolev;
  x.inx := none;
  IF x.typ.comp = csOpenArr THEN BEGIN
    x.descReg := FP;
    x.descOffs := x.adr + (x.typ.n+1) * OPM.LIntSize
  END ELSE
    x.descReg := none;
END;

PROCEDURE loadDynArr(VAR z: Item); FORWARD;

PROCEDURE ArrayLen(VAR len: Item; arr: Item; n: LONGINT);
VAR
  typ: OPT.Structs;
  i: LONGINT;
BEGIN
  IF arr.typ.sysflag = notag THEN
    OPM.err(200);
  typ:=arr.typ;
  i:=n;
  WHILE i>0 DO BEGIN
    typ := typ.BaseTyp;
    DEC(i);
  END;
  ASSERT(typ.comp IN ArraySet{, 100});
  IF typ.comp = csStaticArr THEN
    MakeCon(len, OPT.linttyp, typ.n)
  ELSE IF typ.comp = csSDynArr THEN BEGIN
    ASSERT(typ.link<>NIL{, 101});
    MakeVar(len, typ.link);
  END ELSE BEGIN	(*dynamic size*)
    IF arr.descReg=none THEN BEGIN
      IF typ.comp=csDynArr THEN
        loadDynArr(arr)
      ELSE BEGIN
        ASSERT(arr.mode IN [omVar, omVarPar]{, 103});
        arr.descReg := FP;
        arr.descOffs := arr.adr + (arr.typ.n+1) * OPM.LIntSize;
      END
    END;
    ASSERT(arr.descReg <> none{, 102});	(*array descriptor must be set*)
    IF arr.descReg = FP THEN BEGIN
      len.mode := omVar;
      len.adr := arr.descOffs - OPM.LIntSize * n;
      len.offs := 0
    END ELSE BEGIN
      len.mode := omRegRel;
      len.adr := arr.descReg;
      len.offs := arr.descOffs - OPM.LIntSize * n
    END;
    len.typ := OPT.linttyp;
    len.node := NIL;
    len.mnolev := arr.mnolev;
    len.inx := none
  END
END;
	
PROCEDURE FNLen (VAR len, x, y: Item);	(* len := LEN (x, y), where x is a dynamic array and y the dimension *)
BEGIN
  ASSERT(y.mode = omCon{, 100});
  ASSERT(x.typ.comp IN [csOpenArr, csDynArr]{, 101});
  ArrayLen(len, x, y.adr)
END;
	
PROCEDURE ArrayBaseSize(VAR len: Item; arr: Item);
VAR
  typ: OPT.Structs;
  dim: LONGINT;
  t, tt: Item;
BEGIN
  ASSERT(arr.typ.comp IN ArraySet{, 100});
  typ := arr.typ.BaseTyp;
  dim := 1;
  MakeCon(len, OPT.linttyp, 1);
  WHILE typ.comp IN ArraySet DO BEGIN
    ArrayLen(t, arr, dim);  
    IF (t.mode=omCon) AND (len.mode=omCon) THEN
      len.adr:=len.adr*t.adr
    ELSE BEGIN
      tt := len;
      FNMul(len, t, tt, sfLInt)
    END;
    typ := typ.BaseTyp;
    INC(dim)
  END;
  MakeCon(t, OPT.linttyp, typ.size); 
  IF (t.mode=omCon) AND (len.mode=omCon) THEN
    len.adr:=len.adr*t.adr
  ELSE BEGIN
    tt := len;
    FNMul(len, t, tt, sfLInt)
  END;
END;

PROCEDURE TypeSize(x: Item; VAR size: Item; VAR step, scale: SHORTINT);	(*used for moves, RESULT:=s the scale*)
VAR
  typ: OPT.Structs;
  dim, rest: LONGINT;
  t, tt: Item;
BEGIN
  typ := x.typ;
  dim := 0;
  MakeCon(size, OPT.linttyp, 1);
  WHILE typ.comp IN ArraySet DO BEGIN
    ArrayLen(t, x, dim);
    IF (t.mode=omCon) AND (size.mode=omCon) THEN
      size.adr:=size.adr*t.adr
    ELSE BEGIN
      tt := size;
      FNMul(size, t, tt, sfLInt)
    END;
    typ := typ.BaseTyp;
    INC(dim)
  END;
  (*base type has a fixed size*)
  IF _SYSTEM._MOD(typ.size,4) = 0 THEN BEGIN
    scale := OPOi386.Scale4;
    step := OPOi386.Bit32;
    rest := _SYSTEM._DIV(typ.size,4)
  END ELSE IF _SYSTEM._MOD(typ.size,2) = 0 THEN BEGIN
    scale := OPOi386.Scale2;
    step := OPOi386.Bit16;
    rest := _SYSTEM._DIV(typ.size,2)
  END ELSE BEGIN
    scale := OPOi386.Scale1;
    step := OPOi386.Bit8;
    rest := typ.size
  END;
  IF rest<>0 THEN BEGIN
    IF size.mode=omCon THEN
      size.adr:=size.adr*rest
    ELSE BEGIN
      MakeCon(t, OPT.linttyp, rest);
      tt := size;
      FNMul(size, t, tt, sfLInt)
    END
  END;
  ASSERT(size.mode IN [omCon, omReg]{, 110})
END;

PROCEDURE FNLength (VAR len, x: Item); (* len := LEN (x) *)
BEGIN
  IF x.typ.form = sfString THEN
    MakeCon (len, OPT.linttyp, x.offs)
  ELSE
    ArrayLen(len, x, 0);
END;
	
PROCEDURE loadDynArr(VAR z: Item);	(*z points to the ArrBlk. make z point to the first elem and set desc*)
BEGIN
  ASSERT(z.typ.comp = csDynArr{, 100});
  ASSERT(z.descReg = none{, 101});
  load(z);
  z.descReg := z.adr;
  z.descOffs := ArrBlkLastDim + z.typ.n*4;	(*point to the first dim*)
  z.mode := omRegRel;
  z.offs := ArrBlkLastDim + 4 + 8*(_SYSTEM._DIV((1+z.typ.n),2));
END;
	
PROCEDURE FNIndex(VAR z, index: Item); (* z := z[index] *)
VAR
  t, len, size: Item;
  typ: OPT.Structs;
  scale, comp, check: LONGINT;
BEGIN
  check := z.mode;
  scale:=0;
  ASSERT(check IN [omVar, omVarPar, omAbs, omRegRel]{, 100});
  typ := z.typ;
  comp := typ.comp;
  ASSERT(comp IN ArraySet{, 101});
  IF (z.descReg=none) AND (comp<>csStaticArr) THEN BEGIN	(*first access: load base address, ev. array desc (dym/open/sdyn)*)
    ASSERT(z.mode <> omReg{, 102});
    IF comp = csDynArr THEN
      loadDynArr(z)
    ELSE IF (comp = csSDynArr) AND (z.mode <> omRegRel) THEN BEGIN
      ASSERT(z.mode IN [omVar, omVarPar]{, 103});
      z.mode := omVar;
      load(z);
      z.mode := omRegRel;
      z.offs := 0;
    END ELSE IF comp = csOpenArr THEN BEGIN
      ASSERT(z.mode IN [omVar, omVarPar]{, 103});
      z.descReg := FP;
      z.descOffs := z.adr + (z.typ.n+1) * OPM.LIntSize;
      z.mode := omVar;
      load(z);
      z.mode := omRegRel;
      z.offs := 0;
    END
  END;
		
  (*load index*)
  IF index.typ.form <> sfLInt THEN Convert(index, sfLInt);
		
  (*index check*)
  IF (OPM.inxchk IN OPM.options) AND ((index.mode<>omCon) OR (comp<>csStaticArr))THEN BEGIN
    ArrayLen(len, z, 0);
    CheckIndex(index, len)					(*side effect of CheckIndex: index is loaded*)
  END ELSE IF NOT (index.mode IN [omCon, omReg]) THEN
    load(index);
		
  (*perform index*)
  ASSERT(index.mode IN [omCon, omReg]{, 106});
  ASSERT(z.mode IN [omAbs, omVar, omVarPar, omRegRel]{, 107});
  IF (z.mode = omVar) AND (level - z.mnolev > 0 ) THEN (* we need static link *)
    CompleteSL (z);

  IF (index.mode <> omCon) OR (index.adr <> 0) THEN	BEGIN(*something to do*)
    ArrayBaseSize(size, z);
    ASSERT(size.mode IN [omCon, omReg]{, 105});
    IF (size.mode = omCon) AND (size.adr = 0) THEN
      (*skip*)
    ELSE IF (index.mode = omCon) AND (size.mode=omCon) THEN BEGIN
      index.adr := index.adr*size.adr;
      IF z.mode = omAbs THEN
        INC(z.adr, index.adr)
      ELSE IF z.mode = omVar THEN
        INC(z.adr, index.adr)
      ELSE IF z.mode <> omRegRel THEN BEGIN
        loadAdr(z);
        z.mode := omRegRel;
        z.offs := index.adr
      END ELSE
        INC(z.offs, index.adr);
    END ELSE BEGIN
      IF z.inx = none THEN BEGIN	(*index.mode = Reg*)
        IF z.mode = omVar THEN BEGIN
          z.mode := omRegRel;
          z.offs := z.adr;
          z.adr := FP
        END ELSE IF z.mode = omVarPar THEN BEGIN
          loadAdr(z);
          z.mode := omRegRel
        END;
        z.inx := index.adr;
        z.scale := scale;
      END ELSE BEGIN
        loadAdr(z);
        z.mode := omRegRel;
        z.offs := 0;
        z.inx := index.adr;
        z.scale := scale
      END;
      IF (size.mode=omCon) AND (size.adr<=8) AND (size.adr IN [1, 2, 4, 8]) THEN BEGIN
        CASE size.adr OF
          1: z.scale := OPOi386.Scale1;
          2: z.scale := OPOi386.Scale2;
          4: z.scale := OPOi386.Scale4;
          8: z.scale := OPOi386.Scale8;
        END;
      END ELSE BEGIN
        FNMul(t, size, index, sfLInt);
        index := t
      END;
      z.inx := index.adr
    END
  END;
  z.typ := typ.BaseTyp;
  DEC(z.descOffs, OPM.LIntSize);
END;

PROCEDURE MoveBlock (VAR z, x, size: Item; moveSize: SHORTINT);
VAR
  Vtype: OPT.Structs;
  src: Item;
BEGIN
  ASSERT (moveSize IN [OPOi386.Bit8, OPOi386.Bit16, OPOi386.Bit32]);
  IF (size.mode = omCon) AND (moveSize IN [OPOi386.Bit16, OPOi386.Bit8])  THEN BEGIN(* try to optimize *)
    IF (_SYSTEM._MOD(size.adr,4) = 0) AND (moveSize = OPOi386.Bit8) THEN BEGIN
      size.adr := _SYSTEM._DIV(size.adr,4);
      moveSize := OPOi386.Bit32
    END ELSE IF _SYSTEM._MOD(size.adr,2) = 0 THEN BEGIN
      size.adr := _SYSTEM._DIV(size.adr,2);
      DEC (moveSize, 8) (* 8 bit -> 16 bit; 16  bit -> 32 bit *)
    END
  END;
  IF (size.mode = omCon) AND (size.adr = 1) THEN BEGIN	(* block size = 1, 2, 4 -> use assignment, do  not loop *)
    ASSERT(moveSize IN [OPOi386.Bit8, OPOi386.Bit16, OPOi386.Bit32]);
    IF moveSize = OPOi386.Bit8 THEN
      Vtype := OPT.sinttyp
    ELSE IF moveSize = OPOi386.Bit16 THEN
      Vtype := OPT.inttyp
    ELSE
      Vtype := OPT.linttyp;
    IF x.mode = omReg THEN BEGIN
      x.mode := omRegRel;
      x.offs := 0;
      x.inx := none;
    END;
    x.typ := Vtype;
    load(x);
    IF (z.mode = omVarPar) OR ((z.mode = omVar) AND (level - z.mnolev <> 0)) THEN	(*Abs, RegRel handled by Store*)
      loadAdr(z);
    IF z.mode = omReg THEN BEGIN
      z.mode := omRegRel;
      z.offs := 0;
      z.inx := none
    END;
    z.typ := Vtype;
    GenStore(rcstore, z, x)
  END ELSE IF (size.mode = omCon) AND (size.adr = 3) AND (moveSize = OPOi386.Bit8) THEN	BEGIN(* special case, often used *)
    IF x.mode <> omReg THEN loadAdr(x);
    IF z.mode <> omReg THEN loadAdr(z);
    x.mode := omRegRel;
    x.offs := 0;
    x.inx := none;
    x.typ := OPT.inttyp;
    z.mode := omRegRel;
    z.offs := 0;
    z.inx := none;
    z.typ := OPT.inttyp;
    src := x;
    load(src);
    GenStore(rcstore, z, src);
    x.offs := 2;
    x.typ := OPT.sinttyp;
    z.offs := 2;
    z.typ := OPT.sinttyp;
    load(x);
    GenStore(rcstore, z, x)
  END ELSE BEGIN
    IF (size.typ.form <> sfLInt) AND (size.mode <> omCon) THEN BEGIN
      Convert (size, sfLInt);
      OPL.InstrTab [size.adr].hint := OPL.useECX;
    END ELSE BEGIN
      IF size.mode <> omReg THEN BEGIN
        size.typ := OPT.linttyp;
        load (size);
        OPL.InstrTab [size.adr].hint := OPL.useECX
      END ELSE IF (size.adr > 0) AND (OPL.InstrTab [size.adr].hint = OPL.noHint)  THEN
        OPL.InstrTab [size.adr].hint := OPL.tryECX;
    END;
    IF x.mode <> omReg THEN BEGIN
      loadAdr (x);
      OPL.InstrTab [x.adr].hint := OPL.useESI
    END ELSE IF (x.adr >= 0) AND (OPL.InstrTab [x.adr].hint = OPL.noHint) THEN
      OPL.InstrTab [x.adr].hint := OPL.tryESI;
    IF z.mode <> omReg THEN BEGIN
      loadAdr (z);
      OPL.InstrTab [z.adr].hint := OPL.useEDI
    END ELSE IF (z.adr >= 0) AND (OPL.InstrTab [z.adr].hint = OPL.noHint) THEN
      OPL.InstrTab [z.adr].hint := OPL.tryEDI;
    Gen0 (rccld);
    GenRepMovs (z, x, size, moveSize,  FALSE);
    IncReg (size.adr);
    IncReg (x.adr);
    IncReg (z.adr)
  END;
END;

PROCEDURE MoveBlockReversed (VAR z, x, size: Item; moveSize: SHORTINT);
(* same as MoveBlock, but reversed! *)
VAR
  scale: SHORTINT;
  t: LONGINT;
BEGIN
  scale:=0;
  ASSERT (moveSize IN [OPOi386.Bit8, OPOi386.Bit16, OPOi386.Bit32]);
  IF (size.mode = omCon) AND (moveSize IN [OPOi386.Bit16, OPOi386.Bit8])  THEN BEGIN (* try to optimize *)
    IF (_SYSTEM._MOD(size.adr,4) = 0) AND (moveSize = OPOi386.Bit8) THEN BEGIN
      size.adr := _SYSTEM._DIV(size.adr,4);
      moveSize := OPOi386.Bit32
    END ELSE IF _SYSTEM._MOD(size.adr,2) = 0 THEN BEGIN
      size.adr := _SYSTEM._DIV(size.adr,2);
      DEC (moveSize, 8) (* 8 bit -> 16 bit; 16  bit -> 32 bit *)
    END;
  END;
  IF (size.typ.form <> sfLInt) AND (size.mode <> omCon) THEN BEGIN
    Convert (size, sfLInt);
    OPL.InstrTab [size.adr].hint := OPL.useECX;
  END ELSE BEGIN
    IF size.mode <> omReg THEN BEGIN
      size.typ := OPT.linttyp;
      load (size);
      OPL.InstrTab [size.adr].hint := OPL.useECX
    END ELSE IF (size.adr > 0) AND (OPL.InstrTab [size.adr].hint = OPL.noHint)  THEN
      OPL.InstrTab [size.adr].hint := OPL.tryECX;
  END;
  ASSERT(size.mode=omReg);
  CASE moveSize OF
    OPOi386.Bit8: scale:=OPOi386.Scale1;
    OPOi386.Bit16: scale:=OPOi386.Scale2;
    OPOi386.Bit32: scale:=OPOi386.Scale4;
  END;
  t := _SYSTEM.ASH(1,scale);
		
  IF x.mode = omReg THEN BEGIN
    x.mode := omRegRel;
    x.offs := 0
  END ELSE IF (x.mode <> omRegRel) OR (x.inx <> none) OR (z.typ.form=sfComp) THEN BEGIN
    loadAdr(x);
    x.mode := omRegRel;
    x.offs := 0;
  END;
  ASSERT(x.mode=omRegRel);
  x.typ := OPT.linttyp;
  x.inx := size.adr;
  x.scale := scale;
  DEC(x.offs, t);
  loadAdr (x);
  OPL.InstrTab [x.adr].hint := OPL.useESI;
		
  IF z.mode = omReg THEN BEGIN
    z.mode := omRegRel;
    z.offs := 0
  END ELSE IF (z.mode <> omRegRel) OR (z.inx <> none) OR (z.typ.form=sfComp) THEN BEGIN
    loadAdr(z);
    z.mode := omRegRel;
    z.offs := 0;
  END;
  ASSERT(z.mode=omRegRel);
  z.typ := OPT.linttyp;
  z.inx := size.adr;
  z.scale := scale;
  DEC(z.offs, t);
  loadAdr (z);
  OPL.InstrTab [z.adr].hint := OPL.useEDI;
		
  Gen0 (rcstd);
  GenRepMovs (z, x, size, moveSize,  TRUE);
  IncReg (size.adr); IncReg (x.adr);
  IncReg (z.adr);
  Gen0 (rccld)
END ;

PROCEDURE FNMove(VAR z, x: Item); (* z := x, where the size of x is byte, word or dword *)
VAR
  src, dst: Item;
BEGIN
  IF z.mode = omVarPar THEN BEGIN
    loadAdr (z);
    z.mode := omRegRel
  END ELSE
    CompleteSL (z);
  IF x.typ = OPT.hinttyp THEN BEGIN
    x.typ := OPT.linttyp;
    z.typ := OPT.linttyp;
    IF x.mode = omVarPar THEN BEGIN
      loadAdr(x);
      x.mode := omRegRel ;
      x.inx := none
    END;
    NextWord(src, x);
    NextWord(dst, z);
    load(src);
    GenStore(rcstore, dst, src);
  END;
  load (x);
  GenStore (rcstore, z, x);
END;

PROCEDURE SYSmove (VAR z, x, nofBytes: Item); (* z := SYSTEM.MOVE (x, y) *)
BEGIN
  ASSERT ((z.typ.form IN [sfLInt, sfPointer]) AND (x.typ.form IN [sfLInt, sfPointer]));
  IF x.mode <> omReg THEN BEGIN
    load (x);
    OPL.InstrTab [x.adr].hint := OPL.useESI
  END ELSE IF OPL.InstrTab [x.adr].hint = OPL.noHint THEN
    OPL.InstrTab [x.adr].hint := OPL.tryESI;
  IF z.mode <> omReg THEN BEGIN
    load (z);
    OPL.InstrTab [z.adr].hint := OPL.useEDI
  END ELSE IF OPL.InstrTab [z.adr].hint = OPL.noHint THEN
    OPL.InstrTab [z.adr].hint := OPL.tryEDI;
  MoveBlock (z, x, nofBytes, OPOi386.Bit8)
END;
	
PROCEDURE FNCopy (VAR z, x: Item);	(* COPY (x, z) *)
VAR
  zLen, xLen, zLen0, xLen0, char0X, len, z0, x0, t: Item;
  Lfix, L: BLabel;
BEGIN
  MakeCon (char0X, OPT.chartyp, 0);
  FNLength (zLen, z);
  FNLength (xLen, x);
  IF (zLen.mode = omCon) AND (xLen.mode = omCon) THEN BEGIN
    IF zLen.adr >= xLen.adr THEN
      MoveBlock (z, x, xLen, OPOi386.Bit8)
    ELSE BEGIN
      DEC (zLen.adr); (* last char is always 0X *)
      MoveBlock (z, x, zLen, OPOi386.Bit8);
      t.adr := z.adr;
      t.mode := omRegRel;
      t.offs := 0;
      t.inx := none;
      t.typ := OPT.chartyp;
      t.node := NIL;
      FNMove (t, char0X) (* write 0X at the end of z *)
    END;
  END ELSE BEGIN
    (* pre-load the addresses (it will be done only once and will free some regs!!) *)
    IF x.mode<>omReg THEN
      loadAdr (x)
    ELSE BEGIN
      x0:=x;
      MakeCon(len, OPT.linttyp, 0);
      FNAdd(x, x0, len, sfLInt)
    END;
    OPL.InstrTab [x.adr].hint := OPL.tryESI;
    loadAdr (z);
    OPL.InstrTab [z.adr].hint := OPL.tryEDI;
    IF zLen.mode = omCon THEN BEGIN
      xLen0 := xLen;
      IF xLen.typ.form <> sfLInt THEN
        Convert (xLen0, sfLInt)
      ELSE
        load (xLen0);
      OPL.InstrTab [xLen0.adr].hint := OPL.tryECX;
      GenFlags (rccmp, xLen0, zLen);
      Lfix := -2-pc;
      Gen1 (rcjg, aNil, NIL)
    END ELSE BEGIN
      zLen0 := zLen;
      IF zLen0.typ.form <> sfLInt THEN
        Convert (zLen0, sfLInt)
      ELSE
        load (zLen0);
      OPL.InstrTab [zLen0.adr].hint := OPL.tryECX;
      xLen0 := xLen;
      IF xLen0.mode <> omCon THEN BEGIN
        IF xLen0.typ.form <> sfLInt THEN
          Convert (xLen0, sfLInt)
        ELSE
          load (xLen0);
      END;
      GenFlags (rccmp, zLen0, xLen0);
      Lfix := -2-pc;
      Gen1 (rcjl, aNil, NIL)
    END;
    (* first case: |src| <= |trg|: copy all *)
    z0 := z;
    x0 := x;
    MoveBlock (z0, x0, xLen, OPOi386.Bit8);
    L := -2-pc;
    Gen1 (rcjmp, aNil, NIL);
    FixLink (Lfix);
    (* second case: |src| > |trg|: copy |trg|-1 bytes *)
    IF zLen.mode = omCon THEN
      DEC (zLen.adr)
    ELSE BEGIN
      MakeCon (len, OPT.linttyp, 1);
      load (zLen);
      OPL.InstrTab [zLen.adr].hint := OPL.useECX;
      t := zLen;
      FNSub (zLen, t, len, sfLInt)
    END;
    MoveBlock (z, x, zLen, OPOi386.Bit8);
    t.mode := omRegRel;
    t.adr := z.adr;
    t.offs := 0;
    t.inx := none;
    t.typ := OPT.chartyp;
    t.node := NIL;
    FNMove (t, char0X); (* last char is 0X *)
    FixLink (L)
  END 
END;

PROCEDURE CopyDynArr(VAR x, y: Item); FORWARD;

PROCEDURE FNAssign(VAR z, x: Item);	(* z  := x *)
VAR
  zform, xform, step, scale : SHORTINT;
  temp, zero, sizeItem: Item;
  proc: OPT.Objects;
BEGIN
  zform := z.typ.form;
  xform := x.typ.form;
  IF (zform IN [sfInt, sfLInt, sfReal, sfLReal]) AND (zform <> xform) THEN Convert (x, zform);
  CASE zform OF
    sfBool:
            BEGIN
              IF x.mode = omCoc THEN BEGIN
                temp := x;
                MoveCoc (x, temp)
              END ELSE
                load (x);
              IF z.mode = omVarPar THEN BEGIN
                loadAdr (z);
                z.mode := omRegRel;
              END ELSE
                CompleteSL (z);
              GenStore (rcstore, z, x);
            END;
    sfByte, sfChar, sfSInt, sfSet, sfInt, sfLInt, sfHInt: FNMove (z, x);
    sfReal:
            BEGIN
              loadf (x);
              IF z.mode = omVarPar THEN BEGIN
                loadAdr (z);
                z.mode := omRegRel
              END ELSE
                CompleteSL (z);
              GenStore (fstore, z, x);
            END;
    sfLReal:
            BEGIN
              loadf (x);
              IF z.mode = omVarPar THEN BEGIN
                loadAdr (z);
                z.mode := omRegRel
              END ELSE
                CompleteSL (z);
              GenStore (fstore, z, x)
            END;
    sfPointer:
            BEGIN
              IF xform = sfNilTyp THEN BEGIN
                MakeCon (zero, OPT.linttyp, Onil);
                load (zero);
                IF z.mode = omVarPar THEN BEGIN
                  loadAdr (z);
                  z.mode := omRegRel
                END ELSE
                  CompleteSL (z);
                GenStore (rcstore, z, zero);
              END ELSE BEGIN
                load (x);
                IF z.mode = omVarPar THEN BEGIN
                  loadAdr (z);
                  z.mode := omRegRel
                END ELSE
                  CompleteSL (z);
                temp:=z;
                GenStore (rcstore, z, x);
              END;
            END;
    sfProcTyp:
            IF (z.typ = x.typ) OR (xform = sfNilTyp) THEN BEGIN (* procedure variable assignment *)
              IF xform = sfNilTyp THEN
                MakeCon (x, OPT.linttyp, Onil);
              FNMove (z, x)
            END ELSE BEGIN(* procedure assignment *)
              proc := x.node.obj;
              IF x.mnolev = 0 THEN (* local procedure assignment *)
                GenLoadProc (rcldProc, x, _SYSTEM._MOD(proc.adr,$10000), x.node)
              ELSE
                GenLoadProc (rcldXProc, x, _SYSTEM._DIV(proc.adr,$10000), x.node);
              FNMove (z, x);
            END;
    sfComp:
            BEGIN
              ASSERT ((z.typ.comp=csRecord) OR ((z.typ.comp IN [csStaticArr, csDynArr, csSDynArr])));
              IF x.typ.form = sfString THEN
                FNCopy(z, x)
              ELSE BEGIN
                IF (z.typ.comp IN [(*StaticArr,*) csSDynArr]) THEN BEGIN
                  TypeSize(z, sizeItem, step, scale);
                  temp:=z;
                  MoveBlock (z, x, sizeItem, step);
                END ELSE IF z.typ.comp = csDynArr THEN
                  CopyDynArr(z, x)
                ELSE IF x.typ.size<>4 THEN BEGIN
                  MakeCon (sizeItem, OPT.linttyp, z.typ.size);
                  step:=OPOi386.Bit8;	(* optimize later in MoveBlock *)
                  temp:=z;
                  MoveBlock (z, x, sizeItem, step)
                END ELSE BEGIN
                  temp:=z;
                  FNMove(z, x)
                END;
              END;
            END;
  ELSE
    HALT (BUG)
  END
END;

PROCEDURE IncDec (VAR z, x: Item; increment: BOOLEAN); (* INC / DEC (z, x) *)
VAR
  z0, temp: Item;
BEGIN
  z0 := z;
  load (z0);
  temp := z0;
  IF increment THEN
    FNAdd (z0, temp, x, z.typ.form)
  ELSE
    FNSub (z0, temp, x, z.typ.form);
  FNAssign (z, z0)
END;

PROCEDURE MulDim(VAR nofelem, len: Item); (* nofelem := nofelem * len *)
VAR
  t: Item;
BEGIN
  IF (nofelem.mode = omCon) AND (len.mode = omCon) THEN
    nofelem.adr := nofelem.adr * len.adr
  ELSE BEGIN
    t := nofelem;
    FNMul (nofelem, t, len, sfLInt)
  END
END;

PROCEDURE PushLen(VAR z: Item); (* push length z -> NEW (dynArr, dim0, dim1,... ) *)
BEGIN
  IF z.mode <> omCon THEN BEGIN
    IF  z.typ.form <> sfLInt THEN BEGIN
      IF z.mode=omVar THEN load(z);
      Convert (z, sfLInt)
    END ELSE
      load (z);
  END;
  GenPush (z)
END;

PROCEDURE PopLen(VAR block: Item);
VAR
  reg: Item;
BEGIN
  reg.typ := OPT.linttyp; (* reg.node := NIL is done in GenPop *)
  GenPop (reg);
  GenStore (rcstore, block, reg)
END;

PROCEDURE NewSys (VAR z, x: Item); (* SYSTEM.NEW (z, x) *)
BEGIN
  IF z.mode <> omReg THEN loadAdr (z);
  GenPush (z);
  IF x.typ.form IN [sfSInt, sfInt] THEN BEGIN
    IF x.mode = omCon THEN
      x.typ := OPT.linttyp
    ELSE
      Convert(x, sfLInt)  (* << mh 30.8.94 *)
  END;
  IF x.mode<>omCon THEN load (x);
  GenPush (x);
  GenCall (rcxcall, OPLi386.NewSysLinkIndex * $10000 + OPLi386.NewSysEntryNr, 8, NIL);
END;
	
PROCEDURE NewArray(VAR z, nofelem: Item; nofdim: LONGINT; typ: OPT.Structs; dimUsed: BOOLEAN);
VAR
  cons, tdesc, t: Item;
  nofArrElems: LONGINT;
BEGIN
  nofArrElems:=1;
  WHILE typ.comp=csStaticArr DO BEGIN
    nofArrElems:=nofArrElems*typ.n;
    typ:=typ.BaseTyp
  END;
  IF nofArrElems<>1 THEN BEGIN
    MakeCon(cons, OPT.linttyp, nofArrElems);
    MulDim(nofelem, cons)
  END;
  loadAdr(z); 
  IF dimUsed THEN GenPush(z);
  IF (typ.form <> sfPointer) AND (typ.comp <> csRecord) THEN BEGIN(* simple type -> use SYSTEM.NEW (...) *)
    IF nofelem.mode=omCon THEN
      nofelem.adr := nofelem.adr*typ.size + ArrBlkFirstElem + 8 + 8*_SYSTEM._DIV(nofdim,2)	(*rounded!*)
    ELSE BEGIN
      MakeCon(cons, OPT.linttyp, typ.size);
      t:=nofelem;
      FNMul(nofelem, t, cons, sfLInt);
      cons.adr := ArrBlkFirstElem + 8 + 8*_SYSTEM._DIV(nofdim,2);
      t:=nofelem;
      FNAdd(nofelem, t, cons, sfLInt)
    END;
    t := z;
    NewSys(t, nofelem)
  END ELSE BEGIN
    t := z;
    GenPush(t);
    IF typ.form=sfPointer THEN
      MakeCon(tdesc, OPT.linttyp, OPM.nilval)
    ELSE BEGIN
      GetTdAdr(typ, tdesc);
      load(tdesc)
    END;
    GenPush(tdesc);
    IF nofelem.mode<>omCon THEN load(nofelem);
    GenPush(nofelem);
    MakeCon(cons, OPT.linttyp, nofdim);
    GenPush(cons);
    GenCall(rcxcall, OPLi386.NewArrayLinkIndex * $10000 + OPLi386.NewArrayEntryNr, 16, NIL)
  END;
  IF dimUsed THEN BEGIN
    GenPop(z);
    z.mode:=omRegRel;
    z.offs := 0;
    load(z);
    z.mode:=omRegRel;
    z.typ:=OPT.ptrtyp;
    z.offs:=ArrBlkFirstElem;
    z.inx:=none
  END
END;

PROCEDURE NewRec (VAR z: Item; typ: OPT.Structs); (* NEW (record), NEW (POINTER TO Static ARRAY) *)
VAR
  tag, size: Item;
  btyp: OPT.Structs;
  len: LONGINT;
BEGIN
  IF typ.comp=csStaticArr THEN BEGIN
    len := typ.n;
    btyp := typ.BaseTyp;
    WHILE btyp.comp = csStaticArr DO BEGIN
      len := len * btyp.n;
      btyp := btyp.BaseTyp
    END;
    MakeCon (size, OPT.linttyp, len);
    IF (btyp.comp = csRecord) OR (btyp.form = sfPointer) THEN
      NewArray (z, size, 1, btyp, FALSE)
    ELSE BEGIN
      size.adr := typ.size+16;
      NewSys (z, size)
    END
  END ELSE BEGIN
    loadAdr (z);
    GenPush (z);
    IF typ.comp=csDynArr THEN
      MakeCon(tag, OPT.linttyp, OPM.nilval)
    ELSE
      GetTdAdr (typ, tag);
    load (tag);
    GenPush (tag);
    GenCall (rcxcall, OPLi386.NewLinkIndex * $10000 + OPLi386.NewEntryNr, 8, NIL)
  END;
END;

PROCEDURE CallRecBody(VAR self, body: Item; typ: OPT.Structs);		(* Active only *)
VAR
  Vcon: Item;
  Vtype: LONGINT;
  tag, t0, t1: Item;
		
  PROCEDURE GetBody(curtype, bodytype: OPT.Structs;  VAR body: Item);
  VAR
    tag: Item;
  BEGIN
    GetTdAdr(curtype, tag);	(*use indirect access, the record may be invisible!*)
    DeRef(tag);
    IF curtype <> bodytype THEN BEGIN
      tag.mode := omRegRel;
      tag.offs := -8 - 4*bodytype.extlev;
      tag.inx := none;
      tag.typ := OPT.ptrtyp;
      load(tag)
    END;
    body.adr := tag.adr; (* x.offs set by OPV.Designator *)	(*body.offs := MethodOffs - 4 * (body.obj.adr DIV 10000H);*)
    body.mode := omRegRel;
    body.typ := OPT.ptrtyp;
  END;
		
BEGIN
  IF (typ.strobj <> NIL) AND (activeObj IN typ.strobj.conval^.setval) THEN BEGIN
    ASSERT(self.typ.form = sfPointer);
    GetBody(self.typ.BaseTyp, typ, body);
    load (body);
    GenPush(body);
    MakeCon(Vcon, OPT.sinttyp, typ.strobj.prio);
    GenPush(Vcon);
    Vtype := 0;
    IF safe IN typ.strobj.conval^.setval THEN INC (Vtype, 1)	(* set bit 0 *);
    MakeCon (Vcon, OPT.sinttyp, Vtype);
    GenPush(Vcon);
    load(self);
    GenPush(self);
    GenCall(rcxcall, OPLi386.StartIndex * $10000 + OPLi386.StartEntryNr, 16, NIL)
  END ELSE BEGIN
    (* typ and self.typ can be different!!!*)
    tag := self;
    loadAdr(tag);
    tag.mode := omRegRel;
    tag.offs := -4;
    tag.inx := none;
    tag.typ := OPT.linttyp;
    load(tag);
    IF OPM.TDMask <> 0 THEN BEGIN
      t1 := tag;
      MakeCon(t0, OPT.linttyp, OPM.TDMask);
      Gen3(rcand, tag, t1, t0);
    END;
    GenPush(tag);
    load(self);
    GenPush(self);
    GetBody(self.typ.BaseTyp, typ, body);
    load (body);
    GenCall(rccallReg, body.adr, 4, NIL)
  END;
END;

PROCEDURE PushArrLen(fp: OPT.Objects; VAR ap, dest: Item; useDest: BOOLEAN);
VAR
  atyp, ftyp: OPT.Structs;
  depth: LONGINT;
  len, size, t: Item;
BEGIN
  atyp := ap.typ;
  ftyp := fp.typ;
  depth := 0;
  WHILE ftyp.comp IN [csOpenArr, csSDynArr] DO BEGIN
    IF ftyp.BaseTyp = OPT.bytetyp THEN BEGIN
      IF (atyp.comp = csRecord) AND (ap.mode = omVarPar) THEN BEGIN	(*size is dynamic*)
        len := ap;
        CompleteSL (len);
        IF len.mode IN [omVar, omVarPar] THEN BEGIN
          INC (len.adr, 4);
          len.mode := omVar  (* type descriptor is on the stack (hidden parameter) *)
        END ELSE BEGIN
          ASSERT (len.mode = omRegRel);
          INC (len.offs, 4)
        END;
        DeRef (len);
        load (len);
      END ELSE IF atyp.comp IN [csOpenArr, csSDynArr, csDynArr] THEN BEGIN	(*size is dynamic*)
        ArrayLen(len, ap, depth);
        ArrayBaseSize(size, ap);
        IF (len.mode=omCon)AND(size.mode=omCon) THEN
          len.adr:=len.adr*size.adr
        ELSE BEGIN
          FNMul(t, len, size, sfLInt);
          len:=t
        END;
      END ELSE
        MakeCon(len, OPT.linttyp, atyp.size);
    END ELSE IF atyp.form = sfString THEN BEGIN
      ASSERT(ap.mode = omAbs);
      MakeCon(len, OPT.linttyp, ap.offs)
    END ELSE BEGIN
      ASSERT(atyp.comp IN ArraySet{, 100});
      ArrayLen(len, ap, depth);
    END;
    len.typ := OPT.linttyp;
    IF useDest THEN BEGIN
      IF len.mode <> omReg THEN load(len);
      DEC(dest.offs, 4);
      t := dest;
      GenStore(rcstore, t, len)
    END ELSE BEGIN
      IF  NOT (len.mode IN [omCon, omReg]) THEN load(len);
      GenPush(len)
    END;
    atyp := atyp.BaseTyp;
    ftyp := ftyp.BaseTyp;
    INC(depth);
  END;
  IF (ap.mode=omRegRel) AND (ap.adr=ESP) THEN INC(ap.offs, 4*depth);
END;
	
PROCEDURE PushComp(fp: OPT.Objects; VAR ap: Item);
VAR
  sizeItem, t: Item;
  size: LONGINT;
BEGIN
  IF ap.typ.form = sfString THEN BEGIN
    size := fp.typ.size;
    INC(size, _SYSTEM._MOD(-size,4));
    MakeCon (sizeItem, OPT.linttyp, size);
    t:=Stack;
    FNSub(t, Stack, sizeItem, sfLInt);
    MakeCon (sizeItem, OPT.linttyp, ap.offs);
    t:=Stack;
    MoveBlock (t, ap, sizeItem, OPOi386.Bit8)
  END ELSE IF ap.typ.size<>4 THEN BEGIN
    size := fp.typ.size;
    INC(size, _SYSTEM._MOD(-size,4));
    MakeCon (sizeItem, OPT.linttyp, size);
    t:=Stack;
    FNSub(t, Stack, sizeItem, sfLInt);
    MakeCon (sizeItem, OPT.linttyp, fp.typ.size);
    t:=Stack;
    MoveBlock (t, ap, sizeItem, OPOi386.Bit8)
  END ELSE BEGIN
    load(ap);
    GenPush(ap)
  END;
END;
	
PROCEDURE ParRecord (fp: OPT.Objects; VAR ap: Item; apIsDeref, alwaysRef: BOOLEAN);
VAR
  t, t0, t1, tag: Item;
  pushtag: BOOLEAN;
BEGIN
  pushtag := fp.typ.sysflag <> notag;
  IF ap.mode = omVarPar THEN BEGIN
    IF (fp.mode = omVarPar) OR alwaysRef THEN BEGIN(* push tag too -> tag is on the stack as a hidden parameter *)
      IF pushtag THEN BEGIN
        t.mode := omVar;
        t.adr := ap.adr + 4;
        t.typ := OPT.ptrtyp;
        t.node := NIL;
        t.inx := none;
        t.offs := 0;
        t.mnolev := ap.mnolev;
        load (t);
        GenPush (t);
        lastTD := t;	(*remember the tag, used by Call later*)
        IF (ap.mode=omRegRel) AND (ap.adr=ESP) THEN  INC(ap.offs, 4);
      END;
      IF ap.mode<>omReg THEN loadAdr(ap);
      GenPush(ap)
    END ELSE
      PushComp(fp, ap);
  END ELSE BEGIN
    IF (fp.mode = omVarPar) OR alwaysRef THEN BEGIN(* push {Var, Abs} record *)
      IF pushtag THEN BEGIN
        IF apIsDeref THEN BEGIN(* get type descriptor of dynamic type *)
          ASSERT((ap.mode = omRegRel) AND (ap.offs = 0));
          tag.mode := omRegRel;
          tag.scale := noScale;
          tag.typ := OPT.ptrtyp;
          tag.offs := -4;
          tag.adr := ap.adr;
          tag.inx := none;
          tag.descReg := none;
          IF OPM.TDMask <> 0 THEN BEGIN
            t1 := tag; load(t1);
            MakeCon(t0, OPT.linttyp, OPM.TDMask);
            Gen3(rcand, tag, t1, t0)
          END;
        END ELSE
          GetTdAdr(ap.typ, tag); (* get typedescriptor address of static type *)
        load(tag);
        GenPush(tag);
        IF (ap.mode=omRegRel)AND(ap.adr=ESP) THEN  INC(ap.offs, 4);
        lastTD := tag;	(*remember the tag, used by Call later*)
      END;
      IF ap.mode<>omReg THEN loadAdr(ap);
      GenPush(ap)
    END ELSE
      PushComp(fp, ap);
  END;
END;

PROCEDURE Parameter (VAR ap: Item; fp: OPT.Objects; apIsDeref, alwaysRef: BOOLEAN);
VAR
  form, apMode: SHORTINT;
  temp, cons, vstack, tag: Item;
  proc: OPT.Objects;
  apTyp, fpTyp: OPT.Structs;
  pushtag: BOOLEAN;
BEGIN
  fpTyp := fp.typ;
  apTyp := ap.typ;
  pushtag := (fpTyp.sysflag <> notag) AND (apTyp.sysflag <> notag);
  IF (ap.typ.form = sfNilTyp) AND ( NOT pushtag OR (fp.mode = omVarPar)) THEN BEGIN
    load(ap);
    GenPush(ap);
    EXIT;
  END;
  IF (apTyp.comp=csDynArr) AND (fpTyp.comp <> csDynArr) AND (ap.descReg=none) THEN loadDynArr(ap);
  IF (fp.mode = omVarPar) OR (fpTyp.comp IN [csRecord, csStaticArr, csOpenArr, csSDynArr]) THEN BEGIN
    form := fpTyp.comp;
    IF form = csRecord THEN
      ParRecord (fp, ap, apIsDeref, alwaysRef)
    ELSE IF (fp.mode=omVar) AND (fpTyp.comp=csStaticArr) AND pushtag THEN	(* static arrays *)
      PushComp(fp, ap)
    ELSE IF (fpTyp.comp=csOpenArr) THEN BEGIN	(* non static arrays *)
      IF pushtag THEN PushArrLen (fp, ap, ap, FALSE);
      IF ap.mode<>omReg THEN loadAdr (ap);
      IF ap.typ = OPT.hinttyp THEN ap.typ := OPT.linttyp;
      GenPush (ap)
    END ELSE BEGIN
      apMode := ap.mode;
      IF apTyp.form = sfString THEN BEGIN
        HALT(100); (* constant *)
        ap.node := NIL;	(*not possible: can only be passed to a value par*)
      END;
      IF apTyp = OPT.hinttyp THEN ap.typ := OPT.linttyp;
      IF ap.mode<>omReg THEN loadAdr (ap);
      GenPush (ap);
      IF (fpTyp = OPT.ptrtyp) AND (apTyp <> OPT.ptrtyp) THEN BEGIN
        (* pass ap static typ to enable run time tests -> hack *)
        IF apTyp.BaseTyp.comp = csRecord THEN
          GetTdAdr (apTyp.BaseTyp, tag)
        ELSE
          MakeCon (tag, OPT.linttyp, Onil);
        ap.mode := omRegRel;
        FNAssign (ap, tag);
      END ELSE IF (fpTyp = OPT.ptrtyp) AND (apTyp = OPT.ptrtyp) AND (apMode <> omVarPar) THEN BEGIN
        (* pass nil to disable runtime tests -> hack *)
        MakeCon (cons, OPT.linttyp, Onil);
        ap.mode := omRegRel;
        FNAssign (ap, cons)
      END
    END
  END ELSE BEGIN
    (*push as value parameter*)
    ASSERT (ap.mode IN [omAbs, omVar, omVarPar, omCon, omReg, omRegRel, omCoc, omXProc, omLProc]);
    form := apTyp.form;
    IF (apTyp <> fpTyp) AND  NOT (ap.mode IN [omXProc, omLProc])
    AND  NOT (form  IN [sfPointer, sfNilTyp, sfNoTyp]) THEN BEGIN
      (* convert actual parameter *)
      Convert (ap, fp.typ.form);
      form := ap.typ.form
    END;
    IF (form IN realSet) AND (ap.mode = omReg) THEN BEGIN(* ap was converted *)
      IF fp.typ.form = sfReal THEN
        MakeCon (cons, OPT.linttyp, 4)
      ELSE
        MakeCon (cons, OPT.linttyp, 8);
      vstack := Stack;
      Gen3 (rcsub, vstack, vstack, cons);
      vstack.mode := omRegRel;
      vstack.offs := 0;
      vstack.inx := none;
      vstack.typ := fp.typ; (* OPT.lrltyp; *)
      GenStore (fstore, vstack, ap)
    END ELSE BEGIN
      IF ap.mode IN [omXProc, omLProc] THEN BEGIN
        proc := ap.node.obj;
        IF ap.mnolev = 0 THEN (* push global procedure *)
          GenLoadProc (rcldProc, ap, _SYSTEM._MOD(proc.adr,$10000), ap.node)
        ELSE (* push external procedure *)
          GenLoadProc (rcldXProc, ap, _SYSTEM._DIV(proc.adr,$10000), ap.node);
      END ELSE IF form IN [sfHInt, sfLReal] THEN BEGIN(* push higher 4 bytes -> lower bytes will be pushed later *)
        temp := ap;
        temp.typ := OPT.linttyp;
        IF temp.mode = omVarPar THEN BEGIN
          loadAdr(temp);
          temp.mode := omRegRel;
          temp.offs := 0
        END; (* << mb 15.2.94 *)
        IF temp.mode = omRegRel THEN
          INC (temp.offs, 4)
        ELSE IF temp.mode = omReg THEN
          temp.adr := temp.offs	(*msb*)
        ELSE
          INC (temp.adr, 4);
        load (temp);
        GenPush (temp)
      END ELSE IF ap.mode = omCoc THEN BEGIN
        temp := ap;
        MoveCoc (ap, temp)
      END;
      IF ap.mode <> omCon THEN BEGIN
        IF ap.typ.form IN realSet+[sfHInt] THEN ap.typ := OPT.linttyp;
        load (ap)
      END;
      GenPush (ap);
      lastTD := ap
    END
  END
END;
	
FUNCTION PrepRetDest(z: Item): LONGINT;
VAR
  t, from: Item;
  size: LONGINT;
BEGIN
  ASSERT(z.typ.form = sfComp);
  size := 4;
  IF z.mode = 0 THEN BEGIN	(*special for RESULT:= Call *)
    IF z.typ.comp = csRecord THEN BEGIN
      INC(size);
      from.mode:=omRegRel;
      from.adr:=FP;
      from.offs:=z.offs+4;
      from.typ:=OPT.linttyp;
      from.inx:=none;
      load(from);
      GenPush(from)
    END;
    from.mode:=omRegRel;
    from.adr:=FP;
    from.offs:=z.offs;
    from.typ:=OPT.linttyp;
    from.inx:=none;
    load(from);
    GenPush(from)
  END ELSE BEGIN
    IF z.typ.comp = csRecord THEN BEGIN
      MakeCon(t, OPT.linttyp, z.typ.size);
      GenPush(t);
      INC(size, 4)
    END;
    IF z.mode IN [omVar, omVarPar, omAbs, omRegRel] THEN
      loadAdr(z);
    GenPush(z)
  END;
  RESULT:= size;
END;
	
PROCEDURE PushRetDesc(VAR x, tos: Item);	(*push RESULT:= descriptor!*)
VAR con: Item;
BEGIN
  ASSERT((tos.mode=omRegRel) AND (tos.adr=ESP){, 101});
  IF x.typ.comp = csRecord THEN BEGIN
    IF (x.mode=omRegRel) AND (x.adr=ESP) THEN INC(x.offs, 4);
    MakeCon(con, OPT.linttyp, x.typ.size);
    GenPush(con);
    INC(tos.offs, 4)
  END;
  loadAdr(x);
  GenPush(x);
  INC(tos.offs, 4)
END;
	
PROCEDURE CorrectStackAdr(VAR x: Item; size, dynblocks: LONGINT);
VAR
  oldoff: LONGINT;
  typ: OPT.Structs;
BEGIN
  typ := x.typ;
  x.typ := OPT.linttyp;	(*to ensure that load works fine*)
  IF (dynblocks<>0) AND (x.mode=omRegRel) AND (x.adr=ESP) THEN BEGIN
    oldoff := x.offs;
    x.offs := size;
    WHILE dynblocks<>0 DO BEGIN
      load(x);
      x.mode := omRegRel;
      x.offs := 0;
      DEC(dynblocks)
    END;
    x.offs := oldoff
    (*oldoff := x.offs; x := oldsp; x.offs := oldoff*)
  END ELSE IF (size<>0) AND (x.mode=omRegRel) AND (x.adr=ESP) THEN	(*if the item is stack relative, correct the position after the allocation of temporary values*)
    INC(x.offs, size);
  x.typ := typ
END;

PROCEDURE ResetStack(dynblocks: LONGINT);
VAR
  t: Item;
BEGIN
  t := Stack;
  WHILE dynblocks<>0 DO BEGIN
    GenPop(t);
    OPL.InstrTab [pc-1].hint := OPL.useESP;
    DEC(dynblocks)
  END
END;
	
PROCEDURE AllocateStatic(VAR tos: Item; size: LONGINT);
VAR
  t, con: Item;
BEGIN
  ASSERT(size >= 0{, 100});
  ASSERT(size MOD 4 = 0{, 101});
  IF size <> 0 THEN BEGIN
    t := Stack;
    MakeCon(con, OPT.linttyp, size);
    FNSub(t, Stack, con, sfLInt);
  END;
  tos.mode := omRegRel;
  tos.adr := ESP;
  tos.offs := 0;
  tos.inx := none;
  tos.typ := OPT.linttyp;
END;

PROCEDURE RemoveStatic(size: LONGINT);
VAR	t, con: Item;
BEGIN
  ASSERT(size >= 0{, 100});
  ASSERT(size MOD 4 = 0{, 101});
  IF size <> 0 THEN BEGIN
    t := Stack;
    MakeCon(con, OPT.linttyp, size);
    FNAdd(t, Stack, con, sfLInt)
  END
END;
	
PROCEDURE CopyDynArr(VAR x, y: Item);
VAR
  dim: LONGINT;
  yy, xx, size, t: Item;
  typ: OPT.Structs;
  step, scale: SHORTINT;
BEGIN
  IF  NOT (y.mode IN [omVar, omVarPar, omAbs]) THEN BEGIN
    load(y);
    GenPush(y)
  END;
  yy:=y;
  typ:=y.typ;
  dim:=0;
  IF y.descReg = none THEN loadDynArr(y);
  MakeCon(size, OPT.linttyp, 1);
  WHILE typ.comp=csDynArr DO BEGIN
    ArrayLen(t, y, dim);
    FNMul(size, size, t, sfLInt);
    typ:=typ.BaseTyp;
    INC(dim)
  END;
    (*allocate x*)
  NewArray(x, size, dim, typ, TRUE);
  (*now x points to the first elem field*)
  INC(x.offs, 4);
  xx:=x;
  INC(x.offs, _SYSTEM._DIV(dim,2)*8+4);
    (*copy*)
  y:=yy;
  IF  NOT (y.mode IN [omVar, omVarPar, omAbs]) THEN GenPop(y);
  IF y.descReg = none THEN loadDynArr(y);
  TypeSize(y, size, step, scale);
  MoveBlock(x, y, size, step);	(*copy data*)
  y.mode:=omRegRel;
  y.adr:=y.descReg;
  y.offs:=ArrBlkLastDim;
  y.typ:=OPT.linttyp;
  x:=xx;
  x.typ:=OPT.linttyp;
  WHILE dim>0 DO BEGIN
    xx:=x;
    yy:=y;
    FNMove(xx, yy);
    INC(x.offs, 4);
    INC(y.offs, 4);
    DEC(dim)
  END
END;
	
FUNCTION ArgSize(par: OPT.Objects): LONGINT; (* ejz *)
VAR
  typ: OPT.Structs;
  size: LONGINT;
BEGIN
  size := 0;
  WHILE par <> NIL DO BEGIN
    typ := par.typ;
    IF (par.mode = omVarPar) OR (typ.form IN [sfString, sfComp]) THEN BEGIN
      INC(size, 4); (* address *)
      IF (typ.comp IN [csDynArr,csRecord]) AND (typ.sysflag <> notag) THEN
        INC(size, 4); (* descriptor *)
    END ELSE IF _SYSTEM._MOD(typ.size,4) <> 0 THEN
      INC(size, typ.size + 4 - _SYSTEM._MOD(typ.size,4))
    ELSE
      INC(size, typ.size);
    par := par.link
  END;
  RESULT:= size
END;

PROCEDURE EndCall(proc: OPT.Objects; typ: OPT.Structs); (* ejz *)
  VAR Z, X, Y: Item;
BEGIN
  IF ((proc <> NIL) AND (proc.sysflag = cdecl)) OR ((typ <> NIL) AND (typ.sysflag = cdecl)) THEN BEGIN
    Z.mode := omReg;
    Z.typ := OPT.ptrtyp;
    Z.node := NIL;
    Z.adr := ESP;
    Z.offs := 0;
    Z.inx := none;
    X.mode := omReg;
    X.typ := OPT.ptrtyp;
    X.node := NIL;
    X.adr := ESP;
    X.offs := 0;
    X.inx := none;
    Y.mode := omCon;
    Y.typ := OPT.ptrtyp;
    Y.node := NIL;
    IF proc <> NIL THEN
      Y.adr := ArgSize(proc.link)
    ELSE 
      Y.adr := ArgSize(typ.link);
    Y.offs := 0;
    Y.inx := none;
    Gen3(rcadd, Z, X, Y)
  END
END;
	
PROCEDURE FNCall(VAR x: Item; proc: OPT.Objects; node: OPT.Nodes);
VAR
  n: INTEGER;
  sl: Item;
  param: OPT.Objects;
  tag, t0, t1: Item;
BEGIN
  IF (x.mode IN [omLProc, omCProc]) OR ((x.mode IN [omXProc]) AND (x.mnolev >= 0)) THEN BEGIN(* local procedure *)	(* was active only *)
    IF slNeeded IN proc.conval^.setval THEN BEGIN
      n := level - x.mnolev;
      sl.mode := omVar;
      sl.typ := OPT.linttyp;
      sl.mnolev := level;
      sl.inx := none;
      sl.node := NIL;
      IF n = 0 THEN BEGIN
        sl.mode := omReg;
        sl.adr := FP;
        GenPush(sl)
      END ELSE IF n = 1 THEN BEGIN
        sl.adr := StaticLinkAdr;
        load(sl);
        GenPush(sl)
      END ELSE BEGIN
        sl.adr := StaticLinkAdr;
        load (sl);
        sl.mode := omRegRel;
        sl.offs := StaticLinkAdr;
        WHILE n > 2 DO BEGIN
          load (sl);
          sl.mode := omRegRel;
          sl.offs := StaticLinkAdr;
          DEC (n)
        END;
        load(sl);
        GenPush(sl)
      END
    END;
    GenCall (rccall, proc.linkadr, proc.conval^.intval - 4, node);
    IF (x.mode <> omCProc) AND (proc.linkadr < 0) THEN (* forward declared procedure -> fixup chain *)
      proc.linkadr := -2 - (pc - 1);;
    EndCall(proc, NIL)
  END ELSE IF x.mode = omXProc THEN BEGIN(* external procedure *)
    GenCall (rcxcall, x.adr, proc.conval^.intval - 4, node);
    EndCall(proc, NIL)
  END ELSE IF x.mode IN [omTProc, omTyp] THEN BEGIN(* type bound procedure *)
    IF node.left.subcl = 1 THEN BEGIN(* super call *)
      IF node.right.obj.typ.form = sfPointer THEN
        GetTdAdr (node.right.obj.typ.BaseTyp.BaseTyp, tag)
      ELSE
        GetTdAdr (node.right.obj.typ.BaseTyp, tag);
      load(tag)
    END ELSE BEGIN
      IF node.obj.mode = omVarPar (* node.left.obj.mode = VarPar *) THEN	(* << mb 21.2.94 *)
        tag := lastTD
      ELSE BEGIN(* pointer *)
        tag.mode := omRegRel;
        tag.typ := OPT.ptrtyp;
        tag.node := NIL;
        tag.adr := ESP;
        tag.offs := 0 (* obj.conval^.intval - 12 *); tag.inx := none;
        DeRef(tag);
        tag.mode := omRegRel;
        tag.offs := -4;
        IF OPM.TDMask <> 0 THEN BEGIN
          t1 := tag;
          load(t1);
          MakeCon(t0, OPT.linttyp, OPM.TDMask);
          Gen3(rcand, tag, t1, t0);
        END;
        DeRef(tag);
      END;
    END;
    x.adr := tag.adr; (* x.offs set by OPV.Designator *)
    x.mode := omRegRel;
    x.typ := OPT.ptrtyp; load (x);
    GenCall(rccallReg, x.adr, proc.conval^.intval - 4, node);
    EndCall(proc, NIL)
  END ELSE BEGIN(* procedure variable *)
    ASSERT (x.typ.form = sfProcTyp);
    load(x);
    param := x.typ.link; 
    IF param = NIL THEN
      GenCall(rccallReg, x.adr, 0, node) (* no parameter *)
    ELSE BEGIN
      GenCall(rccallReg, x.adr, param.adr, node);
      EndCall(proc, NIL)
    END
  END
END;
	
PROCEDURE PopResult(n: OPT.Nodes; VAR z: Item);
BEGIN
  IF n <> NIL THEN BEGIN(* result into ST(0) or EAX *)
    IF n.typ.form IN [sfByte..sfLInt, sfSet, sfPointer, sfProcTyp] THEN BEGIN
      z.typ := n.typ;
      z.mode := omReg;
      z.adr := pc;
      GenPopReg (OPL.useEAX, z.typ.form)
    END ELSE IF n.typ.form = sfHInt THEN BEGIN
      z.typ := n.typ;
      z.mode := omReg;
      z.adr := pc;
      z.offs := pc-1;
      GenPopReg (OPL.useEAX, OPT.linttyp.form);
      OPL.InstrTab[pc-1].src1 := z.offs;
      INC(OPL.InstrTab[pc-1].op, 5);	(*LInt -> HInt size *)
    END ELSE IF n.typ.comp IN [csStaticArr, csSDynArr, csRecord] THEN
      GenPopReg (OPL.noHint, 0) (* no result, already copied *)
    ELSE BEGIN(* z.typ.form IN realSet *)
      ASSERT (n.typ.form IN realSet);
      z.typ := n.typ;
      z.mode := omReg;
      z.adr := pc;
      GenPopReg (OPL.useST, z.typ.form) (* top of stack contains result *)
    END
  END ELSE
    GenPopReg (OPL.noHint, 0); (* no result *)
END;

PROCEDURE AllocSDynArr(VAR proc: OPT.Objects);
VAR
  locals: OPT.Objects;
  t, t0, size, len, x, adj, shift: Item;
BEGIN
  locals:=proc.scope.scope;
  WHILE locals<>NIL DO BEGIN
    IF (locals.typ.comp = csSDynArr)  THEN BEGIN
      x.mode := omVar;
      x.typ := locals.typ;
      x.adr := locals.linkadr;
      x.node :=  NIL;
      x.mnolev := locals.mnolev;	(*SDArr*)
      MakeVar(len, locals.typ.link);
      GenDimTrap(len);
      ArrayBaseSize(size, x);
      IF (len.mode=omCon) AND (size.mode=omCon) THEN
        size.adr:=size.adr*len.adr
      ELSE BEGIN
        t := size;
        FNMul(size, t, len, sfLInt)
      END;
      ASSERT(size.mode IN [omCon, omReg]{, 130});
      MakeCon(adj, OPT.linttyp, 3);
      t0:=size;
      Gen3(rcadd, size, t0, adj);
      MakeCon(shift, OPT.linttyp, 2);
      t0:=size;
      Gen3(rcsar, size, t0, shift);
      Gen0(rcclear);
      OPL.InstrTab[pc-1].src1:=size.adr;
      t := Stack;
      FNMove (x, t)
    END;
    locals:=locals.link
  END
END;
	
PROCEDURE FNEnter (proc: OPT.Objects; dataSize: LONGINT; node: OPT.Nodes);
VAR
  par: OPT.Objects;
  x, z, t0, t1, size: Item;
  step, scale: SHORTINT;
  useRef: BOOLEAN;
BEGIN (* !!later due to a bug this procedure looks no longer very nice *)
  useRef := (proc <> NIL) AND (proc.sysflag <> 0);
  IF proc <> NIL THEN BEGIN
    OPLi386.FixupLocalProcCall(proc); (* fixup local procedure calls and assign target pc to procedure *)
    par := proc.link
  END ELSE
    par := NIL;
  Gen1 (rcenter, dataSize, node);
  WHILE par <> NIL DO BEGIN (* copy val parameter SDynArr, OpenArr *)
    IF par.mode = omVar THEN BEGIN
      NewItem(z);
      NewItem(x);
      IF (par.typ.comp IN [csOpenArr, csSDynArr]) THEN BEGIN
        MakeVar(x, par);
        TypeSize(x, size, step, scale);
        ASSERT(step IN [OPOi386.Bit8, OPOi386.Bit16, OPOi386.Bit32]);
        ASSERT(scale IN [OPOi386.Scale1, OPOi386.Scale2, OPOi386.Scale4]);
        load(size);
        ASSERT(size.mode=omReg);
        Gen2(rcneg, t0, size);
        z.mode:=omRegRel;
        z.adr:=ESP;
        z.inx:=t0.adr;
        z.scale:=scale;
        z.typ:=OPT.linttyp;
        loadAdr(z);
        OPL.InstrTab [pc-1].hint := OPL.useESP;
        IF step <> OPOi386.Bit32 THEN	BEGIN		(*clear the last 2 bits -> grows down!*)
          z.typ := OPT.linttyp;
          MakeCon(t0, OPT.linttyp, -4);
          t1 := z;
          Gen3 (rcand, z, t1, t0);
          OPL.InstrTab [pc-1].hint := OPL.useESP
        END;
        MoveBlock(z, x, size, step);
        MakeVar(x, par);
        x.typ := OPT.linttyp;
        z.mode:=omReg;
        z.adr:=ESP;
        FNMove(x, z);
      END ELSE IF par.typ.comp = csDynArr THEN BEGIN
        MakeCon(x, OPT.linttyp, 0);
        GenPush(x); (*allocate*)
        MakeVar(z, par);
        x.mode:=omRegRel;
        x.adr:=ESP;
        x.offs:=0;
        x.inx:=none;
        x.typ:=par.typ;
        x.descReg:=none;
        CopyDynArr(x, z);
        GenPop(x);
        MakeVar(z, par);
        FNMove(z, x);
      END ELSE IF useRef AND (par.typ.comp IN [csStaticArr, csRecord]) THEN BEGIN(* copy array, record *)
        z.mode := omVar;
        z.typ := OPT.linttyp;
        z.node := NIL;
        z.adr := par.linkadr;
        z.mnolev := par.mnolev;
        x.mode := omVarPar;
        x.typ := OPT.linttyp;
        x.adr := par.adr;
        x.node := NIL;
        x.mnolev := par.mnolev;
        MakeCon (size, OPT.linttyp, par.typ.size);
        MoveBlock (z, x, size, OPOi386.Bit8)
      END
    END;
    par := par.link
  END
END;

PROCEDURE FNReturn(VAR res: Item; proc: OPT.Objects); (* RESULT:= res *)
VAR
  res1, t, size, negsize: Item; pReg: LONGINT;
  procform: SHORTINT;
  typ: OPT.Structs;
  sp, ref, t1: Item;
  step, scale: SHORTINT;
  savedPC, savedFP, callerTOS: Item;
BEGIN
  typ := proc.typ;
  procform:=typ.form;
  IF res.mode = omCoc THEN BEGIN
    t := res;
    IF (t.offs = aNil) AND (t.inx = aNil) THEN BEGIN
      MoveCoc (res, t);
      OPL.InstrTab [res.adr].hint := OPL.useEAX
    END ELSE BEGIN
      MoveCoc (res, t);
      ASSERT (OPL.InstrTab [pc-2].op = rcphi);
      pReg := OPL.InstrTab [pc-2].src1;
      OPL.InstrTab [pReg].hint := OPL.useEAX
    END
  END ELSE BEGIN
    IF res.typ.form = sfNoTyp THEN BEGIN(* RESULT:= procedure *)
      IF res.mnolev = 0 THEN (* local procedure assignment *)
        GenLoadProc (rcldProc, res, _SYSTEM._MOD(res.node.obj.adr,$10000), res.node)
      ELSE
        GenLoadProc (rcldXProc, res, _SYSTEM._DIV(res.node.obj.adr,$10000), res.node);
      OPL.InstrTab [res.adr].hint := OPL.useEAX
    END ELSE IF res.mode IN [omXProc, omLProc] THEN BEGIN(* RESULT:= procedure *)
      //OPM.LogWLn;OPM.LogWStr('ret proc');
      IF res.mnolev = 0 THEN (* local procedure assignment *)
        GenLoadProc (rcldProc, res, _SYSTEM._MOD(res.node.obj.adr,$10000), res.node)
      ELSE
        GenLoadProc (rcldXProc, res, _SYSTEM._DIV(res.node.obj.adr,$10000), res.node);
      OPL.InstrTab [res.adr].hint := OPL.useEAX
    END ELSE IF (procform <> res.typ.form) AND (res.typ.form <> sfNilTyp) THEN BEGIN
      ASSERT (procform <> sfSet);
      Convert (res, procform)
    END;
    IF res.typ.form IN realSet THEN
      loadf (res)
    ELSE IF typ.comp IN [csOpenArr, csSDynArr] THEN BEGIN
        (* prepare callerTOS *)
      callerTOS.mode:=omRegRel;
      callerTOS.adr:=FP;
      callerTOS.offs:=proc.conval^.intval;
      callerTOS.inx:=none;
      callerTOS.typ:=OPT.linttyp;
      load(callerTOS);
      sp:=callerTOS;
        (*push the dimensions at 0[callerTOS] *)
      sp.mode:=omRegRel;
      sp.offs:=0;
      PushArrLen(proc(*formal*), res(*actual*), sp, TRUE);
        (*allocate pointer to the data [ref]*)
      DEC(sp.offs, 4); ref:=sp;
        (*compute array size, align to 4, reserve stack*)
      TypeSize(res, size, step, scale);
      IF size.mode=omCon THEN BEGIN
        MakeCon(negsize, OPT.linttyp, -size.adr - _SYSTEM._MOD(-size.adr,4));
        INC(sp.offs, negsize.adr);
        IF res.mode<>omVar THEN load(negsize);	(*only if needed!*)
      END ELSE BEGIN
        load(size);
        negsize.typ:=OPT.linttyp;
        Gen2(rcneg, negsize, size);
        IF (step IN [OPOi386.Bit8, OPOi386.Bit16]) THEN BEGIN
          MakeCon(t, OPT.linttyp, -4);
          t1:=negsize;
          Gen3(rcand, negsize, t1, t);	(* negsize AND FFFFFFF7 *)
        END;
        sp.inx:=negsize.adr;
        sp.scale:=scale;
      END;
        (* if the RESULT:=ed value is not local, allocate stack for it *) 
      IF res.mode<>omVar THEN BEGIN
        t:=Stack;
        t.mode:=omRegRel;
        t.offs:=-12;
        t.inx:=negsize.adr;
        t.scale:=scale;	(* LEA SP, -12[SP][size*scale]  -12 to cover the dims, ref and link*)
        loadAdr(t);
        OPL.InstrTab[pc-1].hint:=OPL.useESP
      END;
        (* assign pointer to data [ref] *)
      loadAdr(sp);
      t:=sp;
      GenStore(rcstore, ref, t);
      sp.mode:=omRegRel;
      sp.offs:=0;
      sp.inx:=none;	(*load current sp, frees the register with negsize*)
        (* save PC' / EBP' *)
      savedFP.mode:=omRegRel;
      savedFP.adr:=FP;
      savedFP.offs:=0;
      savedFP.inx:=none;
      savedFP.typ:=OPT.linttyp;
      load(savedFP);
      GenPush(savedFP);
      savedPC.mode:=omRegRel;
      savedPC.adr:=FP;
      savedPC.offs:=4;
      savedPC.inx:=none;
      savedPC.typ:=OPT.linttyp;
      load(savedPC);
      GenPush(savedPC);
        (* copy the result *)
      t:=sp;
      MoveBlockReversed(t, res, size, step);
        (* store link to previous block / callerTOS; PC'/EBP' *)
      t:=sp;
      DEC(t.offs, 4);
      GenStore(rcstore, t, callerTOS);
      t:=sp;
      DEC(t.offs, 8);
      GenPop(savedPC);
      GenStore(rcstore, t, savedPC);
      t:=sp;
      DEC(t.offs, 12);
      GenPop(savedFP);
      GenStore(rcstore, t, savedFP);
        (* set the new SP *)
      DEC(sp.offs, 12);
      loadAdr(sp);
      OPL.InstrTab[pc-1].hint:=OPL.useESP;
      ASSERT(sp.typ=OPT.linttyp)
    END ELSE IF typ.comp = csDynArr THEN BEGIN
      t.mode := omRegRel;
      t.adr := FP;
      t.offs := proc.conval^.intval;
      t.typ := res.typ;
      t.inx := none;
      load(t);
      t.mode:=omRegRel;
      t.offs:=0;
      IF res.mode=omVar THEN	(*local, no need to copy it, RESULT:= the var self*)
        FNMove(t, res)
      ELSE
        CopyDynArr(t, res);
    END ELSE IF typ.form = sfComp THEN BEGIN
      ASSERT(typ.comp <> csDynArr);
      t.mode := omRegRel;
      t.adr := FP;
      t.offs := proc.conval^.intval;
      t.typ := res.typ;
      t.inx := none;
      load(t);
      IF typ.comp = csRecord THEN BEGIN
        size.mode := omRegRel;
        size.adr := FP;
        size.offs := proc.conval^.intval+4;
        size.inx := none;
        size.typ := OPT.linttyp
      END ELSE
        MakeCon(size, OPT.linttyp, typ.size);
      MoveBlock(t, res, size, OPOi386.Bit8)
    END ELSE IF typ = OPT.hinttyp THEN BEGIN
      IF res.mode = omReg THEN BEGIN
        OPL.InstrTab[res.adr].hint := OPL.useEAX;
        OPL.InstrTab[res.offs].hint := OPL.useEDX
      END ELSE BEGIN
        res.typ := OPT.linttyp;
        loadAdr(res);
        res.mode := omRegRel;
        res1 := res;
        INC(res1.offs, 4);
        load(res);
        OPL.InstrTab[res.adr].hint := OPL.useEAX;
        load(res1);
        OPL.InstrTab[res1.adr].hint := OPL.useEDX
      END
    END ELSE BEGIN
      load (res);
      OPL.InstrTab [res.adr].hint := OPL.useEAX
    END
  END
END;
	
PROCEDURE FNExit(proc: OPT.Objects);
VAR
  size: LONGINT;
  ebp: Item;
BEGIN
  IF proc = NIL THEN BEGIN(* from module *)
    Gen1 (rcleave, 0, NIL);
    Gen1 (rcret, 0, NIL)
  END ELSE IF proc.typ.comp IN [csOpenArr, csSDynArr] THEN BEGIN
    ebp.typ:=OPT.linttyp;
    GenPop(ebp);
    OPL.InstrTab [pc-1].hint := OPL.useEBP;
    Gen1 (rcret, 0, NIL);
  END ELSE BEGIN(* procedure *)
    Gen1 (rcleave, proc.conval^.intval2, NIL);
    IF proc.sysflag = cdecl THEN
      size := 0
    ELSE IF (proc.mode=omTyp) OR (proc.typ.form <> sfComp) THEN
      size := proc.conval^.intval - 8
    ELSE IF proc.typ.comp = csRecord THEN
      size := proc.conval^.intval - 8 + 8
    ELSE
      size := proc.conval^.intval - 8 + 4;
    Gen1 (rcret, size, NIL)
  END;
END;

PROCEDURE Lock(self, excl: Item);		(* active only *)
BEGIN
  (* push address of the record *)
  IF self.typ.comp = csRecord THEN
    loadAdr (self)
  ELSE (*pointer*)
    load (self);
  GenPush (self);
  GenPush (excl);	(* -> don't load (excl) because it's a boolean constant. This generates better code *)
  GenCall (rcxcall, OPLi386.LockIndex * $10000 + OPLi386.LockEntryNr, 4, NIL);
END;

PROCEDURE Unlock(self, excl: Item);		(* active only *)
BEGIN
  (* push address of the record *)
  IF self.typ.comp = csRecord THEN
    loadAdr (self)
  ELSE (*pointer*)
    load (self);
  GenPush (self);
  GenPush (excl);	(* -> don't load (excl) because it's a boolean constant. This generates better code *)
  GenCall (rcxcall, OPLi386.UnlockIndex * $10000 + OPLi386.UnlockEntryNr, 4, NIL);
END;

PROCEDURE Await(z, self, y: Item);	(* active only *)
BEGIN
  IF z.typ = OPT.niltyp THEN
    MakeCon(z, OPT.linttyp, -1)
  ELSE
    GenLoadProc(rcldProc, z, _SYSTEM._MOD(z.node.obj.adr,$10000), z.node);		(* !! z is a local procedure !! *)
  GenPush(z);
  z.mode := omReg;
  z.adr := FP;
  GenPush(z);
  IF self.mode <> omCon THEN BEGIN
  (* push address of the record *)
    IF self.typ.comp = csRecord THEN
      loadAdr (self)
    ELSE (*pointer*)
      load (self);
  END;
  GenPush(self);
  GenPush(y);
  GenCall(rcxcall, OPLi386.AwaitIndex * $10000 + OPLi386.AwaitEntryNr, 8, NIL);
END;

PROCEDURE FNTrap(n: LONGINT; node: OPT.Nodes); (* HALT (n) *)
BEGIN
  Gen1 (rctrap, n, node)
END;

PROCEDURE GenAsm(n: OPT.Nodes);		(* iASM *)
BEGIN
  Gen1(rcassembler, 0, n)
END;

PROCEDURE GenDimTrap(VAR len: Item);
VAR
  Vconst: Item;
BEGIN
  IF (OPM.inxchk IN OPM.options) THEN BEGIN
    MakeCon (Vconst, OPT.linttyp, -1);
    load (len);
    GenFlags (rccmp, len, Vconst);
    Gen1 (rctle, OPL.DimTrap , NIL)
  END
END;

PROCEDURE FNCase(VAR x: Item; low, high: LONGINT; VAR tab: LONGINT; VAR L: BLabel; node: OPT.Nodes);
VAR
  c, y: Item;
BEGIN
  MakeCon (c, OPT.linttyp, low);
  IF x.typ.form <> sfLInt THEN BEGIN
    IF x.mode = omCon THEN BEGIN
      x.typ := OPT.linttyp;
      load (x)
    END ELSE
      Convert (x, sfLInt);
  END ELSE
    load (x);
  IF low <> 0 THEN
    Gen3 (rcsub, x, x, c);
  c.adr := high-low;
  GenFlags (rccmp, x, c);
  L := -2-pc;
  Gen1 (rcja, aNil, node);
  OPL.AllocCaseTab (low, high, tab);
  c.mode := omAbs;
  c.typ := OPT.linttyp;
  c.adr := tab;
  c.inx := x.adr;
  c.scale := OPOi386.Scale4;
  c.node := NIL;
  GenLoad (rcld, y, c);
  Gen1 (rcjmpReg, y.adr, node)
END;
	
PROCEDURE CaseFixup (tab: LONGINT; elseLabel: BLabel; len: LONGINT); (* case *)
BEGIN
  InitInstr (rccase, 0, 0, NIL);
  OPL.InstrTab [pc].src1 := tab;
  OPL.InstrTab [pc].src2 := elseLabel;
  OPL.InstrTab [pc].inx := len * 4;
  INC (pc)
END;

PROCEDURE OPCBegin;
BEGIN
	MakeCon (VTrue, OPT.booltyp, Otrue);
	MakeCon (VFalse, OPT.booltyp, Ofalse);
	Stack.mode := omReg;
  Stack.adr := ESP;
  Stack.node := NIL;
  Stack.typ := OPT.linttyp;
  Stack.offs := 0;
  Stack.inx := none;
	ebp.mode := omReg;
  ebp.adr := FP;
  ebp.node := NIL;
  ebp.typ := OPT.linttyp;
	formTab[sfUndef]    := 0;
	formTab[sfByte]     := OPOi386.Bit8  DIV 8;
  formTab[sfBool]     := OPOi386.Bit8  DIV 8;
	formTab[sfChar]     := OPOi386.Bit8  DIV 8;
  formTab[sfSInt]     := OPOi386.Bit8  DIV 8;
	formTab[sfInt]      := OPOi386.Bit16 DIV 8;
	formTab[sfLInt]     := OPOi386.Bit32;
  formTab[sfSet]      := OPOi386.Bit32;
	formTab[sfPointer]  := OPOi386.Bit32;
  formTab[sfProcTyp]  := OPOi386.Bit32;
	formTab[sfNilTyp]   := OPOi386.Bit32;
  formTab[sfComp]     := OPOi386.Bit32;
	formTab[sfReal]     := 3;
	formTab[sfLReal]    := 4;
	CCtab[0,0] := EQ;
  CCtab [0,1] := NE;
  CCtab [0,2] := LT;
  CCtab [0,3] := LE;
  CCtab [0,4] := GT;
  CCtab [0,5] := GE;
	CCtab[1,0] := EQ;
  CCtab [1,1] := NE;
  CCtab [1,2] := BL;
  CCtab [1,3] := BE;
  CCtab [1,4] := AB;
  CCtab [1,5] := AE;
	InvCCtab [0] := NE;
  InvCCtab [1] := EQ;
  InvCCtab [2] := GE;
  InvCCtab [3] := GT;
  InvCCtab [4] := LE;
  InvCCtab [5] := LT;
	InvCCtab [6] := BE;
  InvCCtab [7] := BL;
  InvCCtab [8] := AE;
  InvCCtab [9] := AB;
  InvCCtab [10] := CC;
  InvCCtab [11] := CS;
END;

END.
