UNIT OPVi386;

INTERFACE

USES
  Classes, OPT, OPM, OPCi386, OPOi386, OPL, OPLi386, OPV, _SYSTEM,
  SYSUTILS;

CONST
  MaxAdr 	= MaxLInt;
  MinAdr 	= MinLInt;

  RecVarParSize 	= 8;
  VarParSize 	= OPM.PointerSize;

VAR
  findpc: BOOLEAN;
  ExitChain: OPCi386.BLabel;
  allocTd: BOOLEAN;
  dummy: OPCi386.Item;

  PROCEDURE Module(prog: OPT.Nodes);

IMPLEMENTATION

USES
  Unit1;

//############################## RefFile #################################

PROCEDURE RefW(VAR S: STRING; ch: CHAR);
BEGIN
  S:=S+Ch;
END;

PROCEDURE RefWNum(VAR S: STRING; x: INT64);
BEGIN
  WHILE (x<-64) OR (x>63) DO BEGIN
    S:=S+CHR(_SYSTEM._MOD(x,$80)+$80);
    x := _SYSTEM._DIV(x,$80);
  END;
  S:=S+CHR(_SYSTEM._MOD(x,$80));
END;

PROCEDURE OutRefPoint (VAR F: STRING; proc: OPT.Objects);
VAR
  obj: OPT.Objects;
  n: LONGINT;
BEGIN
  IF proc = NIL THEN BEGIN
    RefW (F,RefTag);
    RefW (F,#$0F8);
    RefWNum (F,0);
  END ELSE BEGIN
    IF NewRef THEN BEGIN
      RefW(F,#$0F9);
      RefWNum(F,proc.linkadr);
      obj := proc.link;
      n := 0;
      WHILE obj <> NIL DO BEGIN
        INC(n);
        obj := obj.link
      END;
      RefWNum(F,n);
      n := proc.typ.form;
      IF n IN [sfUndef, sfNilTyp, sfNoTyp] THEN
        n := 0
      ELSE IF n = sfComp THEN
        n := $10 + proc.typ.comp;
      RefW(F,CHR(n));
      RefW(F,CHR(proc.mnolev));
      IF slNeeded IN proc.conval^.setval THEN
        RefW(F,#1)
      ELSE
        RefW(F,#0)
    END ELSE BEGIN
      RefW (F,#$0F8);
      RefWNum (F,proc.linkadr)
    END
  END
END;

PROCEDURE OutRefName(VAR F: STRING; name: STRING);
  VAR
    ch: CHAR;
    i: INTEGER;
BEGIN
  i := 1;
  REPEAT
    ch := name [i];
    RefW(F,ch);
    INC (i)
  UNTIL ch = #0
END;

PROCEDURE OutRefs(VAR F: STRING; obj: OPT.Objects);
VAR
  Tf: INTEGER;
BEGIN
  IF obj <> NIL THEN BEGIN
    OutRefs(F,obj.left);
    IF obj.mode IN [omVar, omVarPar] THEN BEGIN
      Tf := obj.typ.form;
      IF NewRef THEN BEGIN

        IF (Tf IN [sfByte..sfSet, sfPointer, sfProcTyp]) OR (obj.typ.comp = csRecord) OR
            (obj.typ.comp IN [csStaticArr, csOpenArr])
            AND (obj.typ.BaseTyp.form IN [sfByte..sfSet, sfPointer, sfProcTyp]) THEN BEGIN

          IF (obj.mode = omVar) AND (obj.typ.comp <> csOpenArr) THEN
            RefW (F,#1)	(* direct *)
          ELSE
            RefW (F,#3);	(* indirect *)

          IF obj.typ.comp = csStaticArr THEN BEGIN
            RefW (F,CHR ($80+obj.typ.BaseTyp.form));
            RefWNum (F,obj.typ.n)
          END ELSE IF obj.typ.comp = csOpenArr THEN BEGIN
            RefW (F,CHR ($80 + obj.typ.BaseTyp.form));
            RefWNum (F,0)
          END ELSE IF obj.typ.comp = csRecord THEN BEGIN
            RefW(F,CHR($10+csRecord));
            RefWNum(F,obj.typ.tdadr)
          END ELSE IF (Tf = sfPointer) AND (obj.typ.BaseTyp <> NIL)
          AND (obj.typ.BaseTyp.comp = csRecord) THEN BEGIN
            RefW(F,CHR($10+sfPointer));
            RefWNum(F,obj.typ.BaseTyp.tdadr)
          END ELSE
            RefW (F,CHR (Tf));

          RefWNum (F,obj.linkadr);
          OutRefName(F,obj.name)

        END;

      END ELSE BEGIN
        IF (Tf IN [sfByte..sfSet, sfPointer, sfProcTyp]) OR
            (obj.typ.comp IN [csStaticArr, csOpenArr])
            AND (obj.typ.BaseTyp.form IN [sfByte..sfSet, sfPointer, sfProcTyp]) THEN BEGIN

          IF (obj.mode = omVar) AND (obj.typ.comp <> csOpenArr) THEN
            RefW (F,#1)	(* direct *)
          ELSE
            RefW (F,#3);

          IF obj.typ.comp = csStaticArr THEN BEGIN
            RefW (F,CHR ($80 + obj.typ.BaseTyp.form));
            RefWNum (F,obj.typ.n)
          END ELSE IF obj.typ.comp = csOpenArr THEN BEGIN
            RefW (F,CHR ($80 + obj.typ.BaseTyp.form));
            RefWNum (F,0)
          END ELSE
            RefW (F,CHR (Tf));

          RefWNum (F,obj.linkadr);
          OutRefName (F,obj.name);

        END
      END
    END;
    OutRefs (F, obj.right)
  END
END;

//######################################################################

PROCEDURE Expression (n: OPT.Nodes; VAR z: OPCi386.Item; selfForUnlock: OPT.Nodes); FORWARD;
PROCEDURE Call(proc: OPT.Nodes; VAR res: OPCi386.Item; assignment: BOOLEAN);FORWARD;
PROCEDURE Designator (n: OPT.Nodes; VAR x: OPCi386.Item); FORWARD;
PROCEDURE Statement (n, selfForUnlock: OPT.Nodes); FORWARD;

PROCEDURE IncAdr(VAR adr: LONGINT; s: LONGINT);
BEGIN
  IF (s >= 0) AND (adr <= MaxAdr - s) OR (s < 0) AND (adr >= MinAdr - s) THEN
    INC(adr, s)
  ELSE
    OPM.err(242)
END;

PROCEDURE AlignFld(VAR offset: LONGINT; elemsize: LONGINT);
BEGIN
  IF elemsize >= 4 THEN IncAdr(offset, (-offset) MOD 4)
  ELSE IF elemsize = 2 THEN IncAdr(offset, offset MOD 2);
END;

PROCEDURE AllocFld(VAR offset, this: LONGINT; elemsize: LONGINT);
BEGIN
  this := offset;
  IncAdr(offset, elemsize)
END;
	
PROCEDURE AlignRec(VAR size: LONGINT);
BEGIN 
  IncAdr(size, (-size) MOD 4)
END;

PROCEDURE AlignPar(VAR elemsize: LONGINT);
BEGIN
  IncAdr(elemsize, (-elemsize) MOD 4)
END;

FUNCTION ArgSize (par: OPT.Objects): LONGINT;
VAR
  size: LONGINT;
  comp: SHORTINT;
  typ: OPT.Structs;
BEGIN
  size:= 0;
  WHILE par <> NIL DO BEGIN
    typ:= par.typ;
    comp:= typ.comp;
    IF (comp=csOpenArr) THEN
      INC (size, typ.size)
    ELSE IF par.mode = omVarPar THEN BEGIN
      IF comp = csRecord THEN
        INC (size, 8)
      ELSE
        INC (size, 4);
    END ELSE IF par.typ.form = sfLReal THEN
      INC (size, 8)
    ELSE IF comp IN [csStaticArr, csDynArr, csRecord] THEN BEGIN
      INC(size, typ.size);
      AlignPar(size);
    END ELSE
      INC (size, 4);
    par:= par.link
  END;
  AlignPar (size);
  RESULT:= size;
END;

PROCEDURE CallLR(proc: OPT.Nodes; VAR res: OPCi386.Item; assignment: BOOLEAN);
VAR
  firstap, ap: OPT.Nodes;
  firstfp, fp: OPT.Objects;
  x, dynptr, tos: OPCi386.Item;
  s, dynblocks, staticsize: LONGINT;
  Vtproc: BOOLEAN;
	
  PROCEDURE AllocTempSpace(VAR dynblocks, staticsize: LONGINT; ap: OPT.Nodes; fp: OPT.Objects);	(*traverse in reverse order!*)
  VAR
    tos: OPCi386.Item;
  BEGIN
    IF ap <> NIL THEN BEGIN
      AllocTempSpace(dynblocks, staticsize, ap.link, fp.link);
      IF (ap.clas = Ncall) AND (ap.typ.form=sfComp) THEN BEGIN
        IF ap.typ.comp IN [csOpenArr, csSDynArr] THEN	BEGIN(*dynamic*)
          OPCi386.AllocateStatic(tos, ap.typ.n*4+8);	(*get tos, allocate space for the len and adrptr*)
          INC(tos.offs, ap.typ.n*4+8);
          tos.typ := OPT.linttyp (*ap.typ*);
          Call(ap, tos, FALSE);	(*the procedure allocates the space when the array is copied back!*)
          INC(dynblocks)
        END ELSE IF (fp.typ.comp=csOpenArr) OR (fp.mode = omVarPar) THEN	(*static, indirect*)
          INC(staticsize, ap.typ.size + (-ap.typ.size) MOD 4);
      END
    END
  END;

BEGIN
  IF NOT assignment THEN OPCi386.PushRegs;	(*be sure nothing is pushed in that case!*)
  firstap := proc.right;
  firstfp := proc.obj;
  Vtproc := (proc.left.obj<>NIL) AND (proc.left.obj.mode = omTProc);
  IF Vtproc THEN BEGIN
    firstap := firstap.link;
    firstfp := firstfp.link
  END;	(*skip self*)
		
    (*temporary values*)
  staticsize := 0; dynblocks := 0;
  AllocTempSpace(dynblocks, staticsize, firstap, firstfp);
  OPCi386.AllocateStatic(tos, staticsize);
		
    (*return descriptor*)
  IF proc.typ.form=sfComp THEN BEGIN
    OPCi386.CorrectStackAdr(res, staticsize, dynblocks);	(*correction for static allocated ret areas*)
    OPCi386.PushRetDesc(res, tos);
  END;
  dynptr := tos;
  OPCi386.CorrectStackAdr(dynptr, staticsize, 0);

    (*parameters*)
  ap := firstap;
  fp := firstfp;
  WHILE ap <> NIL DO BEGIN
    IF (ap.clas = Ncall) AND (ap.typ.form = sfComp) THEN BEGIN
      IF ap.typ.comp IN [csOpenArr, csSDynArr] THEN	BEGIN(*dynamic, indirect*)
        x := dynptr;
        OPCi386.DeRef(dynptr);
        x.typ := ap.typ;
        INC(x.offs, 4);
        x.descReg:=dynptr.adr;
        x.descOffs := dynptr.offs - 4;
        OPCi386.Parameter(x, fp, FALSE, FALSE);
      END ELSE IF (*(ap.typ.comp=StaticArr)&*)(fp.typ.comp=csOpenArr) OR (fp.mode = omVarPar) THEN BEGIN	(*static, indirect*)
        x := tos;
        x.typ := ap.typ;
        Call(ap, x, FALSE);
        x := tos;
        x.typ := ap.typ;
        OPCi386.Parameter(x, fp, FALSE, FALSE);
        OPCi386.CorrectStackAdr(tos, ap.typ.size+(-ap.typ.size) MOD 4, 0);	(*prepare for next static*)
      END ELSE BEGIN	(*static, direct*)
        OPCi386.AllocateStatic(x, fp.typ.size);
        x.typ:=fp.typ;
        Call(ap, x, FALSE)
      END
    END ELSE BEGIN
      Expression(ap, x, NIL);
      OPCi386.Parameter(x, fp, ap.clas=Nderef, FALSE);
    END;
      (*correct tos*)
    IF (fp.mode = omVarPar) THEN BEGIN
      IF (fp.typ.comp = csRecord) THEN
        s:= RecVarParSize
      ELSE IF (fp.typ.comp = csOpenArr) THEN
        s:= fp.typ.size
      ELSE
        s:= VarParSize;
    END ELSE
      s:= fp.typ.size;
    AlignPar(s);
    OPCi386.CorrectStackAdr(tos, s, 0);
    OPCi386.CorrectStackAdr(dynptr, s, 0);
    ap := ap.link;
    fp := fp.link
  END;

  IF Vtproc THEN BEGIN
    Expression(proc.right, x, NIL);
    OPCi386.Parameter(x, proc.obj,proc.right.clas=Nderef, FALSE)
  END;

    (*activation*)
  IF proc.subcl = 2 THEN	(*inlined*)
    Statement(proc.left, NIL)
  ELSE BEGIN
    Designator(proc.left, x);
    IF proc.left.clas = Nproc THEN
      OPCi386.FNCall(x, proc.left.obj, proc)
    ELSE
      OPCi386.FNCall(x, NIL, proc);
  END;
    (*cleanup*)
  IF NOT (proc.typ.comp IN [csOpenArr, csSDynArr]) THEN BEGIN
    OPCi386.RemoveStatic(staticsize);
    OPCi386.ResetStack(dynblocks)
  END;
  IF (proc.typ=OPT.notyp) OR (proc.typ.form=sfComp) THEN  (* pop register and put result into z, also restore regs *)
    OPCi386.PopResult(NIL, res)
  ELSE
    OPCi386.PopResult(proc, res);
END;

PROCEDURE CallRL(proc: OPT.Nodes; VAR res: OPCi386.Item; assignment: BOOLEAN);
VAR
  ap: OPCi386.Item;
  n, p, q: OPT.Nodes;
  fp, formPar: OPT.Objects;
BEGIN
  n := proc.right;
  fp := proc.obj;
  IF NOT assignment THEN OPCi386.PushRegs;
  IF n <> NIL THEN BEGIN
    p := n;
    formPar := fp;
    WHILE p.link <> NIL DO BEGIN
      p := p.link;
      formPar := formPar.link
    END;
    WHILE p <> n DO BEGIN
      Expression (p, ap, NIL);
      OPCi386.Parameter (ap, formPar, p.clas = Nderef, TRUE); (* mh 16.2.94 *)
      q := n;
      formPar := fp;
      WHILE q.link <> p DO BEGIN
        q := q.link;
        formPar := formPar.link
      END;
      p := q
    END;
    Expression (n, ap, NIL);
    OPCi386.Parameter (ap, fp, n.clas = Nderef, TRUE); (* mh 16.2.94 *)
  END;
  IF proc.subcl = 2 THEN	(*inlined*)
    Statement(proc.left, NIL)
  ELSE BEGIN
    Designator(proc.left, res);
    IF proc.left.clas = Nproc THEN
      OPCi386.FNCall(res, proc.left.obj, proc)
    ELSE
      OPCi386.FNCall(res, NIL, proc);
  END;
  IF (proc.typ=OPT.notyp) THEN
    OPCi386.PopResult(NIL, res)
  ELSE
    OPCi386.PopResult(proc, res);
END;

PROCEDURE Call(proc: OPT.Nodes; VAR res: OPCi386.Item; assignment: BOOLEAN);
VAR
  obj: OPT.Objects;
  n: OPT.Nodes;
BEGIN
  n := proc.left;
  IF (n.clas = Nproc) OR (n.clas = Nassembler) THEN	BEGIN(*procedure*)
    obj := n.obj;
    IF obj.sysflag = 0 THEN
      CallLR(proc, res, assignment)
    ELSE
      CallRL(proc, res, assignment);
  END ELSE BEGIN	(*procedure variable*)
    ASSERT(n.typ.form = sfProcTyp);
    IF n.typ.sysflag = 0 THEN
      CallLR(proc, res, assignment)
    ELSE
      CallRL(proc, res, assignment);
  END
END;

PROCEDURE Dim (VAR z, nofelem: OPCi386.Item; n: OPT.Nodes; typ: OPT.Structs; nofdim: LONGINT);
VAR
  len: OPCi386.Item;
BEGIN
  ASSERT(typ.comp IN [csStaticArr, csDynArr, csOpenArr]);
  Expression(n, len, NIL);
  IF len.mode <> omCon THEN BEGIN
    OPCi386.PushLen(len);
    OPCi386.GenDimTrap(len)
  END;

  IF nofdim = 1 THEN  BEGIN
    nofelem:=len;
    nofelem.typ:=OPT.linttyp
  END ELSE
    OPCi386.MulDim(nofelem, len);

  IF n.link <> NIL THEN
    Dim(z, nofelem, n.link, typ.BaseTyp, nofdim+1)
  ELSE BEGIN
    OPCi386.NewArray(z, nofelem, nofdim, typ.BaseTyp, TRUE);
    INC(z.offs, 4)
  END;
  IF len.mode <> omCon THEN
    OPCi386.PopLen(z)
  ELSE BEGIN
    z.typ:=OPT.linttyp;
    OPCi386.FNAssign(z, len)
  END;
  INC(z.offs, 4)
END;

PROCEDURE IfStat (n, selfForUnlock: OPT.Nodes; VAR Lfix, Lcfix: OPCi386.BLabel; Velse: BOOLEAN);
VAR
  x: OPCi386.Item;
BEGIN
  REPEAT
    OPM.errpos.CodePos := n.conval^.intval;
    Expression (n.left, x, selfForUnlock);
    OPCi386.Jncc (x, Lcfix, n);
    Statement (n.right, selfForUnlock);
    IF n.link = NIL THEN BREAK;
    OPCi386.FNJmp (Lfix, n);
    OPCi386.FixLink (Lcfix);
    n := n.link
  UNTIL FALSE;
  IF Velse THEN OPCi386.FNJmp (Lfix, n);
END;

PROCEDURE CaseStat (n, selfForUnlock: OPT.Nodes);
VAR
  Vcase, interval: OPT.Nodes;
  x: OPCi386.Item;
  L, elseLabel, dummy: OPCi386.BLabel;
  low, high, tab: LONGINT;
BEGIN
  Expression (n.left, x, selfForUnlock);
  low := n.right.conval^.intval;
  high := n.right.conval^.intval2;
  L := OPCi386.aNil;
  OPCi386.FNCase (x, low, high, tab, elseLabel, n);
  Vcase := n.right.left;
  WHILE Vcase <> NIL DO BEGIN
    OPCi386.DefLabel (dummy); (* label entry: used for common subexpression elimination *)
    interval := Vcase.left;
    WHILE interval <> NIL DO BEGIN
      OPLi386.CaseJump (OPCi386.pc, tab, interval.conval^.intval - low, interval.conval^.intval2 - low);
      interval := interval.link
    END;
    Statement(Vcase.right, selfForUnlock);
    OPCi386.FNJmp (L, n);
    Vcase := Vcase.link;
  END;
  OPCi386.FixLink (elseLabel);
  elseLabel := OPCi386.pc;
  IF n.right.conval^.setval <>[] THEN
    Statement(n.right.right, selfForUnlock) (* ELSE part *)
  ELSE
    OPCi386.FNTrap(OPL.CaseTrap, n);
  OPCi386.CaseFixup (tab, elseLabel, high - low + 1);
  OPCi386.FixLink (L)
END;

//#####################################################################

PROCEDURE Designator (n: OPT.Nodes; VAR x: OPCi386.Item);
VAR
  obj   : OPT.Objects;
  y     : OPCi386.Item;
  index : LONGINT;
  varRec: BOOLEAN;
BEGIN
  CASE n.clas OF
    Nvar, Nvarpar:
                  BEGIN
                    obj := n.obj;
                    WHILE obj.scope <> NIL DO BEGIN
                      obj := obj.scope;
                      n.obj := obj
                    END;	(* WITH-bug fix *)
                    x.node := n;
                    x.mnolev := obj.mnolev;
                    x.scale := OPOi386.noScale;
                    x.inx := OPLi386.none;
                    IF obj.mnolev <= 0 THEN BEGIN(* global or imported variable *)
                      x.mode := omAbs;
                      IF obj.mnolev < 0 THEN BEGIN(* imported variable *)
                        IF _SYSTEM._DIV(obj.adr,$10000) = 0 THEN BEGIN(* no VarConsLink index yet *)
                          OPLi386.NewVarCons (-obj.mnolev, obj.adr, index);
                          obj.adr := index * $10000 + obj.adr;	(* index entry *)
                        END;
                        x.adr := 0
                      END ELSE
                        x.adr := obj.linkadr
                    END ELSE BEGIN(* Var, VarPar *)
                      x.mode := obj.mode;
                      x.adr := obj.linkadr
                    END;
                    x.offs := 0;
                    x.descReg := OPLi386.none
                  END;
    Nfield:
                  BEGIN
                    Designator (n.left, x);
                    OPCi386.Field (x, n.obj.adr)
                  END;
    Nderef:
                  BEGIN
                    Designator (n.left, x);
                    OPCi386.DeRef (x);
                  END;
    Nindex:
                  BEGIN
                    Expression (n.right, y, NIL);
                    Designator (n.left, x);
                    OPCi386.FNIndex (x, y);
                  END;
    Nguard:
                  BEGIN
                    varRec := (n.left.clas = Nvarpar) AND (n.left.typ.comp = csRecord);
                    Designator (n.left, x);
                    OPCi386.TypeTest (x, n.typ, TRUE, FALSE, varRec)
                  END;
    Neguard:
                  BEGIN
                    varRec := (n.left.clas = Nvarpar) AND (n.left.typ.comp = csRecord);
                    Designator (n.left, x);
                    OPCi386.TypeTest (x, n.typ, TRUE, TRUE, varRec);
                  END;
    Nproc:
                  BEGIN
                    IF (n.obj.mode IN [omXProc, omTProc]) AND (n.obj.conval^.intval = -1) THEN
                      n.obj.conval^.intval := ArgSize (n.obj.link) + (* 4 *) 8;
                    OPCi386.FNProcedure (x, n);
                  END;
  END;
  x.typ := n.typ
END;

PROCEDURE Expression (n: OPT.Nodes; VAR z: OPCi386.Item; selfForUnlock: OPT.Nodes);
VAR
  x, y: OPCi386.Item;
  Tf: INTEGER;
  Vreal: Single;
  Vcon: OPT.Consts;
  typ: OPT.Structs;
  s: STRING;

  PROCEDURE AllocConst (VAR bytes: STRING; len, align: LONGINT);
  VAR
    con1, con2: OPT.Consts;
  BEGIN
    con1 := Vcon;
    IF n.obj = NIL THEN
      OPL.AllocConst(bytes, len, align, con1^.intval)
    ELSE BEGIN
      con2 := n.obj.conval;
      IF con2^.intval = OPM.ConstNotAlloc THEN
        OPL.AllocConst (bytes, len, align, con2^.intval);
      con1^.intval := con2^.intval
    END
  END;
		
BEGIN
  z.node := NIL;
  CASE n.clas OF
    Nconst:
              BEGIN
                z.typ := n.typ;
                z.mnolev := 0;
                z.node := NIL; (* constants have no node *)
                Vcon := n.conval;
                CASE z.typ.form OF
                  sfByte..sfLInt, sfNilTyp, sfPointer:
                            BEGIN
                              z.mode := omCon;
                              z.adr := Vcon^.intval;
                            END;
                  sfSet:
                            BEGIN
                              z.mode := omCon;
                              z.adr := INTEGER(Vcon^.setval);
                            END;
                  sfString:
                            BEGIN
                              AllocConst (Vcon^.ext, Vcon.intval2, 4);
                              z.mode := omAbs;
                              z.adr  := Vcon^.intval;
                              z.offs := Vcon^.intval2 (* length *);
                              z.inx  := OPLi386.none
                            END;
                  sfReal:
                            BEGIN
                              Vreal := Vcon^.realval;
                              S:=_SYSTEM.ToSysByte(_SYSTEM.RealToIEEE(SINGLE(Vreal)),4);
                              AllocConst (S, 4, 4);
                              z.mode := omAbs;
                              z.adr := Vcon^.intval;
                              z.inx := OPLi386.none;
                            END;
                  sfLReal:
                            BEGIN
                              S:=_SYSTEM.ToSysByte(_SYSTEM.DoubleToIEEE(Vcon^.realval),8);
                              AllocConst (S, 8, 4);
                              z.mode := omAbs;
                              z.adr := Vcon^.intval;
                              z.inx := OPLi386.none
                            END;
                END;
              END;
    Nupto:
              BEGIN
                Expression (n.left, x, selfForUnlock);
                Expression (n.right, y, selfForUnlock);
                OPCi386.SetRange (z, x, y);
              END;
    Nmop:
              BEGIN
                IF (n.subcl <> vis) AND (n.subcl <> ytype) THEN
                  Expression (n.left, x, selfForUnlock);
                z := x;
                CASE n.subcl OF
                  vnot:   OPCi386.FNNot (z, x);
                  vminus:  OPCi386.FNNeg (z, x);
                  vis:
                          BEGIN
                            Designator (n.left, z);
                            IF n.obj.typ.form = sfPointer THEN
                              OPCi386.TypeTest (z, n.obj.typ, FALSE, FALSE, FALSE)
                            ELSE
                              OPCi386.TypeTest (z, n.obj.typ, FALSE, FALSE, TRUE);
                          END;
                  vconv:
                          BEGIN
                            IF n.typ.form = sfSet THEN
                              OPCi386.SetElem (z, x)
                            ELSE BEGIN
                              z := x;
                              OPCi386.Convert (z, n.typ.form);
                            END;
                          END;
                  vabs:   OPCi386.AbsVal (z, x);
                  vcap:   OPCi386.FNCap (z, x);
                  vodd:   OPCi386.FNOdd (z, x);
                  yadr, ycc, yval, yget8, yget16, yget32: OPCi386.SYSmop (z, x, n.subcl, n.typ);
                  ytype:
                          BEGIN
                            typ := n.left.typ;
                            IF typ.form = sfPointer THEN
                              typ := typ.BaseTyp;
                            ASSERT(typ.comp = csRecord);
                            OPCi386.GetTdAdr(typ, z);
                            z.typ := OPT.linttyp
                          END;
                END;
              END;
    Ndop:
              BEGIN
                Expression (n.left, x, selfForUnlock);
                Tf := x.typ.form;
                IF n.subcl = vand THEN
                  OPCi386.CondAnd (x)
                ELSE IF n.subcl = vor THEN
                  OPCi386.CondOr (x)
                ELSE IF (n.subcl >= veql) AND (n.subcl <= vgeq) THEN
                  OPCi386.Relation (x);
                Expression (n.right, y, selfForUnlock);
                CASE n.subcl OF
                  vtimes:        OPCi386.FNMul (z, x, y, Tf);
                  vslash, vdiv:  OPCi386.FNDiv (z, x, y, Tf);
                  vmod:          OPCi386.FNMod (z, x, y);
                  vand:          OPCi386.FNAnd (z, x, y);
                  vplus:         OPCi386.FNAdd (z, x, y, Tf);
                  vminus:        OPCi386.FNSub (z, x, y, Tf);
                  vor:           OPCi386.FNOr (z, x, y);
                  veql..vgeq:    OPCi386.FNCmp (z, x, y, n.subcl);
                  vin:           OPCi386.FNIn (z, x, y);
                  vash:          OPCi386.FNAsh (z, x, y);
                  vmsk:          OPCi386.Msk (z,  x, y);
                  vlen:          OPCi386.FNLen (z, x, y);
                  ybit, ylsh, yrot:OPCi386.SYSdop (z, x, y, n.subcl);
                END
              END;
    Ncall:    Call(n, z, FALSE);
  ELSE
    Designator (n, z)
  END;
  z.typ := n.typ;
END;

PROCEDURE Statement (n, selfForUnlock: OPT.Nodes);
VAR
  x, y, z, times: OPCi386.Item;
  t: OPT.Structs;
  p: OPT.Nodes;
  L, Lc, prevExitChain: OPCi386.BLabel;
  proc, par: OPT.Objects;
  name: STRING;
BEGIN
  WHILE NOT OPOi386.CodeErr AND (n <> NIL) DO BEGIN
    OPCi386.NewStat(n.conval^.intval);
    CASE n.clas OF
      Nenter    :
                  BEGIN
                    IF n.obj = NIL THEN BEGIN(* enter module *)
                      //OPM.LogWStrLn('Nenter NIL');
                      OPCi386.pc := 0;
                      OPCi386.FNEnter (NIL, 0, NIL);
                      Statement(n.right, NIL);
                      OPCi386.FNExit (NIL);
                      OPLi386.GenCode(OPCi386.pc);
                      OutRefPoint (RefStr,NIL);
                      OutRefName(RefStr,'$$');
                      OutRefs(RefStr,OPT.topScope);
                      INC(OPCi386.level);
                      Statement (n.left, NIL);
                      DEC(OPCi386.level);
                    END ELSE BEGIN(* procedure *)
                      //OPM.LogWStrLn('Nenter NOT NIL');
                      proc := n.obj;
                      par := proc.link;
                      INC(OPCi386.level);
                      Statement (n.left, NIL);
                      DEC(OPCi386.level);
                      OPCi386.pc := 0;
                      IF hasBody IN proc.conval^.setval THEN BEGIN
                        OPCi386.FNEnter (proc, proc.conval^.intval2, n);
                        IF proc.code<>NIL THEN BEGIN
                          Statement(proc.code, NIL);
                          OPCi386.AllocSDynArr(proc);
                        END;
                        Statement (n.right, NIL);
                        IF (proc.typ <> OPT.notyp) AND NOT (asmProc IN proc.conval^.setval)
                        AND (proc.mode <> omTyp) THEN
                          OPCi386.FNTrap (OPL.FuncTrap, n)
                        ELSE
                          OPCi386.FNExit (proc);
                        OutRefPoint (RefStr,proc);
                        IF proc.mode IN [omTProc, omTyp] THEN BEGIN
                          t := proc.link.typ;
                          IF (t.comp = csRecord) AND (t.ptr <> NIL) AND (t.ptr.strobj <> NIL) THEN
                            par := t.ptr.strobj
                          ELSE IF (t.comp = csRecord) OR (t.BaseTyp.strobj.name = '') THEN
                            par := t.strobj
                          ELSE
                            par := t.BaseTyp.strobj;
                          name := par.name+'.';
                          IF proc.mode = omTProc THEN
                            name := name+proc.name
                          ELSE IF proc.mode = omTyp THEN			(* Type body *)
                            name := name+'$$';
                        END ELSE
                          name:=proc.name;
                        OutRefName (RefStr,name);
                        OutRefs (RefStr,proc.scope.right);
                        OPLi386.GenCode (OPCi386.pc);
                      END;
                    END;

                  END;

      Ninittd   : ;//OPM.LogWStrLn('Ninittd');
      Nassign   :
                  BEGIN
                    //OPM.LogWStrLn('Nassign');
                    IF n.subcl = vassign THEN BEGIN  
                      IF (n.right.clas = Ncall) AND (n.right.typ.form = sfComp) THEN BEGIN
                        OPCi386.PushRegs;
                        Expression(n.left, z, selfForUnlock);	(*dest*)
                        Call(n.right, z, TRUE);
                      END ELSE BEGIN
                        Expression(n.right, x, selfForUnlock);
                        OPCi386.Relation(x);		(*load the Condition Code, if generated by expression*)
                        Expression(n.left, z, selfForUnlock);
                        OPCi386.FNAssign (z, x);
                      END
                    END ELSE IF (n.subcl = stifn) OR (n.subcl = clifn) THEN					(* New Functions *)
                      OPCi386.SYSinterrupt(n.subcl=clifn)
                    ELSE BEGIN
                      IF (n.subcl <> lockfn) THEN
                        Expression (n.left, z, selfForUnlock);
                      IF (n.subcl <> newfn) THEN
                        Expression (n.right, x, selfForUnlock);
                      IF n.subcl = movefn (* SYSTEM.Move *) THEN BEGIN
                        Expression (n.right.link, times, selfForUnlock);
                        OPCi386.SYSmove (z, x, times)
                      END ELSE BEGIN
                        CASE n.subcl OF
                          incfn, decfn:   OPCi386.IncDec (z, x, n.subcl = incfn);
                          inclfn, exclfn: OPCi386.Include (z, x, n.subcl = inclfn);
                          getfn, putfn:   OPCi386.SYSgetput (z, x, n.subcl = getfn);
                          getrfn, putrfn: OPCi386.SYSgetputReg (z, x, n.subcl = getrfn);
                          pinfn, poutfn:  OPCi386.SYSportio (z, x, n.subcl = pinfn);					(* New Functions *)
                          awaitfn:
                                          BEGIN
                                            Expression(n.right.link, y, selfForUnlock);
                                            OPCi386.Await(z, x, y);
                                          END;
                          lockfn:
                                          BEGIN
                                            Expression (n.right.link, z, NIL);
                                            OPCi386.Lock(x, z);
                                            Statement(n.left, n.right);
                                            Expression (n.right, x, NIL);
                                            Expression (n.right.link, z, NIL);
                                            OPCi386.Unlock(x, z);
                                          END;
                          copyfn:         OPCi386.FNCopy (z, x);
                          sysnewfn:       OPCi386.NewSys (z, x);
                          newfn:
                                          BEGIN
                                            t := n.left.typ;
                                            IF (n.right<>NIL) AND ((t.comp IN ArraySet)OR (t.BaseTyp.comp IN ArraySet)) THEN BEGIN
                                              IF t.form = sfPointer THEN BEGIN
                                                t := t.BaseTyp;
                                                IF t.comp=csDynArr THEN BEGIN
                                                  x := z;
                                                  OPCi386.NewRec (x, t);
                                                  OPCi386.DeRef(z)
                                                END;	(*allocate the descriptor*)
                                              END;
                                              Dim(z, x, n.right, t, 1)
                                            END ELSE BEGIN
                                              ASSERT(t.form = sfPointer);
                                              t := t.BaseTyp;
                                              OPCi386.NewRec (z, z.typ.BaseTyp);
                                              IF (t.comp<>csStaticArr) AND (t.strobj <> NIL) AND (t.strobj.conval <> NIL) THEN BEGIN
                                                IF n.left.link <> NIL THEN	BEGIN(* object parametrisation *)
                                                  n.left.link.conval := n.conval;
                                                  Statement(n.left.link, selfForUnlock);
                                                END;
                                                p := n.right;
                                                WHILE p <> NIL DO BEGIN
                                                  Expression(n.left, z, selfForUnlock);
                                                  Expression(p, x, selfForUnlock);
                                                  OPCi386.CallRecBody(z, x, p.obj.typ(*z.typ.BaseTyp*));
                                                  p := p.link
                                                END;
                                              END;
                                            END;
                                          END;
                        END (* CASE *)
                      END;

                    END;
                  END;

      Ncall     :
                  BEGIN
                    //OPM.LogWStrLn('Ncall');
                    Call(n, x, FALSE);
                  END;
      Nifelse   :
                  BEGIN
                    //OPM.LogWStrLn('Nifelse');
                    IF (n.subcl <> assertfn) OR OPV.bassert THEN BEGIN
                      L := OPCi386.aNil;
                      Lc := OPCi386.aNil;
                      IfStat (n.left, selfForUnlock, L, Lc, n.right <> NIL);
                      OPCi386.FixLink (Lc);
                      IF n.right <> NIL THEN Statement (n.right, selfForUnlock);
                      OPCi386.FixLink (L);
                    END;
                  END;
      Ncase     :
                  BEGIN
                    //OPM.LogWStrLn('Ncase');
                    CaseStat (n, selfForUnlock);
                  END;
      Nwhile    :
                  BEGIN
                    //OPM.LogWStrLn('Nwhile');
                    L := OPCi386.aNil;
                    OPCi386.FNJmp (L, n);
                    OPCi386.DefLabel (Lc);
                    Statement (n.right, selfForUnlock);
                    OPCi386.FixLink (L);
                    Expression (n.left, x, selfForUnlock);
                    OPCi386.FNJcc (x, Lc, n);
                  END;
      Nrepeat   :
                  BEGIN
                    //OPM.LogWStrLn('Nrepeat');
                    OPCi386.DefLabel (L);
                    Statement (n.left, selfForUnlock);
                    Expression (n.right, x, selfForUnlock);
                    OPCi386.Jncc (x, L, n)
                  END;
      Nloop     :
                  BEGIN
                    //OPM.LogWStrLn('Nloop');
                    prevExitChain := ExitChain;
                    ExitChain := OPCi386.aNil;
                    OPCi386.DefLabel (L);
                    Statement (n.left, selfForUnlock);
                    OPCi386.FNJmp (L, n);
                    OPCi386.FixLink (ExitChain);
                    ExitChain := prevExitChain;
                  END;
      Nexit     :
                  BEGIN
                    //OPM.LogWStrLn('Nexit');
                    OPCi386.FNJmp (ExitChain, n);
                  END;
      Nreturn   :
                  BEGIN
                    //OPM.LogWStrLn('Nreturn');
                    IF n.left <> NIL THEN BEGIN(* function *)
                      IF (n.left.clas = Ncall) AND (n.left.typ.form = sfComp) THEN BEGIN(*optimization*)
                        OPCi386.PushRegs;
                        dummy.mode := omVarPar;
                        dummy.mnolev := OPCi386.level;	(*fake mnolev: this is a temporary address on the stack, no allocated variable*)
                        dummy.adr:=n.obj.conval^.intval;		(*param size, needed to find the ret adr*)
                        dummy.typ:=n.left.typ;
                        Call(n.left, dummy, TRUE);
                        dummy.mode:=0
                      END ELSE BEGIN
                        Expression(n.left, x, NIL);
                        OPCi386.FNReturn(x, n.obj)
                      END
                    END;
                    IF selfForUnlock <> NIL THEN BEGIN
                      OPCi386.PushRegs;
                      Expression(selfForUnlock, x, NIL);
                      Expression (selfForUnlock.link, z, NIL);
                      OPCi386.Unlock(x, z);
                      OPCi386.PopResult(NIL, x);
                    END;
                    OPCi386.FNExit (n.obj);
                  END;
      Nwith     :
                  BEGIN
                    //OPM.LogWStrLn('Nwith');
                    L := OPCi386.aNil;
                    Lc := OPCi386.aNil;
                    IfStat (n.left, selfForUnlock, L, Lc, TRUE); (* L label for jump to END of if statement, Lc for the ELSE statement *)
                    OPCi386.FixLink (Lc);
                    IF n.subcl = 1 (*  # NIL *) THEN
                      Statement (n.right, selfForUnlock)
                    ELSE
                      OPCi386.FNTrap (OPL.WithTrap, n);
                    OPCi386.FixLink (L);
                  END;
      Ntrap     :
                  BEGIN
                    //OPM.LogWStrLn('Ntrap');
                    OPCi386.FNTrap(n.right.conval^.intval,n);
                  END;
      Nassembler:
                  BEGIN
                    //OPM.LogWStrLn('Nassembler');
                    OPCi386.GenAsm(n);
                  END;
    END;
    n:=n.link;
  END;
END;

PROCEDURE Module(prog: OPT.Nodes);
BEGIN
  dummy.mode:=0;
  Statement(prog, NIL);
  IF findpc AND OPM.noerr THEN OPM.err (254);
END;

END.
