UNIT OPA;

INTERFACE

USES
	_SYSTEM, OPM, OPS, OPT, OPB, OER, Classes;

CONST
  none = -1;
	
  (*processor targets*)
  i386 = 0;  i486 = 1;  Pentium = 2;  PentiumPro = 3; 
  FPU = 16;  Privileged = 17;
		
  (* node classes for inline instructions*)
  Nassembler = 29;
  Ncode = 30;
  Nfixup = 31;
		
  (* Own Symbol Table types *)
  StLabel = 0; StFwdLabel = 1;
  StConst = 3; StType = 4; StVar = 5; StVarPar = 6; StGlobal = 7;
  StMod = 11;

  (* Own Addressing modes, also used in the St and Sc *)
  Reg8 = $08; Reg16 = $10; Reg32 = $20;
  RegAL = $09; RegAX = $11; RegEAX = $21;
  RegDX = $12;								(* special case for IN/OUT *)
  RegCL = $13;								(* special case for rotate/shift *)
  Imm8 = $30; Imm16 = $31; Imm32 = $32; SImm8 = $35; Imm = $36;
  Const1 = $33;								(* special case for rotate/shift *)
  Const3 = $34;								(* special case for INT 3 *)
  RM = $40; RM8 = $41; RM16 = $42; RM32 = $43;
  M0 = $48; M8 = $49; M16 = $4A; M32 = $4B; M64 = $4C; M80 = $4D;
  (*M2 = 4EH;*)
  SReg = $50;	(* Segment Register *)
  RegCS = $51; RegDS = $52; RegES = $53; RegFS = $54;
  RegGS = $55; RegSS = $56;
  Rel8 = $60; Rel16 = $61; Rel32 = $62; FwdJmp = $63;
  CRReg = $70; DRReg = $71; TRReg = $72;
  FReg = $80; FRegST = $81;
		
  (* Scanner: ASCII codes *)
  TAB   = #$09;
  CR    = #$0D;
  LF    = #$0A;
  SPACE = #$20;

  (* Structure sizes *)
  NameLen = 32;
  MaxStrLen = 256;
  MnemoLen = 12;
		
  (* Opcodes modifiers *)
  OmRM = 1;	OmOp = 2;		(* register stored in the RM byte / register added to the opcode *)
		
  (* Match evaluation *)
  NoMatch = 0; SizeCast = 1; TypeCast = 2; ConstCast = 3; Hit = 4;

TYPE 
  Name = STRING;

  (*Object: The symbol table. It's an ordered binary tree. First element is root.right *)
  AObjects = ^ObjDesc;
  FixPnt = ^FixPntDesc;
		
  InstrDesc = RECORD
        name: STRING;
        start, vend: INTEGER;	 (* index in the opcode-tab *)
        target: TSET;
      END;

  (* Opcode: this structure contains the definition of an opcode, with the
    expected dst/src operands and the information needed to encode it.
    Ex:
      if opc # -1 => modr/m . opc = opc. ( extension of the opcode )
  *)
  OpCode = RECORD
      op: ARRAY [0..3] OF INTEGER;		(* mode of the dst/src operands, if needed *)
      op0, op1: INTEGER;	   (* opcode *)
      opc: INTEGER;			(* value stored in reg/opcode of modr/m. -1 = none *)
      reg: INTEGER;			(* where is the register stored: none / rm /op *)
    END;
		
  Operand = RECORD
        mode: INTEGER;		(* the mode: Reg8.... *)
        imm: LONGINT;		(* the constant *)
        disp: LONGINT;
        index, base, reg: INTEGER;
        seg, scale: INTEGER;	(* segment override *)
        obj: AObjects;			(* the jmp label for forward jumps / the object referenced *)
      END;
				
  ObjDesc = RECORD
            typ: INTEGER;
            val, size:  LONGINT;		(* if typ <0 => pos of first use of this label *)
            errpos   : TErrorPos;
            name: Name;
            left, right, next: AObjects;
            fix: FixPnt;
            obj: OPT.Objects
         END;
		
  FixPntDesc = RECORD
          pc: LONGINT;
          node: OPT.Nodes;
          pos: LONGINT;
          next: FixPnt;
        END;

VAR
(* Global structures *)
  root:  AObjects;	(* symbol table *)
  OpTab: ARRAY OF OpCode;	(* Table with the opcodes *)
  OpTabLen: INTEGER;		(* nof opcodes in the table *)
  InstrTab: ARRAY OF InstrDesc;	(* Table with the instructions !!Hash *)
  InstrTabLen: INTEGER;
  SizeTab: ARRAY [0..16] OF SHORTINT;		(* size of the oberon types *)
  TargetTab: ARRAY [0..32] OF RECORD  name: Name;  flag: _SYSTEM.TSET  END;
  TargetCount: LONGINT;
		
(* Configuration *)
  export, Cinline: BOOLEAN;		(* code is exported / inlined *)
  FirstTime: BOOLEAN;		(* First time the module is used *)
  ATarget: SET OF 0..31;						(* The target processor*)
  pc: LONGINT;	(* program counter. Should be used only for pc-rel jumps !! *)
		
(* Scanner variables *)
  (*ch: CHAR;*)			(* look-ahead char is OPS.ch*)
  sym: TToken; 	(* last token read *)
  val: LONGINT;	  (* last value read *)
  OPAstr: STRING;

(* Instruction nodes *)
  first, last: OPT.Nodes;	(* first and last nodes in the instruction stream *)
  fixlist: OPT.Nodes;	(* list of fixups to be done by the back-end *)

  PROCEDURE Assemble(VAR Scaner: TScaner; VAR x: OPT.Nodes;  exported, inlined: BOOLEAN);
  PROCEDURE OPABegin;
  PROCEDURE HashFn(CONST name: STRING; VAR k, n: INTEGER);
  PROCEDURE FindInstruction(VAR name: STRING; VAR inx: INTEGER; Flag: BOOLEAN);

IMPLEMENTATION

USES
  OPAData;

(* ------------------------------------------------------------------- *)
(* Special functions *)

FUNCTION FNAND(a,b: LONGINT):LONGINT;
BEGIN
  RESULT:=a AND b;
END;

(* ------------------------------------------------------------------- *)
(* Addressing Mode functions *)
	
FUNCTION ModeSize(mode: INTEGER): INTEGER;
BEGIN
  CASE mode OF
    Imm8, Const1, Const3, Rel8, RegAL, RegCL, Reg8, RM8, M8, SImm8: RESULT:=1;
    Imm16, Rel16, RegAX, RegDX, Reg16, RM16, M16,
    SReg, RegCS, RegDS, RegES, RegFS, RegGS, RegSS:                 RESULT:=2;
    Imm32, Rel32, FwdJmp, RegEAX, Reg32, RM32, M32:                 RESULT:=4;
    M64:                                                            RESULT:=8;
    M80:                                                            RESULT:=10;
    RM:                                                             RESULT:=0;
  ELSE
    RESULT:=0;
  END
END;
	
FUNCTION ConstSize(i: LONGINT; signed: BOOLEAN):INTEGER;
BEGIN
  IF (MIN(_SHORTINTS) <= i) AND (MAX(_SHORTINTS) >= i) OR (NOT signed AND (FNAND(i,$0FFFFFF00)=0)) THEN
    RESULT:=1
  ELSE IF (MIN(_INTEGERS) <= i) AND (MAX(_INTEGERS) >= i) OR (NOT signed AND (FNAND(i,$0FFFF0000)=0))THEN
    RESULT:=2
  ELSE
    RESULT:=4;
END;
	
FUNCTION IsRegMode(m: INTEGER; strict: BOOLEAN):BOOLEAN;
BEGIN
  RESULT:=((m=Reg8)OR(m=Reg16)OR(m=Reg32)OR
    (NOT strict AND((m=RegAL)OR(m=RegAX)OR(m=RegEAX)OR(m=RegDX))))
END;
	
FUNCTION IsSRegMode(m: INTEGER; strict: BOOLEAN):BOOLEAN;
BEGIN
  RESULT:=((m=SReg)OR (NOT strict AND(m=RegCS)OR(m=RegDS)OR(m=RegES)OR
    (m=RegFS)OR(m=RegGS)OR(m=RegSS)))
END;
	
FUNCTION IsSpecRegMode(m: INTEGER):BOOLEAN;
BEGIN
  RESULT:=((m=CRReg)OR(m=DRReg)OR(m=TRReg))
END;
	
FUNCTION IsMMode(m: INTEGER):BOOLEAN;
BEGIN
  RESULT:=((m=M0)OR(m=M8)OR(m=M16)OR(m=M32)OR(m=M64)OR(m=M80))
END;
	
FUNCTION IsModRmMode(m: INTEGER):BOOLEAN;
BEGIN
  RESULT:=((m=RM)OR(m=RM8)OR(m=RM16)OR(m=RM32) (*OR IsMMode(m)*))
END;
	
FUNCTION IsFRegMode(m: INTEGER; strict: BOOLEAN):BOOLEAN;
BEGIN
  RESULT:=((m=FReg) OR (NOT strict AND (m=FRegST)))
END;

FUNCTION NeedModRM(mode, regpos: INTEGER): BOOLEAN;
BEGIN
  RESULT:=(IsModRmMode(mode) OR IsMMode(mode) OR
          ( (regpos = OmRM) AND (IsRegMode(mode,TRUE)
            OR IsSRegMode(mode, FALSE) OR IsSpecRegMode(mode))))
END;
	
FUNCTION IsImmMode(m: INTEGER):BOOLEAN;
BEGIN
  RESULT:=((m=Imm8)OR(m=Imm16)OR(m=Imm32)OR(m=SImm8)OR(m=Imm))
END;
	
FUNCTION IsRelMode(m: INTEGER):BOOLEAN;
BEGIN
  RESULT:=((m=Rel8)OR(m=Rel16)OR(m=Rel32))
END;
	
FUNCTION Evaluate(VAR op: Operand; mode: INTEGER):INTEGER;
BEGIN
  IF mode = op.mode THEN BEGIN
    RESULT:=(Hit);	(* 80% of the checks, now the 20% special cases *)
    EXIT;
  END ELSE IF IsModRmMode(mode) THEN BEGIN
    IF IsMMode(op.mode) THEN BEGIN
      IF ModeSize(op.mode)=ModeSize(mode) THEN BEGIN
        RESULT:= Hit;
        EXIT;
      END ELSE IF (op.mode = RM) OR (op.mode = M0) THEN BEGIN
        RESULT:= SizeCast;
        EXIT;
      END ELSE BEGIN
        RESULT:= NoMatch;
        EXIT;
      END;
    END ELSE IF IsRegMode(op.mode, FALSE) AND (ModeSize(op.mode)=ModeSize(mode)) THEN BEGIN
      RESULT:= TypeCast;
      EXIT;
    END ELSE BEGIN
      RESULT:= NoMatch;
      EXIT;
    END;
  END ELSE IF IsRegMode(mode,TRUE) THEN	BEGIN(* cannot be a specific reg (AL,AX,EAX) *)
    IF ModeSize(op.mode)=ModeSize(mode) THEN BEGIN
      IF IsRegMode(op.mode,TRUE) THEN BEGIN
        RESULT:= Hit;
        EXIT;
      END ELSE IF IsRegMode(op.mode, FALSE) THEN BEGIN
        RESULT:= TypeCast;
        EXIT;
      END ELSE BEGIN
        RESULT:= NoMatch;
        EXIT;
      END;
    END ELSE BEGIN
      RESULT:= NoMatch;
      EXIT;
    END;
  END ELSE IF mode = RegDX THEN BEGIN
    IF (op.mode = Reg16) AND (op.reg = 2) THEN BEGIN
      RESULT:= Hit;
      EXIT;
    END ELSE BEGIN
      RESULT:= NoMatch;
      EXIT;
    END;	(* special case for IN/OUT which uses DX *)
  END ELSE IF mode = RegCL THEN BEGIN
    IF (op.mode = Reg8) AND (op.reg = 1) THEN BEGIN
      RESULT:= Hit;
      EXIT;
    END ELSE BEGIN
      RESULT:= NoMatch;
      EXIT;
    END	(* special case for shift/rotate which uses CL *)
  END ELSE IF IsRegMode(mode, FALSE) THEN BEGIN
    RESULT:= NoMatch;
    EXIT;
  END ELSE IF IsImmMode(mode) THEN BEGIN
    IF op.mode = Imm THEN BEGIN
      IF ModeSize(mode)  > ConstSize(op.imm, mode = SImm8) THEN BEGIN
        RESULT:= ConstCast;
        EXIT;
      END ELSE IF ModeSize(mode) = ConstSize(op.imm, mode = SImm8) THEN BEGIN
        RESULT:= Hit;
        EXIT;
      END ELSE BEGIN
        RESULT:= NoMatch;
        EXIT;
      END;
    END ELSE IF IsImmMode(op.mode) THEN BEGIN
      IF ModeSize(mode) > ModeSize(op.mode) THEN BEGIN
        RESULT:= ConstCast;
        EXIT;
      END ELSE IF ModeSize(mode) = ModeSize(op.mode) THEN BEGIN
        RESULT:= Hit;
        EXIT;
      END ELSE BEGIN
        RESULT:= NoMatch;
        EXIT;
      END
    END ELSE BEGIN
      RESULT:= NoMatch;
      EXIT;
    END;
  END ELSE IF mode = Const1 THEN BEGIN
    IF IsImmMode(op.mode) AND (op.imm = 1) THEN BEGIN
      RESULT:= Hit;
      EXIT;
    END ELSE BEGIN
      RESULT:= NoMatch;
      EXIT;
    END
  END ELSE IF mode = Const3 THEN BEGIN
    IF IsImmMode(op.mode) AND (op.imm = 3) THEN BEGIN
      RESULT:= Hit;
      EXIT;
    END ELSE BEGIN
      RESULT:= NoMatch;
      EXIT;
    END
  END ELSE IF IsMMode(mode) THEN BEGIN
    IF IsMMode(op.mode) THEN BEGIN
      IF ModeSize(op.mode)=ModeSize(mode) THEN BEGIN
        RESULT:= Hit;
        EXIT;
      END ELSE IF (op.mode = M0) OR (mode = M0) THEN BEGIN
        RESULT:= SizeCast;
        EXIT;
      END ELSE BEGIN
        RESULT:= NoMatch;
        EXIT;
      END
    END ELSE BEGIN
      RESULT:= NoMatch;
      EXIT;
    END
  END ELSE IF mode = SReg THEN BEGIN
    IF IsSRegMode(op.mode,FALSE) THEN BEGIN
      RESULT:= Hit;
      EXIT;
    END ELSE BEGIN
      RESULT:= NoMatch;
      EXIT;
    END;
  END ELSE IF IsRelMode(mode) THEN BEGIN
    IF (mode = Rel32) AND (op.mode = FwdJmp) THEN BEGIN
      RESULT:= Hit;
      EXIT;
    END ELSE IF IsImmMode(op.mode) THEN BEGIN
      IF ModeSize(mode) = ConstSize(op.imm, TRUE) (*ModeSize(op.mode)*) THEN BEGIN
        RESULT:= Hit;
        EXIT;
      END ELSE IF ModeSize(mode) > ConstSize(op.imm, TRUE) (*ModeSize(op.mode)*) THEN BEGIN
        RESULT:= ConstCast;
        EXIT;
      END ELSE BEGIN
        RESULT:= NoMatch;
        EXIT;
      END
    END ELSE BEGIN
      RESULT:= NoMatch;
      EXIT;
    END;
  END ELSE IF mode = FReg THEN BEGIN
    IF IsFRegMode(op.mode, TRUE) THEN BEGIN
      RESULT:= Hit;
      EXIT;
    END ELSE IF IsFRegMode(op.mode, FALSE) THEN BEGIN
      RESULT:= TypeCast;
      EXIT;
    END ELSE BEGIN
      RESULT:= NoMatch;
      EXIT;
    END
  END ELSE BEGIN
    RESULT:= NoMatch;
    EXIT;
  END;
END;
	
	
PROCEDURE Match(VAR ind: INTEGER; VAR op: ARRAY OF Operand);
VAR
  start, vend, i, j, k, best: INTEGER;
BEGIN
  start := InstrTab[ind].start;
  vend := InstrTab[ind].vend;
  ind := -1;
  best := 0;
  WHILE start < vend DO BEGIN
    i := Evaluate(op[0], OpTab[start].op[0]);
    IF (i = NoMatch) THEN
    ELSE BEGIN
      j := Evaluate(op[1], OpTab[start].op[1]);
      IF j = NoMatch  THEN
        i := NoMatch
      ELSE BEGIN
        k := Evaluate(op[2], OpTab[start].op[2]);
        IF (i < j) AND (i < k) THEN
        ELSE IF j < k THEN i :=  j
        ELSE i := k;
      END
    END;
    IF i <> NoMatch THEN BEGIN
      IF i = Hit THEN BEGIN
        ind := start;
        EXIT;
      END ELSE IF i > best THEN BEGIN
        ind := start;
        best := i
      END ELSE IF (i = best) AND (i = SizeCast) THEN BEGIN
        OPM.Mark(512, OPM.errpos);
        ind := -1;
        EXIT;
      END
    END;
    INC(start);
  END;
  IF ind = -1 THEN OPM.Mark(501, OPM.errpos);
END; 	

(* ------------------------------------------------------------------- *)
(* Symbol Table / Instr Table / Op Table  functions *)

PROCEDURE insert(VAR name: STRING; VAR obj: AObjects);
VAR
  ob0, ob1: AObjects;
  d: INTEGER;
BEGIN
  ob0 := root;
  ob1 := ob0.right;
  d := 1;
  REPEAT
    IF ob1 = NIL THEN BEGIN(*insert undefined label*)
      NEW(ob1);
      ob1.name:=name;
      ob1.typ := StFwdLabel;
      ob1.left := NIL;
      ob1.right := NIL;
      ob1.fix := NIL;
      ob1.obj := NIL;
      ob1.next := root.next;
      root.next := ob1;
      ob1.errpos := OPM.errpos;
      IF d < 0 THEN ob0.left := ob1 ELSE ob0.right := ob1;
    END ELSE IF name < ob1.name THEN BEGIN
      d := -1;
      ob0 := ob1;
      ob1 := ob1.left
    END ELSE IF name > ob1.name THEN BEGIN
      d :=  1;
      ob0 := ob1;
      ob1 := ob1.right;
    END ELSE (*found, no insertion*)
      BREAK;
  UNTIL FALSE;
  obj := ob1
END;

FUNCTION ConvertObj(ob: OPT.Objects): AObjects;
VAR
  obj: AObjects;
BEGIN
  obj:=NIL;
  IF ob <> NIL THEN BEGIN
    NEW(obj);
    CASE ob.mode OF
      1: (* Var *)
            BEGIN
              obj.val := 1;
              obj.size := SizeTab[ob.typ.form];
              IF ob.mnolev <= 0 THEN
                obj.typ := StGlobal
              ELSE
                obj.typ := StVar
            END;
      2: (* VarPar *)
            BEGIN
              obj.typ := StVarPar;
              obj.val := 1;
              obj.size := 4;
            END;
      3: (* Con *)
            BEGIN
              obj.typ := StConst;
              obj.val := ob.conval^.intval;
              IF (ob.typ.form > 6) OR (ob.typ.form = 2) THEN OPM.err(51);
            END;
      5: (* Type *)
            BEGIN
              obj.typ := StType;
              obj.size := SizeTab[ob.typ.form];
              OPM.err(514);
            END;
      11: (* Mod *) obj.typ := StMod;
    ELSE
      obj.typ := StConst;
      obj.val := 0;
      OPM.err(514);
    END;
    obj.obj := ob;
  END;
  RESULT:= obj
END;

PROCEDURE find(VAR Scaner: TScaner; VAR name: STRING; VAR obj: AObjects);
VAR
  ob0: AObjects;
  ob : OPT.Objects;
BEGIN
  ob0 := root.right;

  WHILE (ob0 <> NIL) AND (ob0.name <> name) DO BEGIN
    IF name < ob0.name THEN
      ob0 := ob0.left
    ELSE
      ob0 := ob0.right;
  END;

  obj := ob0;
  IF obj = NIL THEN BEGIN
    Scaner.name:=name;
    OPT.Find(Scaner.name,ob);
    IF ob <> NIL THEN
      obj := ConvertObj(ob)
    ELSE
      insert(name, obj);		(* create forward jump *)
  END;
END;

PROCEDURE HashFn(CONST name: STRING; VAR k, n: INTEGER);
VAR
  i: INTEGER;
BEGIN
  IF n = 0 THEN BEGIN
    i := 1;
    k := 0;
    WHILE i <LENGTH(name) DO BEGIN
      k := (k*16 + ORD(name[i])) MOD InstrTabLen;
      INC(i);
    END;
    n := 1
  END ELSE BEGIN
    k := (k + n) MOD InstrTabLen;
    INC(n, 1);
  END
END;

PROCEDURE FindInstruction(VAR name: STRING; VAR inx: INTEGER; Flag: BOOLEAN);
VAR
  n: INTEGER;
BEGIN
  n := 0;
  REPEAT
    HashFn(name, inx, n);
  UNTIL (name = InstrTab[inx].name) OR (InstrTab[inx].name = '') OR (n >InstrTabLen);
  IF (InstrTab[inx].name = '') OR (n >InstrTabLen) THEN
    inx := -1
  ELSE IF (InstrTab[inx].target+ATarget <> ATarget) AND Flag THEN
    OPM.err(515);
END;
	
(* ------------------------------------------------------------------- *)
(* Scanner functions *)

PROCEDURE skipBlanks(VAR Scaner: TScaner);
BEGIN
  WHILE (Scaner.ch = SPACE) OR (Scaner.ch = TAB) DO Scaner.GetChar(Scaner.ch);
  IF Scaner.ch = ';' THEN BEGIN
    WHILE (Scaner.ch <> CR) AND (Scaner.ch <> LF)
    AND (Scaner.ch <> OPM.Eot) DO Scaner.GetChar(Scaner.ch);
  END;
END;

PROCEDURE GetIdent(VAR Scaner: TScaner);
BEGIN
  //OPM.errpos := OPM.curpos;
  Scaner.name:='';
  REPEAT
    IF (NameLen>1) THEN Scaner.name := Scaner.name+Scaner.ch;
    Scaner.GetChar(Scaner.ch);
  UNTIL NOT (('A' <= CAP(Scaner.ch)) AND (CAP(Scaner.ch) <= 'Z') OR ('0' <= Scaner.ch) AND (Scaner.ch <= '9'));
END;

PROCEDURE Get(VAR Scaner: TScaner; VAR sym: TToken);
	
  PROCEDURE Comment;
  BEGIN
    Scaner.GetChar(Scaner.ch);	(* skip '*' *)
    REPEAT
      REPEAT
        Scaner.GetChar(Scaner.ch);
        IF Scaner.ch = '(' THEN BEGIN
          Scaner.GetChar(Scaner.ch);
          IF Scaner.ch = '*' THEN Comment;
        END;
      UNTIL (Scaner.ch = '*') OR (Scaner.ch = OPM.Eot);
      Scaner.GetChar(Scaner.ch);
    UNTIL (Scaner.ch = ')') OR (Scaner.ch = OPM.Eot);
    Scaner.GetChar(Scaner.ch);
  END;
		
  PROCEDURE Str(VAR Scaner: TScaner);
  VAR
    och: CHAR;
    i: LONGINT;
  BEGIN
    och := Scaner.ch;
    i := 1;
    REPEAT
      Scaner.GetChar(Scaner.ch);
      IF Scaner.ch = och THEN
        BREAK
      ELSE IF Scaner.ch < ' ' THEN BEGIN
        OPM.err(3);
        BREAK;
      END ELSE IF i = MaxStrLen-1 THEN BEGIN
        OPM.err(241);
        BREAK;
      END;
      OPAstr[i] := Scaner.ch;
      INC(i)
    UNTIL FALSE;
    Scaner.GetChar(Scaner.ch);
    OPAstr[i] := #0;
    val := i;
  END;
		
BEGIN
  skipBlanks(Scaner);
  //OPM.errpos := OPM.curpos;
  CASE Scaner.ch OF
    'A' .. 'Z',
    'a' .. 'z' :
          BEGIN
            GetIdent(Scaner);
            IF Scaner.ch = ':' THEN BEGIN
              Scaner.GetChar(Scaner.ch);
              sym := ScLabel;
            END ELSE IF Scaner.name = 'END' THEN
              sym := ScEnd
            ELSE
              sym := ScIdent
          END;
    '0'..'9':
          BEGIN
            val := 0;
            sym := ScNumber;
            Scaner.Number;
            ASSERT(Scaner.numtyp <= _real);
            val := Scaner.intval
          END;
    ';':  BEGIN Scaner.GetChar(Scaner.ch); sym := ScSemicolon; END;
    '@':  BEGIN Scaner.GetChar(Scaner.ch); sym := ScAt;        END;
    ',':  BEGIN Scaner.GetChar(Scaner.ch); sym := ScComma;     END;
    '[':  BEGIN Scaner.GetChar(Scaner.ch); sym := ScLBrak;     END;
    ']':  BEGIN Scaner.GetChar(Scaner.ch); sym := ScRBrak;     END;
    '{':  BEGIN Scaner.GetChar(Scaner.ch); sym := ScLBrace;    END;
    '}':  BEGIN Scaner.GetChar(Scaner.ch); sym := ScRBrace;    END;
 CR,LF :	BEGIN Scaner.GetChar(Scaner.ch); sym := ScCR;        END;
    '*':	BEGIN Scaner.GetChar(Scaner.ch); sym := ScMult;      END;
    '/':	BEGIN Scaner.GetChar(Scaner.ch); sym := ScDiv;       END;
    '+':	BEGIN Scaner.GetChar(Scaner.ch); sym := ScPlus;      END;
    '-':	BEGIN Scaner.GetChar(Scaner.ch); sym := ScMinus;     END;
    '(':
          BEGIN
            Scaner.GetChar(Scaner.ch);
            IF Scaner.ch = '*' THEN BEGIN
              Comment;
              Get(Scaner,sym);
            END ELSE
              sym := ScLPar
          END;
    ')':  BEGIN Scaner.GetChar(Scaner.ch); sym := ScRPar;      END;
    ':':	BEGIN Scaner.GetChar(Scaner.ch); sym := ScColon;     END;
    '.':	BEGIN Scaner.GetChar(Scaner.ch); sym := ScPoint;     END;
    #22,
    #27:  BEGIN Str(Scaner);     sym := ScString;    END;
  ELSE
    sym := ScUndef;
    Scaner.GetChar(Scaner.ch);
  END;
END;
	
PROCEDURE Check(VAR Scaner: TScaner; s: TToken);
BEGIN
  IF sym <> s THEN OPM.err(ORD(s));
  Get(Scaner,sym)
END;

PROCEDURE Qualident(VAR Scaner: TScaner; VAR obj: AObjects);
VAR
  i: LONGINT;
  o: OPT.Objects;
BEGIN
  obj := NIL;
  IF sym = ScIdent THEN BEGIN
    IF Scaner.name = 'SYSTEM' THEN BEGIN
      IF NOT OPT.SYSimported THEN OPM.err(135);
      Scaner.name := 'SYSTEM';
      OPT.Find(Scaner.name,o);
      IF o <> NIL THEN
        INCL(o.flag, OPT.used);
      Get(Scaner,sym);
      IF sym=ScPoint THEN BEGIN
        Get(Scaner,sym);
        IF sym = ScIdent THEN BEGIN
          i := 0;
          WHILE (i < TargetCount) AND (TargetTab[i].name <> Scaner.name) DO INC(i);
          IF i = TargetCount THEN
            OPM.err(0)
          ELSE BEGIN
            NEW(obj);
            obj.typ := StConst;
            obj.val := i;
          END;
        END ELSE
          OPM.err(0);
        Get(Scaner,sym)
      END ELSE
        OPM.err(18);
    END ELSE BEGIN
      Get(Scaner,sym);
      find(Scaner,Scaner.name, obj);
      IF (obj<>NIL) AND (sym=ScPoint) AND (obj.typ=StMod) THEN BEGIN
        Get(Scaner,sym);
        OPT.FindImport(Scaner.name,obj.obj, obj.obj);
        obj := ConvertObj(obj.obj);
      END;
      IF obj=NIL THEN OPM.err(0);
    END;
  END ELSE
    OPM.err(40)
END;
	
PROCEDURE Expression(VAR Scaner: TScaner; VAR x: LONGINT);
VAR
  y: LONGINT;
  op : TToken;

  PROCEDURE Factor(VAR x: LONGINT);
  BEGIN
    IF sym = ScNumber THEN BEGIN
      x := val;
      Get(Scaner,sym);
    END ELSE IF sym = ScLPar THEN BEGIN
      Get(Scaner,sym);
      Expression(Scaner,x);
      Check(Scaner,ScRPar);
    END ELSE
      OPM.err(601);
  END;
	
  PROCEDURE Term(VAR x: LONGINT);
  VAR
    y: LONGINT;
    op : TToken;
  BEGIN
    Factor(x);
    WHILE (sym = ScMult) OR (sym = ScDiv) DO BEGIN
      op := sym;
      Get(Scaner,sym);
      Factor(y);
      IF op = ScMult THEN
        x := x * y
      ELSE
        x := x DIV y;
    END;
  END;
	
BEGIN
  IF sym = ScMinus THEN BEGIN
    Get(Scaner,sym);
    Term(x);
    x := -x
  END ELSE
    Term(x);
  WHILE (sym = ScPlus) OR (sym = ScMinus) DO BEGIN
    op := sym;
    Get(Scaner,sym);
    Term(y);
    IF op = ScPlus THEN
      x := x + y
    ELSE
      x := x - y;
  END;
END;

PROCEDURE GetBaseIndex(VAR Scaner: TScaner; VAR o: Operand; size: LONGINT);
VAR
  obj: AObjects;
BEGIN
  CASE size OF
      0 : o.mode := M0;
      1 : o.mode := M8;
      2 : o.mode := M16;
      4 : o.mode := M32;
      8 : o.mode := M64;
      10: o.mode := M80;
  END;
  o.disp := o.imm;
  o.imm := 0;		(* this is not a const but a rm !! *)
  WHILE (sym = ScLBrak) OR (sym = ScPlus) DO BEGIN
    Get(Scaner,sym);
    IF sym = ScIdent THEN find(Scaner,Scaner.name, obj) ELSE obj := NIL;
    IF (obj = NIL) OR NOT IsRegMode(obj.typ, FALSE) OR (ModeSize(obj.typ) <> 4) THEN BEGIN
      OPM.err(506);
      EXIT;
    END;
    Get(Scaner,sym);
    IF sym = ScMult THEN BEGIN		(* This is the index, get the scale *)
      IF o.index <> none THEN BEGIN
        OPM.err(509);
        EXIT;
      END;
      o.index := obj.val;
      Get(Scaner,sym);
      IF (sym <> ScNumber) OR ((val <> 1) AND (val <> 2) AND (val <> 4) AND (val <> 8)) THEN BEGIN
        OPM.err(508);
        EXIT;
      END;
      o.scale := val;
      Get(Scaner,sym);
    END ELSE IF (sym = ScRBrak) OR (sym = ScPlus) THEN BEGIN
      IF o.base = none THEN
        o.base := obj.val
      ELSE IF o.index = none THEN
        o.index := obj.val
      ELSE
        OPM.err(509);
    END ELSE BEGIN
      OPM.err(510);
      EXIT;
    END;
    IF sym = ScRBrak THEN Get(Scaner,sym);
  END; (* WHILE ScLBrak *)
END;

PROCEDURE GetOperand(VAR Scaner: TScaner; VAR o: Operand);
VAR
  obj: AObjects;
  size: LONGINT;
BEGIN
  o.reg := none;
  o.disp := 0;
  o.base := none;
  o.index := none;
  o.imm := 0;
  o.seg := none;
  o.scale := none;
  o.obj := NIL;
  size := 0;	(* unknown size *)
  IF sym = ScIdent THEN BEGIN
    find(Scaner,Scaner.name, obj);
    IF (obj <> NIL) AND (obj.typ = SReg) THEN BEGIN
      o.seg := obj.val;
      Get(Scaner,sym);
      Check(Scaner,ScColon);
      IF sym = ScIdent THEN find(Scaner,Scaner.name, obj);
    END;
  END;
  IF (sym = ScIdent) AND (obj <> NIL) AND (obj.typ = StType) THEN BEGIN
    size := obj.size;
    Get(Scaner,sym);
    IF (sym <> ScIdent) OR (Scaner.name <> 'PTR') THEN (* OPM.err(511) *)	(* PTR is optional *)
    ELSE
      Get(Scaner,sym);
    IF sym = ScIdent THEN find(Scaner,Scaner.name, obj);
  END;
  CASE sym OF
    ScAt:
              BEGIN
                Get(Scaner,sym);
                IF sym <> ScNumber THEN OPM.err(-601);
                o.disp := val;
                o.mode := RM32;
                Get(Scaner,sym);
              END;
    ScLBrak:  GetBaseIndex(Scaner,o, size);
    ScIdent, ScNumber, ScMinus:
              BEGIN
                IF sym = ScIdent THEN BEGIN
                  IF obj = NIL THEN BEGIN
                    OPM.err(500);
                    EXIT;
                  END;
					
                  IF size = 0 THEN size := obj.size;		(* don't override the xyz PTR command *)
					
                  IF IsRegMode(obj.typ, FALSE) THEN BEGIN
                    o.mode := obj.typ;
                    o.reg := obj.val;
                    Get(Scaner,sym);
                    EXIT;
                  END ELSE IF IsSRegMode(obj.typ, FALSE) THEN BEGIN
                    o.mode := obj.typ;
                    o.reg := obj.val;
                    Get(Scaner,sym);
                    EXIT;
                  END ELSE IF IsSpecRegMode(obj.typ) THEN BEGIN
                    IF NOT (Privileged IN ATarget) THEN OPM.err(515);
                    IF NOT (Pentium IN ATarget) AND (obj.typ = CRReg) AND (obj.val = 4) THEN OPM.err(515);
                    o.mode := obj.typ;
                    o.reg := obj.val;
                    Get(Scaner,sym);
                    EXIT;
                  END ELSE IF IsFRegMode(obj.typ, FALSE) THEN BEGIN
                    o.mode := obj.typ;
                    Get(Scaner,sym);
                    IF sym = ScLPar THEN BEGIN
                      Expression(Scaner,size);
                      o.reg := size;
                    END ELSE BEGIN
                      o.reg := 0;
                      o.mode := FRegST
                    END;
                    EXIT;
                  END ELSE IF obj.typ = StLabel THEN BEGIN
                    o.imm := obj.val;
                    o.mode := Rel8;
                    Get(Scaner,sym);
                    EXIT;
                  END ELSE IF obj.typ = StFwdLabel THEN BEGIN
                    o.obj := obj;
                    o.mode := FwdJmp;
                    Get(Scaner,sym);
                    EXIT;
                  END ELSE IF obj.typ = StConst THEN BEGIN
                    o.imm := obj.val;
                    Get(Scaner,sym);
                  END ELSE IF (obj.typ = StVar) OR (obj.typ = StVarPar) THEN BEGIN
                    IF Cinline THEN  OPM.err(518);
                    o.imm := obj.val;
                    o.obj := obj;
                    Get(Scaner,sym);
                  END ELSE IF (obj.typ = StGlobal) THEN BEGIN
                    o.disp := $012345678;
                    o.obj := obj;
                    CASE size OF
                      1 : o.mode := M8;
                      2 : o.mode := M16;
                      4 : o.mode := M32;
                      8 : o.mode := M64;
                      10: o.mode := M80;
                    ELSE
                      o.mode := M32;
                    END;
                    Get(Scaner,sym);
                    EXIT;		(* Global Var *)
                  END ELSE BEGIN
                    OPM.err(600);
                    EXIT;
                  END;
                END ELSE
                  Expression(Scaner,o.imm);

                CASE size OF
                  1: o.mode := Imm8;
                  2: o.mode := Imm16;
                  4: o.mode := Imm32;
                ELSE
                  o.mode := Imm;
                END;

                IF sym = ScLBrak THEN GetBaseIndex(Scaner,o, size);
              END;
  ELSE
    o.mode := none;
    OPM.err(511)
  END;  (* CASE sym OF *)
END;
	
(* ------------------------------------------------------------------- *)
(* Fixpoint handling functions *)

PROCEDURE CheckLabels;
VAR
  obj, obj1, obj2: AObjects;
BEGIN
  obj1 := root.next;
  obj := NIL; (* invert list *)
  WHILE obj1 <> NIL DO BEGIN
    obj2 := obj1.next;
    obj1.next := obj;
    obj := obj1;
    obj1 := obj2
  END;
  WHILE obj <> NIL DO BEGIN
    IF obj.typ = StFwdLabel THEN
      OPM.Mark(128, obj.errpos);	(* unsatisfied forward reference *)
    obj := obj.next;
  END;
END;

PROCEDURE InsertFix(VAR obj: AObjects; pc: LONGINT);
VAR
  fix: FixPnt;
  x: OPT.Nodes;
BEGIN
  CASE obj.typ OF
    StFwdLabel:
          BEGIN
            NEW(fix);
            fix.pc := pc;
            fix.node := last;
            fix.pos := ORD(last.conval^.code[0])-3;	(* displacement position *)
            fix.next := obj.fix;
            obj.fix := fix;
          END;
    StVar, StVarPar, StGlobal:
          BEGIN
            IF export AND Cinline THEN  OPM.err(517);
            x := OPT.NewNode(Nfixup);
            x.obj := obj.obj;
            x.conval := OPT.NewConst();
            x.conval^.intval := pc;
            x.conval^.intval2 := OPM.errpos.CodePos;
            x.link := fixlist;
            fixlist := x;
          END;
  ELSE
    OPM.err(607)
  END
END;

PROCEDURE FixDisp(VAR fix: FixPnt);
VAR
  pos, disp: LONGINT;
BEGIN
  disp := pc - fix.pc;
  FOR pos := fix.pos TO fix.pos+3 DO BEGIN
    fix.node.conval^.code[pos] := CHR(disp);
    disp := _SYSTEM._DIV(disp,256);
  END;
END;
	
(* ------------------------------------------------------------------- *)
(* Code Generation functions *)

PROCEDURE NewNode;
VAR
  x: OPT.Nodes;
BEGIN
  x := OPT.NewNode(Ncode);
  x.conval := OPT.NewConst();
  FILLCHAR(x.conval^.code,LENGTH(x.conval^.code),#0);
  x.conval^.intval := OPM.errpos.CodePos;
  OPB.Link(first, last, x);
END;

PROCEDURE PutByte(b: LONGINT);
VAR pos: INTEGER;
BEGIN
  pos := ORD(last.conval^.code[0]) + 1;
  IF pos > 255 THEN BEGIN
    NewNode;
    pos := 1
  END;
  last.conval^.code[pos] := CHR(b);
  last.conval^.code[0] := CHR(pos);
  INC(pc);
END;

PROCEDURE PutBytes(b, size: LONGINT);
VAR pos: INTEGER;
BEGIN
  pos := ORD(last.conval^.code[0]);
  IF pos > 255-size THEN BEGIN
    NewNode;
    pos := 0
  END;
  WHILE size>0 DO BEGIN
    INC(pos);
    last.conval^.code[pos] := CHR(b);
    b := _SYSTEM._DIV(b,256);
    INC(pc);
    DEC(size);
  END;
  last.conval^.code[0] := CHR(pos);
END;

PROCEDURE ModRM(VAR modrm: LONGINT; VAR op: Operand; mode: INTEGER; UseSpecReg: BOOLEAN);
VAR
  Vmod: LONGINT;
BEGIN
  IF (IsRegMode(mode,TRUE) AND NOT UseSpecReg) OR IsSRegMode(mode, FALSE) OR IsSpecRegMode(mode) THEN BEGIN
    (* put reg index into the reg/opcode field *)
    ASSERT(op.reg <> none);
    modrm := modrm + op.reg * $8;
  END ELSE IF IsModRmMode(mode) OR IsMMode(mode) OR (IsRegMode(mode,TRUE) AND UseSpecReg) THEN BEGIN
    IF op.reg <> none THEN BEGIN
      (* put reg index into the r/m field, 3H into the Mod field *)
      modrm := modrm + op.reg + $0C0;
    END ELSE BEGIN
      (* set the mod field *)
      IF (op.disp = 0) AND NOT ((op.base = 5) AND (op.index = none)) THEN 	(* avoid case 0[EBP] *)
        Vmod := 0
      ELSE IF ConstSize(op.disp, TRUE)=1 THEN
        Vmod := 1
      ELSE IF (op.base = none) AND (op.index = none) THEN
        Vmod := 0
      ELSE
        Vmod := 2;
      modrm := modrm + Vmod * $40;
      IF op.index <> none THEN
        modrm := modrm + 4		(* use the sib byte *)
      ELSE IF op.base <> none THEN
        modrm := modrm + op.base
      ELSE
        modrm := modrm + 5;			(* special case: disp32 *)
    END;
  END ELSE
    OPM.err(1000);
END;

FUNCTION SIB(op: Operand): SHORTINT;
VAR
  val: INTEGER;
BEGIN
  val:=0;
  IF op.index = 4 THEN 	(* ESP not allowed *)
    OPM.err(501)
  ELSE IF op.index <> none THEN BEGIN
    val := op.base + op.index*$08;
    CASE op.scale OF
      none, 1:;
      2: val := val + 1 * $40;
      4: val := val + 2 * $40;
      8: val := val + 3 * $40;
    END
  END ELSE
    val := op.base + $04*$08;
  RESULT:=val;
END;

PROCEDURE GenCode(ind: INTEGER; VAR op: ARRAY OF Operand);
VAR
  i: LONGINT;
  name: STRING;
  fixobj: AObjects;
  UseRM, UseImm, UseSpecReg,
  UseSegReg, UseDisp: BOOLEAN;
  seg, reg, Vmodrm, Vsib,
  imm, immSize, disp: LONGINT;
BEGIN
  (* Initialisation *)
  disp := 0;
  imm:=0;
  immSize:=0;
  IF IsRelMode(op[0].mode) THEN	BEGIN	(* compute the size of the rel operand *)
    CASE ConstSize(op[0].imm-2 - pc, TRUE) OF		(* +2 = place for prefix AND opcode0 *)
      1   : op[0].mode := Rel8;
      2, 4: op[0].mode := Rel32;
    END
  END;
  name:=InstrTab[ind].name;
  Match(ind, op);
  IF ind < 0 THEN EXIT;	(* no matching opcode found *)
  (* Instruction prefix *)
  UseSpecReg := IsSpecRegMode(OpTab[ind].op[0]) OR IsSpecRegMode(OpTab[ind].op[1]);
  UseSegReg := IsSRegMode(OpTab[ind].op[0], FALSE) OR IsSRegMode(OpTab[ind].op[1], FALSE);
	
  IF NOT UseSpecReg AND NOT UseSegReg
  AND  (name[1] <> 'F')
  AND (((name  = 'OUT') AND (ModeSize(OpTab[ind].op[1]) = 2))
  OR  ((name <> 'OUT') AND (ModeSize(OpTab[ind].op[0]) = 2))) THEN BEGIN	(* 16-bit mode *)
    IF (OpTab[ind].op[0] = Rel16) OR (((name ='LEA')OR(name='OUTS'))
    AND (OpTab[ind].op[1] = RM16)) THEN
      PutByte($67)	(* operand size prefix *)
    ELSE IF (name <> 'RET') AND (name <> 'ARPL') AND (name <> 'STR') THEN
      PutByte($66);
  END;
		
(* prepare the instruction*)
  seg := none;
  reg := none;
  Vmodrm := 0;
  UseRM := FALSE;
  Vsib := none;
  UseImm := FALSE;
  fixobj := NIL;
  UseDisp := FALSE;
  i := 0;
  WHILE (i<3) AND (OpTab[ind].op[i] <> none) DO BEGIN
    (* segment prefix *)
    IF op[i].seg <> none THEN BEGIN
      IF seg <> none THEN
        OPM.err(504)
      ELSE
        seg := op[i].seg;
    END;
    (* register to be added to the opcode *)
    IF (OpTab[ind].reg = OmOp) AND (IsRegMode(OpTab[ind].op[i], TRUE)
    OR IsFRegMode(OpTab[ind].op[i], TRUE)) THEN
      reg := op[i].reg;
    (* modrm byte *)
    IF NeedModRM(OpTab[ind].op[i], OpTab[ind].reg) THEN BEGIN
      ModRM(Vmodrm, op[i], OpTab[ind].op[i], UseSpecReg);
      UseRM := TRUE;
      IF IsModRmMode(OpTab[ind].op[i]) OR IsMMode(OpTab[ind].op[i]) THEN BEGIN
        disp := op[i].disp;
        fixobj := op[i].obj;
        UseDisp := (disp <> 0) OR ((op[i].base = 5) AND (op[i].index = none));
      END
    END;
    (* sib byte *)
    IF (op[i].index <> none) OR (op[i].base = 4) THEN	BEGIN		(* process the escape ESP *)
      ASSERT (Vsib = none);
      Vsib := SIB(op[i]);
    END;
    IF IsImmMode(OpTab[ind].op[i]) OR (IsRelMode(OpTab[ind].op[i]) AND IsImmMode(op[i].mode)) THEN BEGIN
      ASSERT(NOT UseImm);
      UseImm := TRUE;
      imm := op[i].imm;
      immSize := ModeSize(OpTab[ind].op[i]);
    END;
    INC(i);
  END;

(* segment override *)
  IF seg <> none THEN PutByte(seg);
		
(* opcode *)
  IF reg = none THEN reg := 0;
  IF OpTab[ind].op1 <> none THEN BEGIN
    PutByte(OpTab[ind].op0);
    PutByte(OpTab[ind].op1 + reg)
  END ELSE
    PutByte(OpTab[ind].op0 + reg);

(* modr/m *)
  IF OpTab[ind].opc <> none THEN BEGIN
    ASSERT(FNAND(Vmodrm, $38) = 0);		(* reg/opcode field not used *)
    Vmodrm := Vmodrm + OpTab[ind].opc * $8;
    UseRM := TRUE
  END;
  IF UseRM THEN PutByte(Vmodrm);
		
(* sib *)
  IF Vsib <> none THEN PutByte(Vsib);
		
(* displacement / immediate *)
  IF  UseDisp THEN BEGIN
    IF fixobj <> NIL THEN InsertFix(fixobj, pc);
    IF ConstSize(disp, TRUE) = 1 THEN
      PutByte(disp)
    ELSE
      PutBytes(disp, 4);
  END;

  IF IsRelMode(OpTab[ind].op[0]) AND NOT IsImmMode(op[0].mode) THEN begin
    PutBytes(op[0].imm-pc-ModeSize(OpTab[ind].op[0]), ModeSize(OpTab[ind].op[0]));
    IF op[0].mode = FwdJmp THEN InsertFix(op[0].obj, pc);
  END;

  IF UseImm THEN PutBytes(imm, immSize);

END;

(* ------------------------------------------------------------------- *)
(* Parser functions *)

PROCEDURE ParseLabel(VAR Scaner: TScaner);
VAR
  obj: AObjects;
  fix: FixPnt;
BEGIN
  ASSERT(sym = ScLabel);		(* Test done by the caller *)
  insert(Scaner.name, obj);
  IF obj.typ = StFwdLabel THEN BEGIN
    fix := obj.fix;
    WHILE fix <> NIL DO BEGIN
      FixDisp(fix);
      fix := fix.next;
    END;
    obj.typ := StLabel;
    obj.val := pc
  END ELSE
    OPM.err(512);
  Get(Scaner,sym);
END;

PROCEDURE ParseInstruction(VAR Scaner: TScaner);
VAR
  ind, size: INTEGER;
  i: LONGINT;
  op: ARRAY [0..3] OF Operand;
  name: STRING;
BEGIN
(* Read Mnemonic *)
  size:=0;
  IF (Scaner.name = 'DB') OR (Scaner.name = 'DW') OR (Scaner.name = 'DD') THEN BEGIN
    CASE Scaner.name[2] OF
      'B': size := 1;
      'W': size := 2;
      'D': size := 4;
    END;
    Get(Scaner,sym);
    WHILE (sym = ScNumber) OR (sym = ScString) DO BEGIN
      IF (sym = ScString) AND (val = 1) THEN
        PutBytes(ORD(OPAstr[1]), size)
      ELSE IF (sym = ScNumber) AND (ConstSize(val, FALSE) <= size) THEN
        PutBytes(val, size)
      ELSE
        OPM.err(203);
      Get(Scaner,sym);
      IF sym = ScComma THEN
        Get(Scaner,sym)
      ELSE
        EXIT;;
    END;
  END ELSE IF (Scaner.name = 'DS') THEN BEGIN
    Get(Scaner,sym);
    IF (sym = ScString) THEN BEGIN
      FOR i := 1 TO val DO PutBytes(ORD(OPAstr[i]), 1);
      Get(Scaner,sym)
    END ELSE
      OPM.err(513);
  END ELSE BEGIN
    FOR i := 0 TO 2 DO op[i].mode := none;

    FindInstruction(Scaner.name, ind, TRUE);	(* index in the hash tab *)
    name:=Scaner.name;
    OPM.errpos := OPM.errpos;
    Get(Scaner,sym);

    IF ind < 0 THEN BEGIN
      OPM.err(502);
      EXIT;
    END;

    IF name = 'REP' THEN BEGIN	(* REP/REPE/REPZ/REPNE/REPNZ *)
      GenCode(ind, op);
      IF sym = ScCR THEN Get(Scaner,sym);		(* there should be no label here *)
      FindInstruction(Scaner.name, ind, TRUE);	(* index in the hash tab *)
      OPM.errpos := OPM.errpos;
      Get(Scaner,sym);
      IF ind < 0 THEN BEGIN
        OPM.err(502);
        EXIT;
      END;
    END;

    i := 0;
    IF sym <> ScCR THEN BEGIN
      REPEAT
        GetOperand(Scaner,op[i]);
        INC(i);
        IF sym <> ScComma THEN BREAK;
        Get(Scaner,sym);
      UNTIL FALSE;
    END;

    GenCode(ind, op);
  END; (* IF DB *)
END;

PROCEDURE ParseTarget(VAR Scaner: TScaner);
VAR
  obj: AObjects;
BEGIN
  REPEAT
    IF sym = ScIdent THEN BEGIN
      Qualident(Scaner,obj);
      IF (obj = NIL) THEN
        OPM.err(0)
      ELSE IF (obj.typ = StConst) THEN
        ATarget := ATarget + TargetTab[obj.val].flag
      ELSE
        OPM.err(0);
    END ELSE
      OPM.err(40);
    IF (sym = ScRBrace) THEN BEGIN
      Get(Scaner,sym);
      BREAK;
    END ELSE IF sym = ScComma THEN
      Get(Scaner,sym)
    ELSE IF sym <> ScIdent THEN BEGIN
      OPM.err(24(*rbrace*));
      BREAK;
    END ELSE
      OPM.err(19(*Comma*));
  UNTIL FALSE;
END;
	
(* ------------------------------------------------------------------- *)
(* Initialisation functions *)

PROCEDURE InsertReg(name: STRING; t, v: INTEGER);
VAR
  obj: AObjects;
BEGIN
  insert(name, obj);
  obj.typ := t;
  obj.val := v;
  obj.size := 0
END;
	
PROCEDURE InsertType(name: STRING; s: INTEGER);
VAR
  obj: AObjects;
BEGIN
  insert(name, obj);
  obj.typ := StType;
  obj.val := none;
  obj.size := s;
END;

PROCEDURE Init;
BEGIN 
  (* Internal Symbol-Table *)
    root.next := NIL;
    root.left := NIL;
    root.right := NIL;
    InsertReg('AL', RegAL, 0);
    InsertReg('AH', Reg8, 4);
    InsertReg('AX', RegAX, 0);
    InsertReg('EAX', RegEAX, 0);
    InsertReg('BL', Reg8, 3);
    InsertReg('BH', Reg8, 7);
    InsertReg('BX', Reg16, 3);
    InsertReg('EBX', Reg32, 3);
    InsertReg('CL', Reg8, 1);
    InsertReg('CH', Reg8, 5);
    InsertReg('CX', Reg16, 1);
    InsertReg('ECX', Reg32, 1);
    InsertReg('DL', Reg8, 2);
    InsertReg('DH', Reg8, 6);
    InsertReg('DX', Reg16, 2);
    InsertReg('EDX', Reg32, 2);
    InsertReg('SP', Reg16, 4);
    InsertReg('ESP', Reg32, 4);
    InsertReg('BP', Reg16, 5);
    InsertReg('EBP', Reg32, 5);
    InsertReg('SI', Reg16, 6);
    InsertReg('ESI', Reg32, 6);
    InsertReg('DI', Reg16, 7);
    InsertReg('EDI', Reg32, 7);
			
    InsertReg('CS', RegCS, $1);
    InsertReg('SS', RegSS, $2);
    InsertReg('DS', RegDS, $3);
    InsertReg('ES', RegES, $0);
    InsertReg('FS', RegFS, $4);
    InsertReg('GS', RegGS, $5);
			
    InsertReg('CR0', CRReg, 0);
    InsertReg('CR2', CRReg, 2);
    InsertReg('CR3', CRReg, 3);
    InsertReg('CR4', CRReg, 4);
    InsertReg('DR0', DRReg, 0);
    InsertReg('DR1', DRReg, 1);
    InsertReg('DR2', DRReg, 2);
    InsertReg('DR3', DRReg, 3);
    InsertReg('DR6', DRReg, 6);
    InsertReg('DR7', DRReg, 7);
    (*InsertReg('TR6', TRReg, 6); InsertReg('TR7', TRReg, 7);*)
			
    InsertReg('ST', FReg, 0);

    InsertType('BYTE', 1);
    InsertType('WORD', 2);
    InsertType('DWORD', 4);
    InsertType('QWORD', 8);
    InsertType('TBYTE', 10);
			
  (* Global variables *)
    pc := 0;
    ATarget := [];
END;
	
PROCEDURE Close;
BEGIN
  root.right := NIL;
END;
	
PROCEDURE InsertTarget(name: STRING; flag: TSET);
BEGIN
  TargetTab[TargetCount].name:=name;
  TargetTab[TargetCount].flag := flag;
  INC(TargetCount);
END;
	
PROCEDURE BodyInit;
BEGIN
  OPM.LogWLn;
  OPM.LogWStrLn('  using OP2-ASM add-on / prk');
  NEW(root);
  root.left := NIL;
  root.right := NIL;
  root.next := NIL;

  OPAData.DataInsert;

  SizeTab[0] := 0;
  SizeTab[1] := 1;
  SizeTab[2] := 1;
  SizeTab[3] := 1;
  SizeTab[4] := 1;
  SizeTab[5] := 2;
  SizeTab[6] := 4;
  SizeTab[7] := 4;
  SizeTab[8] := 8;
  SizeTab[9] := 4;
  SizeTab[10] := 4;
  SizeTab[11] := 4;
  SizeTab[12] := 0;
  SizeTab[13] := 4;
  SizeTab[14] := 4;
  SizeTab[15] := 0;

  TargetCount := 0;
  InsertTarget('i386', [i386]);
  InsertTarget('i486', [i386, i486]);
  InsertTarget('Pentium', [i386, i486, Pentium]);
  InsertTarget('PentiumPro', [i386, i486, Pentium, PentiumPro]);
  InsertTarget('FPU', [FPU]);
  InsertTarget('Privileged', [Privileged]);
END;
(* ------------------------------------------------------------------- *)
(* Entry points and exported functions *)

PROCEDURE Assemble(VAR Scaner: TScaner; VAR x: OPT.Nodes;  exported, inlined: BOOLEAN);
BEGIN
  export := exported;
  Cinline := inlined;

  IF FirstTime THEN BEGIN
    BodyInit;
    FirstTime := FALSE;
  END;

  Init;
  first := NIL;
  last := NIL;
  NewNode;
  fixlist := NIL;
  Get(Scaner,sym);

  IF sym = ScLBrace THEN BEGIN
    Get(Scaner,sym);
    ParseTarget(Scaner);
  END ELSE BEGIN
    Get(Scaner,sym);
    ATarget := [];
  END;

  IF ATarget = [] THEN BEGIN
    OPM.err(516);
    ATarget := [0..31];
  END;

  WHILE (sym <> ScEnd) AND (sym <> ScUndef) DO BEGIN
    IF sym = ScLabel THEN ParseLabel(Scaner);
    IF sym = ScIdent THEN ParseInstruction(Scaner);
    WHILE (sym <> ScEnd) AND (sym <> ScCR) DO BEGIN
      Get(Scaner,sym);
      OPM.err(510)
    END;
    Get(Scaner,sym);
  END;

  IF OPM.noerr THEN CheckLabels;	(* check for undefined forward jumps *)
  Close;
  OPB.Construct(Nassembler, first, fixlist);
  x := first;
  x.conval := NIL;
END;

PROCEDURE OPABegin;
BEGIN
	FirstTime := TRUE;	(* the module initialisation is delayed to the first call *)
END;

END.