UNIT OPAData;

INTERFACE

USES
  _SYSTEM, OPA, OPM;

VAR
  OpTabN: INTEGER;
  Clash : ARRAY [0..32] OF LONGINT;
  InstrCount, Average: LONGINT;
  PROCEDURE DataInsert;
  
IMPLEMENTATION

PROCEDURE InsertMnemo(name: STRING; VAR k: INTEGER; target: TSET);
VAR
  n: INTEGER;
BEGIN
  n := 0;
  REPEAT
    OPA.HashFn(name, k, n);
  UNTIL (InstrTab[k].name = '') OR (n > 2*InstrTabLen) OR (InstrTab[k].name = name);
  IF n > 2*InstrTabLen THEN OPM.LogWStrLn('iOPA, internal instruction table too small!!');
  IF InstrTab[k].name = name THEN BEGIN
    OPM.LogWStr(name);
    OPM.LogWStrLn(' already in the table');
    EXIT;
  END;
  InstrTab[k].name    := name;
  InstrTab[k].target  := target;
  InstrTab[k].start   := OpTabN;
  InstrTab[k].vend    := OpTabN;
  IF n < LENGTH(Clash) THEN
    INC(Clash[n])
  ELSE
    INC(Clash[LENGTH(Clash)-1]);
  Average := Average + n;
  INC(InstrCount);
END;

PROCEDURE InsertAlias(name, alias: STRING);
VAR
  n,a: INTEGER;
BEGIN
  OPA.FindInstruction(name, n, FALSE);
  IF n = -1 THEN BEGIN
    OPM.LogWStr('InsertAlias: the name doesn''t exist ');
    OPM.LogWStrLn('"'+name+'"');
    EXIT;
  END;
  InsertMnemo(alias, a, InstrTab[n].target);
  InstrTab[a].start := InstrTab[n].start;
  InstrTab[a].vend  := InstrTab[n].vend;
END;
	
PROCEDURE InsertOp(idx, dest, src, op0, op1, opc, om: INTEGER; o3: INTEGER = -1);
VAR
  pos: INTEGER;
BEGIN
  pos := InstrTab[idx].vend;
  IF OpTab[pos].op0 <> -1 THEN ;//Message('Op Table, overwriting') END;
  OpTab[pos].op[0] := dest;
  OpTab[pos].op[1] := src;
  IF o3=-1 THEN
    OpTab[pos].op[2] := none	(* dest / src addressing mode *)
  ELSE
    OpTab[pos].op[2] := o3;
  OpTab[pos].op0 := op0;
  OpTab[pos].op1 := op1;	(* opcodes *)
  OpTab[pos].opc := opc;
  OpTab[pos].reg := om;	(* opcode modifier (rm.opcode) / register storing place (opcode /rm byte) *)
  IF OpTabN < OpTabLen-1 THEN BEGIN
    INC(InstrTab[idx].vend);
    INC(OpTabN)
  END ELSE
    ;//Message('Op Table FULL')
END;

PROCEDURE InsertInstr1;
VAR
  i,k: INTEGER;
BEGIN
  InsertMnemo('AAA', k, [i386]);
  InsertOp(k, none, none, $37, none, none, none);

  InsertMnemo('AAD', k, [i386]);
  InsertOp(k, none, none, $0D5, $0A, none, none);

  InsertMnemo('AAM', k, [i386]);
  InsertOp(k, none, none, $0D4, $0A, none, none);

  InsertMnemo('AAS', k, [i386]);
  InsertOp(k, none, none, $3F, none, none, none);

  FOR i:=0 TO 7 DO BEGIN				(* similar integer ops *)
    CASE i OF
      0: InsertMnemo('ADD', k, [i386]);
      1: InsertMnemo('OR' , k, [i386]);
      2: InsertMnemo('ADC', k, [i386]);
      3: InsertMnemo('SBB', k, [i386]);
      4: InsertMnemo('AND', k, [i386]);
      5: InsertMnemo('SUB', k, [i386]);
      6: InsertMnemo('XOR', k, [i386]);
      7: InsertMnemo('CMP', k, [i386]);
    END;
    InsertOp(k, RegAL, Imm8, $04 + i*$08, none, none, none);
    InsertOp(k, RegAX, Imm16, $05 + i*$08, none, none, none);
    InsertOp(k, RegEAX, Imm32, $05 + i*$08, none, none, none);
    InsertOp(k, RM8, Imm8, $80, none, i, none);
    InsertOp(k, RM16, SImm8, $83, none, i, none);
    InsertOp(k, RM32, SImm8, $83, none, i, none);
    InsertOp(k, RM16, Imm16, $81, none, i, none);
    InsertOp(k, RM32, Imm32, $81, none, i, none);
    InsertOp(k, RM8, Reg8, i * $08, none, none, OmRM);
    InsertOp(k, RM16, Reg16, $01 + i * $08, none, none, OmRM);
    InsertOp(k, RM32, Reg32, $01 + i * $08, none, none, OmRM);
    InsertOp(k, Reg8, RM8, $02 + i * $08, none, none, OmRM);
    InsertOp(k, Reg16, RM16, $03 + i * $08, none, none, OmRM);
    InsertOp(k, Reg32, RM32, $03 + i * $08, none, none, OmRM);
  END;

  InsertMnemo('ARPL', k, [i386]);
  InsertOp(k, RM16, Reg16, $63, none, none, OmRM);

  InsertMnemo('BOUND', k, [i386]);
  InsertOp(k, Reg16, M16, $62, none, none, OmRM);
  InsertOp(k, Reg32, M32, $62, none, none, OmRM);

  InsertMnemo('BSF', k, [i386]);
  InsertOp(k, Reg16, RM16, $0F, $0BC, none, none);
  InsertOp(k, Reg32, RM32, $0F, $0BC, none, none);

  InsertMnemo('BSR', k, [i386]);
  InsertOp(k, Reg16, RM16, $0F, $0BD, none, none);
  InsertOp(k, Reg32, RM32, $0F, $0BD, none, none);

  InsertMnemo('BSWAP', k, [i486]);
  InsertOp(k, Reg32, none, $0F, $0C8, none, OmOp);

  FOR i := 0 TO 3 DO BEGIN
    CASE i OF
      0: InsertMnemo('BT' , k, [i386]);
      1: InsertMnemo('BTS', k, [i386]);
      2: InsertMnemo('BTR', k, [i386]);
      3: InsertMnemo('BTC', k, [i386]);
    END;
    InsertOp(k, RM16, Reg16, $0F, $0A3 + i*8, none, OmRM);
    InsertOp(k, RM32, Reg32, $0F, $0A3 + i*8, none, OmRM);
    InsertOp(k, RM16, Imm8, $0F, $0BA, i+4, none);
    InsertOp(k, RM32, Imm8, $0F, $0BA, i+4, none)
  END;

  InsertMnemo('CALL', k, [i386]);
  InsertOp(k, RM16, none, $0FF, none, 2, none);
  InsertOp(k, Rel32, none, $0E8, none, none, none);
  InsertOp(k, RM32, none, $0FF, none, 2, none);

  InsertMnemo('CBW', k, [i386]);
  InsertOp(k, none, none, $066, $098, none, none);

  InsertMnemo('CWDE', k, [i386]);
  InsertOp(k, none, none, $098, none, none, none);

  FOR i := 8 TO $0D DO BEGIN
    CASE i OF
        8  : InsertMnemo('CLC', k, [i386]);
        9  : InsertMnemo('STC', k, [i386]);
        $0A: InsertMnemo('CLI', k, [i386]);
        $0B: InsertMnemo('STI', k, [i386]);
        $0C: InsertMnemo('CLD', k, [i386]);
        $0D: InsertMnemo('STD', k, [i386]);
    END;
    InsertOp(k, none, none, $0F0 + i, none, none, none)
  END;

  InsertMnemo('CLTS', k, [i386, Privileged]);
  InsertOp(k, none, none, $0F, $06, none, none);

  InsertMnemo('CMC', k, [i386]);
  InsertOp(k, none, none, $0F5, none, none, none);

  FOR i := 0 TO 15 DO BEGIN
    CASE i OF
      0: InsertMnemo('CMOVO', k, [PentiumPro]);
      1: InsertMnemo('CMOVNO', k, [PentiumPro]);
      2: InsertMnemo('CMOVB', k, [PentiumPro]);
      3: InsertMnemo('CMOVNB', k, [PentiumPro]);
      4: InsertMnemo('CMOVZ', k, [PentiumPro]);
      5: InsertMnemo('CMOVNZ', k, [PentiumPro]);
      6: InsertMnemo('CMOVBE', k, [PentiumPro]);
      7: InsertMnemo('CMOVNBE', k, [PentiumPro]);
      8: InsertMnemo('CMOVS', k, [PentiumPro]);
      9: InsertMnemo('CMOVNS', k, [PentiumPro]);
      10: InsertMnemo('CMOVP', k, [PentiumPro]);
      11: InsertMnemo('CMOVNP', k, [PentiumPro]);
      12: InsertMnemo('CMOVL', k, [PentiumPro]);
      13: InsertMnemo('CMOVNL', k, [PentiumPro]);
      14: InsertMnemo('CMOVLE', k, [PentiumPro]);
      15: InsertMnemo('CMOVNLE', k, [PentiumPro]);
    END;
    InsertOp(k, Reg16, RM16, $0F, 40 + i, none, none);
    InsertOp(k, Reg32, RM32, $0F, 40 + i, none, none);
  END;
  InsertAlias('CMOVB', 'CMOVC');
  InsertAlias('CMOVB', 'CMOVNAE');
  InsertAlias('CMOVNB', 'CMOVNC');
  InsertAlias('CMOVNB', 'CMOVAE');
  InsertAlias('CMOVZ', 'CMOVE');
  InsertAlias('CMOVNZ', 'CMOVNE');
  InsertAlias('CMOVBE', 'CMOVNA');
  InsertAlias('CMOVNBE', 'CMOVA');
  InsertAlias('CMOVP', 'CMOVPE');
  InsertAlias('CMOVNP', 'CMOVPO');
  InsertAlias('CMOVL', 'CMOVNGE');
  InsertAlias('CMOVNL', 'CMOVGE');
  InsertAlias('CMOVLE', 'CMOVNG');
  InsertAlias('CMOVNLE', 'CMOVG');

  InsertMnemo('CMPSB', k, [i386]);
  InsertOp(k, none, none, $0A6, none, none, none);

  InsertMnemo('CMPSW', k, [i386]);
  InsertOp(k, none, none, $066, $0A7, none, none);

  InsertMnemo('CMPSD', k, [i386]);
  InsertOp(k, none, none, $0A7, none, none, none);

  InsertMnemo('CMPXCHG', k, [i486]);
  InsertOp(k, RM8, Reg8, $0F, $0B0, none, OmRM);
  InsertOp(k, RM16, Reg16, $0F, $0B1, none, OmRM);
  InsertOp(k, RM32, Reg32, $0F, $0B1, none, OmRM);

  InsertMnemo('CMPXCHG8B', k, [Pentium]);
  InsertOp(k, M64, none, $0F, $0C7, 1, none);

  InsertMnemo('CPUID', k, [Pentium]);
  InsertOp(k, none, none, $0F, $0A2, none, none);

  InsertMnemo('CWD', k, [i386]);
  InsertOp(k, none, none, $066, $099, none, none);
  InsertMnemo('CDQ', k, [i386]);
  InsertOp(k, none, none, $099, none, none, none);

  InsertMnemo('DAA', k, [i386]);
  InsertOp(k, none, none, $027, none, none, none);
  InsertMnemo('DAS', k, [i386]);
  InsertOp(k, none, none, $02F, none, none, none);

  InsertMnemo('DEC', k, [i386]);
  InsertOp(k, Reg16, none, $48, none, none, OmOp);
  InsertOp(k, Reg32, none, $48, none, none, OmOp);
  InsertOp(k, RM8, none, $0FE, none, 1, none);
  InsertOp(k, RM16, none, $0FF, none, 1, none);
  InsertOp(k, RM32, none, $0FF, none, 1, none);

  InsertMnemo('DIV', k, [i386]);
  InsertOp(k, RegAL, RM8, $0F6, none, 6, none);
  InsertOp(k, RegAX, RM16, $0F7, none, 6, none);
  InsertOp(k, RegEAX, RM32, $0F7, none, 6, none);
  InsertOp(k, RM8, none, $0F6, none, 6, none);
  InsertOp(k, RM16, none, $0F7, none, 6, none);
  InsertOp(k, RM32, none, $0F7, none, 6, none);

  InsertMnemo('HLT', k, [i386, Privileged]);
  InsertOp(k, none, none, $0F4, none, none, none);

  InsertMnemo('IDIV', k, [i386]);
  InsertOp(k, RegAL, RM8, $0F6, none, 7, none);
  InsertOp(k, RegAX, RM16, $0F7, none, 7, none);
  InsertOp(k, RegEAX, RM32, $0F7, none, 7, none);
  InsertOp(k, RM8, none, $0F6, none, 7, none);
  InsertOp(k, RM16, none, $0F7, none, 7, none);
  InsertOp(k, RM32, none, $0F7, none, 7, none);


  InsertMnemo('IMUL', k, [i386]);
  InsertOp(k, RM8, none, $0F6, none, 5, none);
  InsertOp(k, RM16, none, $0F7, none, 5, none);
  InsertOp(k, RM32, none, $0F7, none, 5, none);
  InsertOp(k, Reg16, RM16, $0F, $0AF, none, OmRM);
  InsertOp(k, Reg32, RM32, $0F, $0AF, none, OmRM);
  InsertOp(k, Reg16, SImm8, $06B, none, none, OmRM);
  InsertOp(k, Reg32, SImm8, $06B, none, none, OmRM);
  InsertOp(k, Reg16, Imm16, $069, none, none, OmRM);
  InsertOp(k, Reg32, Imm32, $069, none, none, OmRM);
  InsertOp(k, Reg16, RM16, $06B, none, none, OmRM, SImm8);
  InsertOp(k, Reg32, RM32, $06B, none, none, OmRM, SImm8);
  InsertOp(k, Reg16, RM16, $069, none, none, OmRM, Imm16);
  InsertOp(k, Reg32, RM32, $069, none, none, OmRM, Imm32);
END;
		
PROCEDURE InsertInstr2;
  VAR i,k: INTEGER;
BEGIN
  InsertMnemo('IN', k, [i386]);
  InsertOp(k, RegAL, Imm8, $0E4, none, none, none);
  InsertOp(k, RegAX, Imm8, $0E5, none, none, none);
  InsertOp(k, RegEAX, Imm8, $0E5, none, none, none);
  InsertOp(k, RegAL, RegDX, $0EC, none, none, none);
  InsertOp(k, RegAX, RegDX, $0ED, none, none, none);
  InsertOp(k, RegEAX, RegDX, $0ED, none, none, none);

  InsertMnemo('INC', k, [i386]);
  InsertOp(k, Reg16, none, $040, none, none, OmOp);
  InsertOp(k, Reg32, none, $040, none, none, OmOp);
  InsertOp(k, RM8, none, $0FE, none, 0, none);
  InsertOp(k, RM16, none, $0FF, none, 0, none);
  InsertOp(k, RM32, none, $0FF, none, 0, none);

  InsertMnemo('INSB', k, [i386]);
  InsertOp(k, none, none, $06C, none, none, none);

  InsertMnemo('INSW', k, [i386]);
  InsertOp(k, none, none, $066, $06D, none, none);
		
  InsertMnemo('INSD', k, [i386]);
  InsertOp(k, none, none, $06D, none, none, none);
		
  InsertMnemo('INT', k, [i386]);
  InsertOp(k, Const3, none, $0CC, none, none, none);
  InsertOp(k, Imm8, none, $0CD, none, none, none);
		
  InsertMnemo('INTO', k, [i386]);
  InsertOp(k, none, none, $0CE, none, none, none);
		
  InsertMnemo('INVD', k, [i486, Privileged]);
  InsertOp(k, none, none, $0F, $08, none, none);
		
  InsertMnemo('INVLPG', k, [i486, Privileged]);
  InsertOp(k, M0, none, $0F, $01, 7, none);
		
  InsertMnemo('IRETD', k, [i386]);
  InsertOp(k, none, none, $0CF, none, none, none);
		
  InsertMnemo('JCXZ', k, [i386]);
  InsertOp(k, Rel8, none, $067, $0E3, none, none);
		
  InsertMnemo('JECXZ', k, [i386]);
  InsertOp(k, Rel8, none, $0E3, none, none, none);
		
  FOR i := 0 TO 15 DO BEGIN
    CASE i OF
      0 : InsertMnemo('JO', k, [i386]);
      1 : InsertMnemo('JNO', k, [i386]);
      2 : InsertMnemo('JB', k, [i386]);
      3 : InsertMnemo('JNB', k, [i386]);
      4 : InsertMnemo('JZ', k, [i386]);
      5 : InsertMnemo('JNZ', k, [i386]);
      6 : InsertMnemo('JBE', k, [i386]);
      7 : InsertMnemo('JNBE', k, [i386]);
      8 : InsertMnemo('JS', k, [i386]);
      9 : InsertMnemo('JNS', k, [i386]);
      10: InsertMnemo('JP', k, [i386]);
      11: InsertMnemo('JNP', k, [i386]);
      12: InsertMnemo('JL', k, [i386]);
      13: InsertMnemo('JNL', k, [i386]);
      14: InsertMnemo('JLE', k, [i386]);
      15: InsertMnemo('JNLE', k, [i386]);
    END;
    InsertOp(k, Rel8, none, $70 + i, none, none, none);
    InsertOp(k, Rel16, none, $0F, $80 + i, none, none);
    InsertOp(k, Rel32, none, $0F, $80 + i, none, none);
  END;
  InsertAlias('JB', 'JC'); InsertAlias('JB', 'JNAE');
  InsertAlias('JNB', 'JNC'); InsertAlias('JNB', 'JAE');
  InsertAlias('JZ', 'JE');
  InsertAlias('JNZ', 'JNE');
  InsertAlias('JBE', 'JNA');
  InsertAlias('JNBE', 'JA');
  InsertAlias('JP', 'JPE');
  InsertAlias('JNP', 'JPO');
  InsertAlias('JL', 'JNGE');
  InsertAlias('JNL', 'JGE');
  InsertAlias('JLE', 'JNG');
  InsertAlias('JNLE', 'JG');
		
  InsertMnemo('JMP', k, [i386]);
  InsertOp(k, Rel8, none, $0EB, none, none, none);
  InsertOp(k, RM16, none, $0FF, none, 4, none);
  InsertOp(k, Rel32, none, $0E9, none, none, none);
  InsertOp(k, RM32, none, $0FF, none, 4, none);
		
  InsertMnemo('LAHF', k, [i386]);
  InsertOp(k, none, none, $09F, none, none, none);
		
  InsertMnemo('LAR', k, [i386]);
  InsertOp(k, Reg16, RM16, $0F, $02, none, OmRM);
  InsertOp(k, Reg32, RM32, $0F, $02, none, OmRM);

  InsertMnemo('LEA', k, [i386]);
  InsertOp(k, Reg16, M16, $08D, none, none, OmRM);
  InsertOp(k, Reg16, M32, $08D, none, none, OmRM);
  (*InsertOp(k, Reg32, M16, 08DH, none, none, OmRM);*)
  InsertOp(k, Reg32, M32, $08D, none, none, OmRM);

  InsertMnemo('LEAVE', k, [i386]);
  InsertOp(k, none, none, $0C9, none, none, none);

  InsertMnemo('LDS', k, [i386]);
  InsertOp(k, Reg16, M16, $0C5, none, none, OmRM);
  InsertOp(k, Reg32, M32, $0C5, none, none, OmRM);
  InsertMnemo('LSS', k, [i386]);
  InsertOp(k, Reg16, M16, $0F, $0B2, none, OmRM);
  InsertOp(k, Reg32, M32, $0F, $0B2, none, OmRM);
  InsertMnemo('LES', k, [i386]);
  InsertOp(k, Reg16, M16, $0C4, none, none, OmRM);
  InsertOp(k, Reg32, M32, $0C4, none, none, OmRM);
  InsertMnemo('LFS', k, [i386]);
  InsertOp(k, Reg16, M16, $0F, $0B4, none, OmRM);
  InsertOp(k, Reg32, M32, $0F, $0B4, none, OmRM);
  InsertMnemo('LGS', k, [i386]);
  InsertOp(k, Reg16, M16, $0F, $0B5, none, OmRM);
  InsertOp(k, Reg32, M32, $0F, $0B5, none, OmRM);
		
  InsertMnemo('LGDT', k, [i386, Privileged]);
  InsertOp(k, M0, none, $0F, $01, 2, none);
		
  InsertMnemo('LIDT', k, [i386, Privileged]);
  InsertOp(k, M0, none, $0F, $01, 3, none);

  InsertMnemo('LOCK', k, [i386]);
  InsertOp(k, none, none, $0F0, none, none, none);

  InsertMnemo('LLDT', k, [i386, Privileged]);
  InsertOp(k, RM16, none, $0F, $00, 2, none);
		
  InsertMnemo('LMSW', k, [i386, Privileged]);
  InsertOp(k, RM16, none, $0F, $01, 6, none);

  InsertMnemo('LODSB', k, [i386]);
  InsertOp(k, none, none, $0AC, none, none, none);

  InsertMnemo('LODSW', k, [i386]);
  InsertOp(k, none, none, $066, $0AD, none, none);

  InsertMnemo('LODSD', k, [i386]);
  InsertOp(k, none, none, $0AD, none, none, none);

  InsertMnemo('LOOP', k, [i386]);
  InsertOp(k, Rel8, none, $0E2, none, none, none);
		
  InsertMnemo('LOOPE', k, [i386]);
  InsertOp(k, Rel8, none, $0E1, none, none, none);
  InsertAlias('LOOPE', 'LOOPZ');
		
  InsertMnemo('LOOPNE', k, [i386]);
  InsertOp(k, Rel8, none, $0E0, none, none, none);
  InsertAlias('LOOPNE', 'LOOPNZ');

  InsertMnemo('LSL', k, [i386]);
  InsertOp(k, Reg16, RM16, $0F, $03, none, OmRM);
  InsertOp(k, Reg32, RM32, $0F, $03, none, OmRM);

  InsertMnemo('LTR', k, [i386, Privileged]);
  InsertOp(k, RM16, none, $0F, $00, 3, none);
END;
	
PROCEDURE InsertInstr3;
  VAR k: INTEGER;
BEGIN
  InsertMnemo('MOV', k, [i386]);
  InsertOp(k, RM8, Reg8, $88, none, none, OmRM);
  InsertOp(k, RM16, Reg16, $89, none, none, OmRM);
  InsertOp(k, RM32, Reg32, $89, none, none, OmRM);
  InsertOp(k, Reg8, RM8,  $8A, none, none, OmRM);
  InsertOp(k, Reg16, RM16, $8B, none, none, OmRM);
  InsertOp(k, Reg32, RM32, $8B, none, none, OmRM);
  InsertOp(k, RM16, SReg, $8C, none, none, OmRM);
  InsertOp(k, SReg, RM16, $8E, none, none, OmRM);
  InsertOp(k, Reg8, Imm8,  $0B0, none, none, OmOp);
  InsertOp(k, Reg16, Imm16,  $0B8, none, none, OmOp);
  InsertOp(k, Reg32, Imm32,  $0B8, none, none, OmOp);
  InsertOp(k, RM8, Imm8,  $0C6, none, none, none);
  InsertOp(k, RM16, Imm16,  $0C7, none, none, none);
  InsertOp(k, RM32, Imm32,  $0C7, none, none, none);
  InsertOp(k, Reg32, CRReg, $0F, $20, none, OmRM);
  InsertOp(k, CRReg, Reg32, $0F, $22, none, OmRM);
  InsertOp(k, Reg32, DRReg, $0F, $21, none, OmRM);
  InsertOp(k, DRReg, Reg32, $0F, $23, none, OmRM);
  InsertOp(k, Reg32, TRReg, $0F, $24, none, OmRM);
  InsertOp(k, TRReg, Reg32, $0F, $26, none, OmRM);

  InsertMnemo('MOVSB', k, [i386]);
  InsertOp(k, none, none, $0A4, none, none, none);
		
  InsertMnemo('MOVSW', k, [i386]);
  InsertOp(k, none, none, $066, $0A5, none, none);
		
  InsertMnemo('MOVSD', k, [i386]);
  InsertOp(k, none, none, $0A5, none, none, none);
		
  InsertMnemo('MOVSX', k, [i386]);
  InsertOp(k, Reg16, RM8, $0F, $0BE, none, OmRM);
  InsertOp(k, Reg32, RM8, $0F, $0BE, none, OmRM);
  InsertOp(k, Reg32, RM16, $0F, $0BF, none, OmRM);
		
  InsertMnemo('MOVZX', k, [i386]);
  InsertOp(k, Reg16, RM8, $0F, $0B6, none, OmRM);
  InsertOp(k, Reg32, RM8, $0F, $0B6, none, OmRM);
  InsertOp(k, Reg32, RM16, $0F, $0B7, none, OmRM);

  InsertMnemo('MUL', k, [i386]);
  InsertOp(k, RegAL, RM8, $0F6, none, 4, none);
  InsertOp(k, RegAX, RM16, $0F7, none, 4, none);
  InsertOp(k, RegEAX, RM32, $0F7, none, 4, none);
  InsertOp(k, RM8, none, $0F6, none, 4, none);
  InsertOp(k, RM16, none, $0F7, none, 4, none);
  InsertOp(k, RM32, none, $0F7, none, 4, none);
		
  InsertMnemo('NEG', k, [i386]);
  InsertOp(k, RM8, none, $0F6, none, 3, none);
  InsertOp(k, RM16, none, $0F7, none, 3, none);
  InsertOp(k, RM32, none, $0F7, none, 3, none);

  InsertMnemo('NOP', k, [i386]);
  InsertOp(k, none, none, $090, none, none, none);
		
  InsertMnemo('NOT', k, [i386]);
  InsertOp(k, RM8, none, $0F6, none, 2, none);
  InsertOp(k, RM16, none, $0F7, none, 2, none);
  InsertOp(k, RM32, none, $0F7, none, 2, none);

  InsertMnemo('OUT', k, [i386]);
  InsertOp(k, Imm8, RegAL, $0E6, none, none, none);
  InsertOp(k, Imm8, RegAX, $0E7, none, none, none);
  InsertOp(k, Imm8, RegEAX, $0E7, none, none, none);
  InsertOp(k, RegDX, RegAL, $0EE, none, none, none);
  InsertOp(k, RegDX, RegAX, $0EF, none, none, none);
  InsertOp(k, RegDX, RegEAX, $0EF, none, none, none);

  InsertMnemo('OUTSB', k, [i386]);
  InsertOp(k, none, none, $06E, none, none, none);
  InsertMnemo('OUTSW', k, [i386]);
  InsertOp(k, none, none, $066, $06F, none, none);
  InsertMnemo('OUTSD', k, [i386]);
  InsertOp(k, none, none, $06F, none, none, none);

  InsertMnemo('POP', k, [i386]);
  InsertOp(k, M32, none, $08F, none, 0, none);
  InsertOp(k, Reg32, none, $058, none, none, OmOp);
  InsertOp(k, RegDS, none, $01F, none, none, none);
  InsertOp(k, RegES, none, $07, none, none, none);
  InsertOp(k, RegSS, none, $017, none, none, none);
  InsertOp(k, RegFS, none, $0F, $0A1, none, none);
  InsertOp(k, RegGS, none, $0F, $0A9, none, none);
		
  InsertMnemo('POPAD', k, [i386]);
  InsertOp(k, none, none, $061, none, none, none);

  InsertMnemo('POPFD', k, [i386]);
  InsertOp(k, none, none, $09D, none, none, none);

  InsertMnemo('PUSH', k, [i386]);
  InsertOp(k, Reg32, none, $050, none, none, OmOp);
  InsertOp(k, SImm8, none, $06A, none, none, none);
  InsertOp(k, Imm32, none, $068, none, none, none);
  InsertOp(k, M32, none, $0FF, none, 6, none);
  InsertOp(k, RegCS, none, $0E, none, none, none);
  InsertOp(k, RegDS, none, $01E, none, none, none);
  InsertOp(k, RegES, none, $06, none, none, none);
  InsertOp(k, RegSS, none, $016, none, none, none);
  InsertOp(k, RegFS, none, $0F, $0A0, none, none);
  InsertOp(k, RegGS, none, $0F, $0A8, none, none);
END;
		
PROCEDURE InsertInstr4;
  VAR i,k: INTEGER;
BEGIN
  InsertMnemo('PUSHAD', k, [i386]);
  InsertOp(k, none, none, $060, none, none, none);

  InsertMnemo('PUSHFD', k, [i386]);
  InsertOp(k, none, none, $09C, none, none, none);

  FOR i := 0 TO 7 DO BEGIN
    CASE i OF
      0: InsertMnemo('ROL', k, [i386]);
      1: InsertMnemo('ROR', k, [i386]);
      2: InsertMnemo('RCL', k, [i386]);
      3: InsertMnemo('RCR', k, [i386]);
      4: InsertMnemo('SHL', k, [i386]);
      5: InsertMnemo('SHR', k, [i386]);
      6: ; (*SAL*)
      7: InsertMnemo('SAR', k, [i386]);
    END;
    IF i = 6 THEN ELSE BEGIN
      InsertOp(k, RM8, Const1, $0D0, none, i, none);
      InsertOp(k, RM8, RegCL, $0D2, none, i, none);
      InsertOp(k, RM8, Imm8, $0C0, none, i, none);
      InsertOp(k, RM16, Const1, $0D1, none, i, none);
      InsertOp(k, RM16, RegCL, $0D3, none, i, none);
      InsertOp(k, RM16, Imm8, $0C1, none, i, none);
      InsertOp(k, RM32, Const1, $0D1, none, i, none);
      InsertOp(k, RM32, RegCL, $0D3, none, i, none);
      InsertOp(k, RM32, Imm8, $0C1, none, i, none)
    END
  END;
  InsertAlias('SHL', 'SAL');
		
  InsertMnemo('REP', k, [i386]);
  InsertOp(k, none, none, $0F3, none, none, none);
  InsertAlias('REP', 'REPE'); InsertAlias('REP', 'REPZ');
  InsertMnemo('REPNE', k, [i386]);
  InsertOp(k, none, none, $0F2, none, none, none);
  InsertAlias('REPNE', 'REPNZ');

  InsertMnemo('RET', k, [i386]);
  InsertOp(k, none, none, $0C3, none, none, none);
  InsertOp(k, Imm16, none, $0C2, none, none, none);

  InsertMnemo('RDMSR', k, [Pentium, Privileged]);
  InsertOp(k, none, none, $0F, $32, none, none);

  InsertMnemo('RDPMC', k, [PentiumPro, Privileged]);
  InsertOp(k, none, none, $0F, $33, none, none);

  InsertMnemo('RDTSC', k, [Pentium]);
  InsertOp(k, none, none, $0F, $31, none, none);

  InsertMnemo('SAHF', k, [i386]);
  InsertOp(k, none, none, $09E, none, none, none);

  InsertMnemo('SCASB', k, [i386]);
  InsertOp(k, none, none, $0AE, none, none, none);
  InsertMnemo('SCASW', k, [i386]);
  InsertOp(k, none, none, $66, $0AF, none, none);
  InsertMnemo('SCASD', k, [i386]);
  InsertOp(k, none, none, $0AF, none, none, none);

  FOR i:=0 TO $0F DO BEGIN
    CASE i OF
      0  : InsertMnemo('SETO', k, [i386]);
      1  : InsertMnemo('SETNO', k, [i386]);
      2  : InsertMnemo('SETB', k, [i386]);
      3  : InsertMnemo('SETNB', k, [i386]);
      4  : InsertMnemo('SETZ', k, [i386]);
      5  : InsertMnemo('SETNZ', k, [i386]);
      6  : InsertMnemo('SETBE', k, [i386]);
      7  : InsertMnemo('SETNBE', k, [i386]);
      8  : InsertMnemo('SETS', k, [i386]);
      9  : InsertMnemo('SETNS', k, [i386]);
      $0A: InsertMnemo('SETP', k, [i386]);
      $0B: InsertMnemo('SETNP', k, [i386]);
      $0C: InsertMnemo('SETL', k, [i386]);
      $0D: InsertMnemo('SETNL', k, [i386]);
      $0E: InsertMnemo('SETLE', k, [i386]);
      $0F: InsertMnemo('SETNLE', k, [i386]);
    END;
    InsertOp(k, RM8, none, $0F, $90 + i, none, none);
  END;
  InsertAlias('SETB', 'SETC');
  InsertAlias('SETNB', 'SETAE');
  InsertAlias('SETZ', 'SETE');
  InsertAlias('SETNBE', 'SETA');
  InsertAlias('SETNL', 'SETGE');
  InsertAlias('SETNLE', 'SETG');
  InsertAlias('SETBE', 'SETNA');
  InsertAlias('SETB', 'SETNAE');
  InsertAlias('SETNB', 'SETNC');
  InsertAlias('SETNZ', 'SETNE');
  InsertAlias('SETLE', 'SETNG');
  InsertAlias('SETL', 'SETNGE');
  InsertAlias('SETP', 'SETPE');
  InsertAlias('SETNP', 'SETPO');

  InsertMnemo('SGDT', k, [i386]);
  InsertOp(k, M0, none, $0F, $01, 0, none);
  InsertMnemo('SIDT', k, [i386]);
  InsertOp(k, M0, none, $0F, $01, 1, none);

  InsertMnemo('SHLD', k, [i386]);
  InsertOp(k, RM16, Reg16, $0F, $0A4, none, OmRM, Imm8);
  InsertOp(k, RM32, Reg32, $0F, $0A4, none, OmRM, Imm8);
  InsertOp(k, RM16, Reg16, $0F, $0A5, none, OmRM, RegCL);
  InsertOp(k, RM32, Reg32, $0F, $0A5, none, OmRM, RegCL);

  InsertMnemo('SHRD', k, [i386]);
  InsertOp(k, RM16, Reg16, $0F, $0AC, none, OmRM, Imm8);
  InsertOp(k, RM32, Reg32, $0F, $0AC, none, OmRM, Imm8);
  InsertOp(k, RM16, Reg16, $0F, $0AD, none, OmRM, RegCL);
  InsertOp(k, RM32, Reg32, $0F, $0AD, none, OmRM, RegCL);

  InsertMnemo('SMSW', k, [i386]);
  InsertOp(k, RM16, none, $0F, $01, 4, none);

  InsertMnemo('SLDT', k, [i386]);
  InsertOp(k, RM16, none, $0F, $00, 0, none);

  InsertMnemo('STOSB', k, [i386]);
  InsertOp(k, none, none, $0AA, none, none, none);

  InsertMnemo('STOSW', k, [i386]);
  InsertOp(k, none, none, $066, $0AB, none, none);

  InsertMnemo('STOSD', k, [i386]);
  InsertOp(k, none, none,  $0AB, none, none, none);

  InsertMnemo('STR', k, [i386]);
  InsertOp(k, RM16, none,  $0F, $00, 1, none);

  InsertMnemo('TEST', k, [i386]);
  InsertOp(k, RegAL, Imm8,  $0A8, none, none, none);
  InsertOp(k, RegAX, Imm16,  $0A9, none, none, none);
  InsertOp(k, RegEAX, Imm32,  $0A9, none, none, none);
  InsertOp(k, RM8, Imm8,  $0F6, none, 0, none);
  InsertOp(k, RM16, Imm16,  $0F7, none, 0, none);
  InsertOp(k, RM32, Imm32,  $0F7, none, 0, none);
  InsertOp(k, RM8, Reg8,  $084, none, none, OmRM);
  InsertOp(k, RM16, Reg16,  $085, none, none, OmRM);
  InsertOp(k, RM32, Reg32,  $085, none, none, OmRM);

  InsertMnemo('VERR', k, [i386]);
  InsertOp(k, RM16, none,  $0F, $00, 4, none);

  InsertMnemo('VERW', k, [i386]);
  InsertOp(k, RM16, none,  $0F, $00, 5, none);

  InsertMnemo('WAIT', k, [i386, FPU]);
  InsertOp(k, none, none,  $09B, none, none, none);

  InsertMnemo('WBINVD', k, [i486, Privileged]);
  InsertOp(k, none, none,  $0F, $09, none, none);

  InsertMnemo('WRMSR', k, [Pentium, Privileged]);
  InsertOp(k, none, none, $0F, $30, none, none);

  InsertMnemo('XADD', k, [i486]);
  InsertOp(k, RM8, Reg8,  $0F, $0C0, none, OmRM);
  InsertOp(k, RM16, Reg16,  $0F, $0C1, none, OmRM);
  InsertOp(k, RM32, Reg32,  $0F, $0C1, none, OmRM);

  InsertMnemo('XLATB', k, [i386]);
  InsertOp(k, none, none,  $0D7, none, none, none);

  InsertMnemo('XCHG', k, [i386]);
  InsertOp(k, RegAX, Reg16, $090, none, none, OmOp);
  InsertOp(k, Reg16, RegAX, $090, none, none, OmOp);
  InsertOp(k, RegEAX, Reg32, $090, none, none, OmOp);
  InsertOp(k, Reg32, RegEAX, $090, none, none, OmOp);
  InsertOp(k, RM8, Reg8, $086, none, none, OmRM);
  InsertOp(k, Reg8, RM8, $086, none, none, OmRM);
  InsertOp(k, RM16, Reg16, $087, none, none, OmRM);
  InsertOp(k, Reg16, RM16, $087, none, none, OmRM);
  InsertOp(k, RM32, Reg32, $087, none, none, OmRM);
  InsertOp(k, Reg32, RM32, $087, none, none, OmRM);
END;
	
PROCEDURE InsertFloatInstr1;
VAR
  k: INTEGER;
BEGIN
		
  InsertMnemo('F2XM1', k, [FPU]);
  InsertOp(k, none, none, $0D9, $0F0, none, none);

  InsertMnemo('FABS', k, [FPU]);
  InsertOp(k, none, none, $0D9, $0E1, none, none);

  InsertMnemo('FADD', k, [FPU]);
  InsertOp(k, M32, none, $0D8, none, 0, none);
  InsertOp(k, M64, none, $0DC, none, 0, none);
  InsertOp(k, FRegST, FReg, $0D8, $0C0, none, OmOp);
  InsertOp(k, FReg, FRegST, $0DC, $0C0, none, OmOp);

  InsertMnemo('FADDP', k, [FPU]);
  InsertOp(k, FReg, FRegST, $0DE, $0C0, none, OmOp);
  InsertOp(k, none, none, $0DE, $0C1, none, none);

  InsertMnemo('FIADD', k, [FPU]);
  InsertOp(k, M32, none, $0DA, none, 0, none);
  InsertOp(k, M16, none, $0DE, none, 0, none);

  InsertMnemo('FBLD', k, [FPU]);
  InsertOp(k, M80, none, $0DF, none, 4, none);

  InsertMnemo('FBSTP', k, [FPU]);
  InsertOp(k, M80, none, $0DF, none, 6, none);

  InsertMnemo('FCHS', k, [FPU]);
  InsertOp(k, none, none, $0D9, $0E0, none, none);

  InsertMnemo('FNCLEX', k, [FPU]);
  InsertOp(k, none, none, $0DB, $0E2, none, none);

  InsertMnemo('FCOM', k, [FPU]);
  InsertOp(k, M32, none, $0D8, none, 2, none);
  InsertOp(k, M64, none, $0DC, none, 2, none);
  InsertOp(k, FReg, none, $0D8, $0D0, none, OmOp);
  InsertOp(k, none, none, $0DC, $0D1, none, none);

  InsertMnemo('FCOMP', k, [FPU]);
  InsertOp(k, M32, none, $0D8, none, 3, none);
  InsertOp(k, M64, none, $0DC, none, 3, none);
  InsertOp(k, FReg, none, $0D8, $0D8, none, OmOp);
  InsertOp(k, none, none, $0D8, $0D9, none, none);

  InsertMnemo('FCOMPP', k, [FPU]);
  InsertOp(k, none, none, $0DE, $0D9, none, none);

  InsertMnemo('FCOS', k, [FPU]);
  InsertOp(k, none, none, $0D9, $0FF, none, none);

  InsertMnemo('FDECSTP', k, [FPU]);
  InsertOp(k, none, none, $0D9, $0F6, none, none);

  InsertMnemo('FDIV', k, [FPU]);
  InsertOp(k, M32, none, $0D8, none, 6, none);
  InsertOp(k, M64, none, $0DC, none, 6, none);
  InsertOp(k, FRegST, FReg, $0D8, $0F0, none, OmOp);
  InsertOp(k, FReg, FRegST, $0DC, $0F8, none, OmOp);

  InsertMnemo('FDIVP', k, [FPU]);
  InsertOp(k, FReg, FRegST, $0DE, $0F8, none, OmOp);
  InsertOp(k, none, none, $0DE, $0F9, none, none);

  InsertMnemo('FIDIV', k, [FPU]);
  InsertOp(k, M32, none, $0DA, none, 6, none);
  InsertOp(k, M16, none, $0DE, none, 6, none);

  InsertMnemo('FDIVR', k, [FPU]);
  InsertOp(k, M32, none, $0D8, none, 7, none);
  InsertOp(k, M64, none, $0DC, none, 7, none);
  InsertOp(k, FRegST, FReg, $0D8, $0F8, none, OmOp);
  InsertOp(k, FReg, FRegST, $0DC, $0F0, none, OmOp);

  InsertMnemo('FDIVRP', k, [FPU]);
  InsertOp(k, FReg, FRegST, $0DE, $0F0, none, OmOp);
  InsertOp(k, none, none, $0DE, $0F1, none, none);

  InsertMnemo('FIDIVR', k, [FPU]);
  InsertOp(k, M32, none, $0DA, none, 7, none);
  InsertOp(k, M16, none, $0DE, none, 7, none);

  InsertMnemo('FFREE', k, [FPU]);
  InsertOp(k, none, none, $0DD, $0C0, none, OmOp);

  InsertMnemo('FICOM', k, [FPU]);
  InsertOp(k, M16, none, $0DE, none, 2, none);
  InsertOp(k, M32, none, $0DA, none, 2, none);

  InsertMnemo('FICOMP', k, [FPU]);
  InsertOp(k, M16, none, $0DE, none, 3, none);
  InsertOp(k, M32, none, $0DA, none, 3, none);

  InsertMnemo('FILD', k, [FPU]);
  InsertOp(k, M16, none, $0DF, none, 0, none);
  InsertOp(k, M32, none, $0DB, none, 0, none);
  InsertOp(k, M64, none, $0DF, none, 5, none);

  InsertMnemo('FINCSTP', k, [FPU]);
  InsertOp(k, none, none, $0D9, $0F7, none, none);

  InsertMnemo('FNINIT', k, [FPU]);
  InsertOp(k, none, none, $0DB, $0E3, none, none);

  InsertMnemo('FIST', k, [FPU]);
  InsertOp(k, M16, none, $0DF, none, 2, none);
  InsertOp(k, M32, none, $0DB, none, 2, none);

  InsertMnemo('FISTP', k, [FPU]);
  InsertOp(k, M16, none, $0DF, none, 3, none);
  InsertOp(k, M32, none, $0DB, none, 3, none);
  InsertOp(k, M64, none, $0DF, none, 7, none);

  InsertMnemo('FLD', k, [FPU]);
  InsertOp(k, M32, none, $0D9, none, 0, none);
  InsertOp(k, M64, none, $0DD, none, 0, none);
  InsertOp(k, M80, none, $0DB, none, 5, none);
  InsertOp(k, FReg, none, $0D9, $0C0, none, OmOp);

  InsertMnemo('FLD1', k, [FPU]);
  InsertOp(k, none, none, $0D9, $0E8, none, none);

  InsertMnemo('FLDL2T', k, [FPU]);
  InsertOp(k, none, none, $0D9, $0E9, none, none);

  InsertMnemo('FLDL2E', k, [FPU]);
  InsertOp(k, none, none, $0D9, $0EA, none, none);

  InsertMnemo('FLDPI', k, [FPU]);
  InsertOp(k, none, none, $0D9, $0EB, none, none);

  InsertMnemo('FLDLG2', k, [FPU]);
  InsertOp(k, none, none, $0D9, $0EC, none, none);

  InsertMnemo('FLDLN2', k, [FPU]);
  InsertOp(k, none, none, $0D9, $0ED, none, none);

  InsertMnemo('FLDZ', k, [FPU]);
  InsertOp(k, none, none, $0D9, $0EE, none, none);

  InsertMnemo('FLDCW', k, [FPU]);
  InsertOp(k, M0, none, $0D9, none, 5, none);

  InsertMnemo('FLDENV', k, [FPU]);
  InsertOp(k, M0, none, $0D9, none, 4, none);

  InsertMnemo('FMUL', k, [FPU]);
  InsertOp(k, M32, none, $0D8, none, 1, none);
  InsertOp(k, M64, none, $0DC, none, 1, none);
  InsertOp(k, FRegST, FReg, $0D8, $0C8, none, OmOp);
  InsertOp(k, FReg, FRegST, $0DC, $0C8, none, OmOp);

  InsertMnemo('FMULP', k, [FPU]);
  InsertOp(k, FReg, FRegST, $0DE, $0C8, none, OmOp);
  InsertOp(k, none, none, $0DE, $0C9, none, none);

  InsertMnemo('FIMUL', k, [FPU]);
  InsertOp(k, M32, none, $0DA, none, 1, none);
  InsertOp(k, M16, none, $0DE, none, 1, none);

  InsertMnemo('FNOP', k, [FPU]);
  InsertOp(k, none, none, $0D9, $0D0, none, none);

  InsertMnemo('FPATAN', k, [FPU]);
  InsertOp(k, none, none, $0D9, $0F3, none, none);

  InsertMnemo('FPREM', k, [FPU]);
  InsertOp(k, none, none, $0D9, $0F8, none, none);

  InsertMnemo('FPREM1', k, [FPU]);
  InsertOp(k, none, none, $0D9, $0F5, none, none);

  InsertMnemo('FPTAN', k, [FPU]);
  InsertOp(k, none, none, $0D9, $0F2, none, none);

  InsertMnemo('FRNDINT', k, [FPU]);
  InsertOp(k, none, none, $0D9, $0FC, none, none);

  InsertMnemo('FRSTOR', k, [FPU]);
  InsertOp(k, M0, none, $0DD, none, 4, none);

  InsertMnemo('FSAVE', k, [FPU]);
  InsertOp(k, M0, none, $09B, $0DD, 6, none);

  InsertMnemo('FNSAVE', k, [FPU]);
  InsertOp(k, M0, none, $0DD, none, 6, none);

  InsertMnemo('FSCALE', k, [FPU]);
  InsertOp(k, none, none, $0D9, $0FD, none, none);

  InsertMnemo('FSIN', k, [FPU]);
  InsertOp(k, none, none, $0D9, $0FE, none, none);

  InsertMnemo('FSINCOS', k, [FPU]);
  InsertOp(k, none, none, $0D9, $0FB, none, none);

  InsertMnemo('FSQRT', k, [FPU]);
  InsertOp(k, none, none, $0D9, $0FA, none, none);
END;
	
PROCEDURE InsertFloatInstr2;
VAR
  k: INTEGER;
BEGIN
  InsertMnemo('FST', k, [FPU]);
  InsertOp(k, M32, none, $0D9, none, 2, none);
  InsertOp(k, M64, none, $0DD, none, 2, none);
  InsertOp(k, FReg, none, $0DD, $0D0, none, OmOp);

  InsertMnemo('FSTP', k, [FPU]);
  InsertOp(k, M32, none, $0D9, none, 3, none);
  InsertOp(k, M64, none, $0DD, none, 3, none);
  InsertOp(k, M80, none, $0DB, none, 7, none);
  InsertOp(k, FReg, none, $0D9, $0D8, none, OmOp);

  InsertMnemo('FSTCW', k, [FPU]);
  InsertOp(k, M0, none, $09B, $0D9, 7, none);

  InsertMnemo('FNSTCW', k, [FPU]);
  InsertOp(k, M0, none, $0D9, none, 7, none);

  InsertMnemo('FSTENV', k, [FPU]);
  InsertOp(k, M0, none, $09B, $0D9, 6, none);

  InsertMnemo('FNSTENV', k, [FPU]);
  InsertOp(k, M0, none, $0D9, none, 6, none);

  InsertMnemo('FSTSW', k, [FPU]);
  InsertOp(k, M0, none, $09B, $0DD, 7, none);

  InsertMnemo('FNSTSW', k, [FPU]);
  InsertOp(k, M0, none, $0DD, none, 7, none);

  InsertMnemo('FSUB', k, [FPU]);
  InsertOp(k, M32, none, $0D8, none, 4, none);
  InsertOp(k, M64, none, $0DC, none, 4, none);
  InsertOp(k, FRegST, FReg, $0D8, $0E0, none, OmOp);
  InsertOp(k, FReg, FRegST, $0DC, $0E8, none, OmOp);

  InsertMnemo('FSUBP', k, [FPU]);
  InsertOp(k, FReg, FRegST, $0DE, $0E8, none, OmOp);
  InsertOp(k, none, none, $0DE, $0E9, none, none);

  InsertMnemo('FISUB', k, [FPU]);
  InsertOp(k, M32, none, $0DA, none, 4, none);
  InsertOp(k, M16, none, $0DE, none, 4, none);

  InsertMnemo('FSUBR', k, [FPU]);
  InsertOp(k, M32, none, $0D8, none, 5, none);
  InsertOp(k, M64, none, $0DC, none, 5, none);
  InsertOp(k, FRegST, FReg, $0D8, $0E8, none, OmOp);
  InsertOp(k, FReg, FRegST, $0DC, $0E0, none, OmOp);

  InsertMnemo('FSUBRP', k, [FPU]);
  InsertOp(k, FReg, FRegST, $0DE, $0E0, none, OmOp);
  InsertOp(k, none, none, $0DE, $0E1, none, none);

  InsertMnemo('FISUBR', k, [FPU]);
  InsertOp(k, M32, none, $0DA, none, 5, none);
  InsertOp(k, M16, none, $0DE, none, 5, none);

  InsertMnemo('FTST', k, [FPU]);
  InsertOp(k, none, none, $0D9, $0E4, none, none);

  InsertMnemo('FUCOM', k, [FPU]);
  InsertOp(k, FReg, none, $0DD, $0E0, none, OmOp);
  InsertOp(k, none, none, $0DD, $0E1, none, none);

  InsertMnemo('FUCOMP', k, [FPU]);
  InsertOp(k, FReg, none, $0DD, $0E8, none, OmOp);
  InsertOp(k, none, none, $0DD, $0E9, none, none);

  InsertMnemo('FUCOMPP', k, [FPU]);
  InsertOp(k, none, none, $0DA, $0E9, none, none);

  InsertMnemo('FWAIT', k, [FPU]);
  InsertOp(k, none, none, $09B, none, none, none);

  InsertMnemo('FXAM', k, [FPU]);
  InsertOp(k, none, none, $0D9, $0E5, none, none);

  InsertMnemo('FXCH', k, [FPU]);
  InsertOp(k, FReg, none, $0D9, $0C8, none, OmOp);
  InsertOp(k, none, none, $0D9, $0C9, none, none);

  InsertMnemo('FXTRACT', k, [FPU]);
  InsertOp(k, none, none, $0D9, $0F4, none, none);

  InsertMnemo('FYL2X', k, [FPU]);
  InsertOp(k, none, none, $0D9, $0F1, none, none);

  InsertMnemo('FYL2XP1', k, [FPU]);
  InsertOp(k, none, none, $0D9, $0F9, none, none);
END;

PROCEDURE DataInsert;
VAR k: INTEGER;
BEGIN
	InstrTabLen := 313;
  OpTabLen := 670;
	OpTabN := 0;
  Average := 0;
  InstrCount := 0;
	SETLENGTH(InstrTab, InstrTabLen);
  SETLENGTH(OpTab, OpTabLen);
	FOR k:= 0 TO OpTabLen-1 DO OpTab[k].op0 := -1;
	FOR k:= 0 TO InstrTabLen-1 DO InstrTab[k].name := '';
	FOR k := 0 TO LENGTH(Clash)-1 DO  Clash[k] := 0;
	InsertInstr1;
  InsertInstr2;
  InsertInstr3;
  InsertInstr4;
  InsertFloatInstr1;
  InsertFloatInstr2;
END;
	
BEGIN

END.
