UNIT OPM;  (** non-portable *)

INTERFACE

USES
  Windows, Classes, SysUtils, OER, _SYSTEM;

TYPE
  TModList = CLASS
  PUBLIC
    count       : integer;
    Text        : STRING;
    CONSTRUCTOR Create;
    DESTRUCTOR  Destroy; OVERRIDE;
    PROCEDURE   Clear;
    FUNCTION    Find(Word: STRING; VAR N: integer): boolean;
    FUNCTION    FindNum(N: integer): STRING;
    PROCEDURE   Add(Word: STRING);
  END;

CONST
  ByteSize     = 1;  (* SYSTEM.BYTE *)
  CharSize     = 1;  (* CHAR *)
  BoolSize     = 1;  (* BOOLEAN *)
  SetSize      = 4;  (* SET *)
  SIntSize     = 1;  (* SHORTINT *)
  IntSize      = 2;  (* INTEGER *)
  LIntSize     = 4;  (* LONGINT *)
  HIntSize     = 8;  (* HUGEINT *)
  RealSize     = 4;  (* REAL *)
  LRealSize    = 8;  (* DOUBLE *)
  ProcSize     = 4;  (* PROCEDURE type *)
  PointerSize  = 4;  (* POINTER type *)
  
  (* value of constant NIL *)
  nilval = 0;
  
  (* maximal index value for array declaration: *)
  MaxIndex = MaxLInt;
  
  (* parametrization of numeric scanner: *)
  MaxHDig      = 8;  (* maximal hexadecimal longint length *)
  MaxRExp      = 38;  (* maximal real exponent *)
  MaxLExp      = 308;  (* maximal DOUBLE exponent *)
  
  (* inclusive range of parameter of standard procedure HALT: *)
  MinHaltNr    = 20;
  MaxHaltNr    = MaxLInt;
  
  (* inclusive range of register number of procedures SYSTEM.GETREG and SYSTEM.PUTREG: *)
  MinRegNr     = 0;  (* interpretation is left to the back-end, e.g. for i386: *)
  MaxRegNr     = 31;  (* F0-F7 if second parameter is REAL or DOUBLE, R0-R7 else *)
  
  (* maximal value of flag used to mark interface structures: *)
  MaxSysFlag   = 0;  (* i386: only 0 is valid, not used *)
  
  (* maximal condition value of parameter of SYSTEM.CC: *)
  MaxCC        = -1;  (* not implemented interpretation is left to the back-end *)
  
  (* initialization of linkadr field in ObjDesc, must be different from any valid link address: *)
  LANotAlloc   = -1;
  
  (* initialization of constant address, must be different from any valid constant address: *)
  ConstNotAlloc = -1;  (* i386: only strings are allocated *)
  
  (* initialization of tdadr field in StrDesc, must be different from any valid address: *)
  TDAdrUndef = -1;
  
  (* maximal number of cases in a case statement: *)
  MaxCases = 128;
  
  (* maximal range of a case statement (higher label - lower label ~ jump table size): *)
  MaxCaseRange = 512;
  
  (* whether field leaf of pointer variable p has to be set to FALSE, when NEW(p) or SYSTEM.NEW(p, n) is used: *)
  
  NEWusingAdr = FALSE;
  
  (* mask for the TD associated with an object instance: mask away GC information *)
  TDMask = 0; (* 0FFFFFFF8H*)
  
  (*Configuration -- temp *)
  OptimizeSelf = FALSE;  (* if possible, make self pointer-based in methods *)
  
  (* special character (< " ") returned by procedure Get, if end of text reached *)
  Eot      = #0;
  SetLen    = 31+1;
  NumErrors = (912+SetLen) DIV SetLen * SetLen;
  
  (** Compiler Options *)
  inxchk         = 0;  (** index check *)
  ovflchk        = 1;  (** overflow check *)
  typchk         = 3;  (** type checking *)
  ptrinit        = 5;  (** pointer initialization *)
  oassert        = 6;  (** assert evaluation *)
  findpc         = 7;  (** find text position of breakpc *)
  systemchk      = 8;  (** prevent compilation of modules that import SYSTEM *) (* jm *)
  nofiles        = 9;  (** don't create any files (dVdW) *)
  newsf          = 10;  (** generation of new symbol file allowed *)
  extsf          = 11;  (** generation of extended symbol file allowed *)
  warning        = 12;  (** display warnings *)
  prefix         = 18;  (** X - retain filename prefix on .Sym & .Obj file (Native) *)
  oberon2        = 19;   (** 2 - compile oberon-2 programs *)
  fullstackinit  = 20;  (** zero all the stack (only if ptrinit)*)
  oberon1        = 21;  (** compile oberon-1 programs *)
  trace          = 22;  (** trace procedure usage *)
  aos            = 23;  (** aos, different SymFile *)
  
  DefOpt = [typchk,inxchk,oassert,fullstackinit,ptrinit,nofiles,warning];

VAR
  MinReal, MaxReal   : SINGLE;
  MinLReal, MaxLReal : DOUBLE;
  noerr              : BOOLEAN;  (* no error found until now *)
  curpos, errpos     : OER.TErrorPos;  (* character and error position in source file *)
  charpos, strpos    : LONGINT;
  breakpos           : LONGINT;  (* set by Mark *)
  breakpc            : LONGINT;  (*set by OPV.Init*)
  errors             : ARRAY [0..NumErrors DIV SetLen] OF TSET;
  outputPath         : STRING;  (* for reading and writing obj files. must include any trailing "/" or ":" *)
  outputPrefix       : ARRAY [0..32] OF CHAR;  (* for writing obj files (in addition to outputPath). *)
  options            : TSET;
  errName            : STRING;
  first              : BOOLEAN;
  AllocFC            : BOOLEAN;

  RefStr             : STRING;

CONST
  OFtag       = #$0BB;  (* object file tag *)
  OFnative    = #$0AF;
  OFportable  = #$088;
  OFaos       = #$01;

  RefTag      = #$8C;
  NewRef      = TRUE;

  extension   = '.Obj';
  moduleext   = '.Mod';

VAR
  modulename: STRING;
  ConNoErr  : BOOLEAN;

  LRealPat: RECORD L, H: LONGINT END ;
  lastpos, pat: LONGINT;  (* last position error in source file *)
  refpatch: LONGINT;  (* OM: last position error in source file *)
  ObjF,SymF,HdrF,ModF    : TMemoryStream;

  PROCEDURE Init(opt: TSET);
  PROCEDURE err(n: INTEGER);
  PROCEDURE Mark(n: INTEGER; pos: TErrorPos);
  PROCEDURE FPrint(VAR fp: LONGINT; val: LONGINT);
  PROCEDURE FPrintSet(VAR fp: LONGINT; Aset: TSET);
  PROCEDURE FPrintReal(VAR fp: LONGINT; real: REAL);
  PROCEDURE FPrintLReal(VAR fp: LONGINT; lr: DOUBLE);

  PROCEDURE LogW(ch: CHAR);
  PROCEDURE LogWStr(CONST s: STRING);
  PROCEDURE LogWStrLn(CONST s: STRING);
  PROCEDURE LogWNum(i, len: LONGINT);
  PROCEDURE LogWHex(i: LONGINT);
  PROCEDURE LogWLn;
  PROCEDURE LogClose;

  PROCEDURE OpenSym(F: TMemoryStream; name: STRING; self: BOOLEAN; VAR done: BOOLEAN);
  PROCEDURE SymWMod(F: TMemoryStream; name: STRING);
  PROCEDURE SymW(F: TMemoryStream; b: CHAR);
  PROCEDURE SymWString(F: TMemoryStream; VAR s: STRING);
  PROCEDURE SymWNum(F: TMemoryStream; x: INT64);
  PROCEDURE SymWInt(F: TMemoryStream; k: LONGINT);
  PROCEDURE SymWLInt(F: TMemoryStream; k: LONGINT);
  PROCEDURE SymWBytes(F: TMemoryStream; x: INTEGER; len: INTEGER);
  PROCEDURE SymWSet(F: TMemoryStream; s: TSET);
  PROCEDURE SymWCh(F: TMemoryStream; ch: CHAR);
  PROCEDURE SymWReal(F: TMemoryStream; r: REAL);
  PROCEDURE SymWLReal(F: TMemoryStream; lr: DOUBLE);

  PROCEDURE ReadInt(VAR F: TMemoryStream; VAR x: INTEGER);
  PROCEDURE ReadSet(VAR F: TMemoryStream; VAR x: TSET);
  PROCEDURE ReadString(VAR F: TMemoryStream; VAR S: STRING);
  PROCEDURE ReadStr(VAR F: TMemoryStream; VAR x: STRING);
  PROCEDURE ReadNum(VAR F: TMemoryStream; VAR x: LONGINT);

  PROCEDURE FNBegin(VAR modName: STRING);
  PROCEDURE EndSym(VAR modName: STRING);
  PROCEDURE CloseObj(modName: STRING);

  PROCEDURE OPMBegin;

IMPLEMENTATION

USES
  OPT;

CONSTRUCTOR TModList.Create;
BEGIN
  INHERITED Create;
  count:=0;
  Text:='';
END;

DESTRUCTOR TModList.Destroy;
BEGIN
  Text:='';
  INHERITED Destroy;  
END;

PROCEDURE TModList.Clear;
BEGIN
  Text:='';
  count:=0;
END;

FUNCTION TModList.FindNum(N: integer): STRING;
var
  i: integer;
  S: STRING;
  P: integer;
BEGIN
  Result:='';
  i:=1;
  S:='';
  P:=-1;
  IF N>=COUNT THEN EXIT;
  WHILE (i<=Length(Text)-2) AND (P<N) DO BEGIN
    S:=S+Text[i];
    IF (Text[i+1]=#13) AND (Text[i+2]=#10) THEN BEGIN
      Result:=S;
      Inc(P);
      S:='';
      Inc(i,2);
    END;
    Inc(i);
  END;
END;

FUNCTION TModList.Find(Word: STRING; VAR N: integer): boolean;
var
  i: integer;
  S: STRING;
BEGIN
  Result:=False;
  N:=0;
  i:=1;
  S:='';
  WHILE (i<=Length(Text)-2) AND NOT Result DO BEGIN
    S:=S+Text[i];
    IF (Text[i+1]=#13) AND (Text[i+2]=#10) THEN BEGIN
      IF SameText(Word,S) THEN Result:=true;
      S:='';
      Inc(i,2);
    END;
    Inc(i);
  END;
  N:=i;
END;

PROCEDURE TModList.Add(Word: STRING);
VAR
  n: integer;
BEGIN
  IF Find(Word,n) THEN Exit;
  Text:=Text+Word+#13#10;
  Inc(count);
END;

//######################################################################

PROCEDURE Init(opt: TSET);
BEGIN
  options := opt;
  noerr   := TRUE;
  RefStr:='';
  //Kernel.GC()
END;

PROCEDURE Close;
BEGIN

END;

PROCEDURE Get(VAR ch: CHAR);  (* read next character from source text, Eot if no more *)
BEGIN

END;

PROCEDURE MakeFileName(VAR path, prefix, modName, fileName: ARRAY OF CHAR; ext: ARRAY OF CHAR);
BEGIN

END;

(* ------------------------- Log Output ------------------------- *)

PROCEDURE AllocCnsl;
CONST
  YellowOnBlue  = FOREGROUND_GREEN OR FOREGROUND_RED OR
                    FOREGROUND_INTENSITY OR BACKGROUND_BLUE;
  WhiteOnBlue   =   FOREGROUND_BLUE OR FOREGROUND_GREEN OR
                  FOREGROUND_RED OR FOREGROUND_INTENSITY OR
                    BACKGROUND_BLUE;
  RedOnWhite    = FOREGROUND_RED OR FOREGROUND_INTENSITY OR
                    BACKGROUND_RED OR BACKGROUND_GREEN OR BACKGROUND_BLUE
                    OR BACKGROUND_INTENSITY;
  WhiteOnRed    = BACKGROUND_RED OR BACKGROUND_INTENSITY OR
                    FOREGROUND_RED OR FOREGROUND_GREEN OR FOREGROUND_BLUE
                    OR FOREGROUND_INTENSITY;
  RedOnBlack    = FOREGROUND_RED OR FOREGROUND_INTENSITY;

VAR
  Coord     : TCoord;
  dwSize, dwCount: DWORD;
  ConHandle : THandle;
  Clr       : BYTE;
BEGIN
  IF NOT AllocFC THEN BEGIN
    AllocConsole;
    SetConsoleTitle(PChar('Debug info!!!'));
  END;

  ConHandle:=GetStdHandle(STD_OUTPUT_HANDLE);
  Coord := GetLargestConsoleWindowSize(ConHandle);
  dwSize := Coord.X*Coord.Y*80;
  Coord.X := 0;
  Coord.Y := 0;
  IF OPM.noerr AND ConNoErr THEN
    Clr:=WhiteOnBlue
  ELSE BEGIN
    Clr:=WhiteOnRed;
    ConNoErr:=FALSE;
  END;
  SetConsoleTextAttribute(ConHandle, Clr);
  FillConsoleOutputAttribute(ConHandle, Clr, dwSize, Coord, dwCount);
  AllocFC:=TRUE;
END;

PROCEDURE LogW(ch: CHAR);
BEGIN
  AllocCnsl;
  Write(ch);
END;

PROCEDURE LogWStr(CONST s: STRING);
BEGIN
  AllocCnsl;
  Write(s);
END;

PROCEDURE LogWStrLn(CONST s: STRING);
BEGIN
  AllocCnsl;
  Write(s);
  WriteLn;
END;

PROCEDURE LogWNum(i, len: LONGINT);
BEGIN
  AllocCnsl;
  Write(i: len);
END;

PROCEDURE LogWHex(i: LONGINT);
BEGIN
  AllocCnsl;
  Write(IntToHex(i,8));
  Write('H');
END;

PROCEDURE LogWLn;
BEGIN
  AllocCnsl;
  WriteLn;
END;

PROCEDURE LogClose;
BEGIN
  FreeConsole;
  AllocFC:=FALSE;
END;

(* ----- Fingerprint calculation ----- *)

PROCEDURE FPrint(VAR fp: LONGINT; val: LONGINT);
BEGIN
  fp:=INTEGER(TSET(_SYSTEM.ROT(fp,7))) XOR INTEGER(TSET(val));
END;

PROCEDURE FPrintSet(VAR fp: LONGINT; Aset: TSET);
BEGIN
  FPrint(fp,LONGINT(Aset));
END;

PROCEDURE FPrintReal(VAR fp: LONGINT; real: REAL);
BEGIN
  FPrint(fp, RealToIEEE(real))
END;

PROCEDURE FPrintLReal(VAR fp: LONGINT; lr: DOUBLE);
BEGIN
  FPrint(fp, DoubleToIEEE(lr));
END;

(* ----- auxiliaries ----- *)

(* ------------------------- Log Output ------------------------- *)
PROCEDURE ErrorMsg(n: INTEGER);
BEGIN

END;

PROCEDURE Mark(n: INTEGER; pos: TErrorPos);
BEGIN
  IF n >= 0 THEN noerr := FALSE;
  OER.Errors(n,pos);
END;

PROCEDURE err(n: INTEGER);
BEGIN
  Mark(n, errpos)
END;

(* ------------------------- Write Symbol File ------------------------- *)

PROCEDURE SymB(F: TMemoryStream; b: BYTE);
BEGIN
  F.Write(b,1);
END;

PROCEDURE SymW(F: TMemoryStream; b: CHAR);
BEGIN
  F.Write(b,1);
END;

PROCEDURE SymWCh(F: TMemoryStream; ch: CHAR);
BEGIN
  F.Write(ch,1)
END;

PROCEDURE SymWTag(F: TMemoryStream; k: INTEGER);
VAR
  Ch: CHAR;
BEGIN
  Ch:=CHR(k);
  F.Write(Ch,1);
END;

PROCEDURE SymWInt(F: TMemoryStream; k: LONGINT);
BEGIN
  F.Write(k,2);
END;

PROCEDURE SymWLInt(F: TMemoryStream; k: LONGINT);
BEGIN
  F.Write(k,4);
END;

PROCEDURE SymWNum(F: TMemoryStream; x: INT64);
BEGIN
  WHILE (x<-64) OR (x>63) DO BEGIN
    SymB(F,_SYSTEM._MOD(x,$80)+$80);
    x := _SYSTEM._DIV(x,$80);
  END;
  SymB(F,_SYSTEM._MOD(x,$80));
END;

FUNCTION SymWPos(F: TMemoryStream):LONGINT;
BEGIN
  RESULT:= F.Position;
END;

PROCEDURE SymWStringByte(VAR F: TMemoryStream; s: STRING; LEN: INTEGER);
VAR
  i: INTEGER;
BEGIN
  I:=1;
  WHILE i<LEN DO BEGIN SymWCh(F,S[i]); INC(i); END;
END;

PROCEDURE SymWString(F: TMemoryStream; VAR s: STRING);
VAR
  i: INTEGER;
  ch: CHAR;
BEGIN
  i:=1;
  ch:=s[i];
  IF  ch=#0  THEN BEGIN
    SymWCh(F, #0);
    EXIT;
  END;
  WHILE (ch<>#0) AND (ch<#$7F) DO BEGIN
    INC(i);
    ch:=s[i];
  END;
  IF i>1 THEN SymWStringByte(F,s,i-1);
  IF ch=#0 THEN
    SymWCh(F,CHR(ORD(s[i-1])+$80))
  ELSE BEGIN
    IF  i>1  THEN SymWCh(F, s[i-1]);
    SymWCh(F,#$7F);
    REPEAT
      SymWCh(F,ch);
      INC(i);
      ch:=s[i]
    UNTIL ch=#0;
    SymWCh(F,#0);
  END;
END;

PROCEDURE SymWMod(F: TMemoryStream; name: STRING);
BEGIN
  SymWString(F,name);
END;

PROCEDURE SymWSet(F: TMemoryStream; s: TSET);
BEGIN
  SymWNum(F,LONGINT(s));
END;

PROCEDURE SymWBytes(F: TMemoryStream; x: INTEGER; len: INTEGER);
VAR
  i: INTEGER;
BEGIN
  i:=0;
  WHILE i < len DO BEGIN
    SymWCh(F,CHR(LO(x)));
    x := x SHR 8;
    INC(i)
  END;
END;

PROCEDURE SymWLBytes(F: TMemoryStream; x: Int64; len: INTEGER);
VAR
  i: INTEGER;
BEGIN
  i:=0;
  WHILE i < len DIV 2 DO BEGIN
    SymWCh(F,CHR(LO(x)));
    SymWCh(F,CHR(HI(x)));
    x := x shr 16;
    INC(i)
  END;
END;

PROCEDURE SymWReal(F: TMemoryStream; r: REAL);
VAR
  tr: INTEGER;
BEGIN
  tr:= RealToIEEE(r);
  SymWBytes(F,tr, RealSize)
END;

PROCEDURE SymWLReal(F: TMemoryStream; lr: DOUBLE);
VAR
  tr: Int64;
BEGIN
  tr:= DoubleToIEEE(lr);
  SymWLBytes(F,tr, LRealSize)
END;

(*-------------------------------------------------------------------*)

PROCEDURE ReadInt(VAR F: TMemoryStream; VAR x: INTEGER);
BEGIN
  F.Read(x,2);
END;

PROCEDURE ReadSet(VAR F: TMemoryStream; VAR x: TSET);
BEGIN
  F.Read(INTEGER(x),4);
END;

PROCEDURE ReadNum(VAR F: TMemoryStream; VAR x: LONGINT);
VAR
  n: LONGINT;
  s: SHORTINT;
  ch: CHAR;
BEGIN
  s := 0;
  n := 0;
  F.Read(ch,1);
  WHILE ORD(ch) >= $80 DO BEGIN
    INC(n, ASH(ORD(ch) - $80, s) );
    INC(s, 7);
    F.Read(ch,1)
  END;
  x := n + ASH(ORD(ch) MOD 64 - ORD(ch) DIV 64 * 64, s);
END;

PROCEDURE ReadStr(VAR F: TMemoryStream; VAR x: STRING);
VAR
  ch: CHAR;
BEGIN
  x:='';
  REPEAT
    f.Read(ch,1);
    x:=x+ch;
  UNTIL ch = #0;
  IF x=#0 THEN x:='';
END;

PROCEDURE ReadString(VAR F: TMemoryStream; VAR S: STRING);
VAR
  ch: char;
BEGIN
  S:='';
  REPEAT
    F.Read(ch,1);
    IF ch = #0 THEN BEGIN
      S:='';
      EXIT;
    END ELSE IF ch < #$7F THEN
      S:=S+ch
    ELSE IF ch > #$7F THEN BEGIN
      S:=S+CHR(ORD(ch)-$80);
      EXIT;
    END ELSE
      BREAK;
  UNTIL FALSE;
  REPEAT
    F.Read(ch,1);
    IF ch = #0 THEN
      EXIT
    ELSE
      S:=S+ch;
  UNTIL FALSE;
END;

(* ------------------------- Write Reference & Object Files ------------------------- *)

(* ----- files ----- *)

PROCEDURE FNBegin(VAR modName: STRING);
VAR
  BUFF      : CHAR;
BEGIN
  HdrF := TMemoryStream.Create;
  HdrF.Position:=0;
  BUFF:=OFtag;
  HdrF.Write(BUFF, 1);
  ModF := TMemoryStream.Create;
  ModF.Position:=0;
  SymF := TMemoryStream.Create;
  SymF.Position:=0;
  ObjF := TMemoryStream.Create;
  ObjF.Position:=0;
  IF aos IN options THEN BEGIN
    BUFF:=OFtag;
    HdrF.Write(BUFF, 1);
  END ELSE BEGIN
    BUFF:=OFnative;
    HdrF.Write(BUFF, 1);
  END;
END;

PROCEDURE EndSym(VAR modName: STRING);
VAR
  sfsize  : LONGINT;
  BUFF    : CHAR;
BEGIN
  BUFF:=#0;
  ModF.Write(BUFF,1);
  sfsize:=SymF.Size+ModF.Size;
  SymWNum(HdrF,sfsize);
  HdrF.Write(ModF.Memory^,ModF.Size);
  SymWInt(ObjF,$00); //RefSize 2 byte
  SymWInt(ObjF,$00); //RefSize 2 byte
  HdrF.Write(SymF.Memory^,SymF.Size);
END;

PROCEDURE CloseObj(modName: STRING);
BEGIN (*ref block*)
  ObjF.Write(RefStr[1],LENGTH(RefStr));
  ObjF.Position:=0;
  SymWLInt(ObjF,LENGTH(RefStr));
  HdrF.Write(ObjF.Memory^,ObjF.Size);
  HdrF.Position:=0;
  HdrF.SaveToFile(modName+extension);
  ModF.Free;
  SymF.Free;
  ObjF.Free;
  HdrF.Free;
  RefStr:='';
END;

PROCEDURE OpenSym(F: TMemoryStream; name: STRING; self: BOOLEAN; VAR done: BOOLEAN);
VAR
  ch, ch1: CHAR;
  symSize: LONGINT;
BEGIN
  done := FALSE;
  name:=outputPath+name+extension;
  IF FileExists(Name) THEN F.LoadFromFile(Name);
  IF F.Size<>0 THEN BEGIN
    done := TRUE;
    F.Read(ch,1);
    IF ch=OFtag THEN BEGIN
      F.Read(ch1,1);
      ReadNum(F,symSize);
      IF ((aos IN options) AND (ch1<>OFaos))
      OR (NOT(aos IN options) AND (ch1<>OFnative)) THEN BEGIN
        err(151);
        done := FALSE
      END
    END ELSE
      done := FALSE;
  END;
END;

PROCEDURE OPMBegin;
BEGIN
  pat         := MinRealPat;
  MOVE(pat,MinReal,4);	      (*i386: -3.40282346E38*)
	pat         := MaxRealPat;
  MOVE(pat,MaxReal,4);	      (*i386: 3.40282346E38*)  
	LRealPat.L  := MinLRealPatL;
  LRealPat.H  := MinLRealPatH;
	MOVE(LRealPat,MinLReal,8);	(*i386: -1.7976931348623157D308*)
	LRealPat.L  := MaxLRealPatL;
  LRealPat.H  := MaxLRealPatH;
	MOVE(LRealPat,MaxLReal,8);	(*i386: 1.7976931348623157D308*)
	first       := TRUE;
  noerr       := TRUE;
  AllocFC     := FALSE;
  (*
  Available options are:
  f,F - prevent/allow creation of files
  n,N - prevent/allow generation of new Symbol table
  e,E - prevent/allow extension of symbol table
  s,S - prevent/allow compilation of modules importing SYSTEM
  w,W - hide/display warning messages

  p,P - don't/do initialize pointers to NIL
  z,Z - don't/do zero all the stack (only if ptrinit)
  a,A - don't/do evaluate ASSERT functions
  x,X - don't/do check indexes
  t,T - don't/do check types
  v,V - don't/do check overflows

  3 - compile Active Oberon
  2 - compile Oberon-2
  1 - compile Oberon-1

  Default options are FnesWPZAXTv3

  "f" nofiles
  "n" newsf
  "e" extsf
  "s" systemchk
  "w" warning
  "p" ptrinit
  "z" fullstackinit
  "a" assert
  "x" inxchk
  "t" typchk
  "v" ovflchk
  "3" -oberon2,-oberon1
  "2" oberon2,-oberon1
  "1" oberon1,-oberon2
  *)
END;

END.











