UNIT OPV;

INTERFACE

USES
  OPS, OPT, OPL, OPM, OER, OPOi386, _SYSTEM, Classes, SysUtils;

CONST
  OAbs = OPT.omAbs;

  (* Arrays extlev *)
  static = 0; dynamic = 1; sdynamic = 2;

  MaxAdr 	= MaxLInt;
  MinAdr 	= MinLInt;
  VarParSize 	= OPM.PointerSize;
  RecVarParSize 	= 8; 	(* push ADR (rec); push TAG (reg) *)
  LProcOffSL 	= 12; 	(* parameter offset with static link *)
  LProcOff 	= 8; 	(* parameter offset without static link *)
  ExtProcOff 	= 8; 	(* parameter offset of external procedures *)

  ArraySet = [csStaticArr, csSDynArr, csDynArr, csOpenArr];

TYPE
  ParamAdrProc = PROCEDURE (VAR firstPar: Objects; VAR psize, dsize: LONGINT; TProc, vInline, allocDesc: BOOLEAN);

VAR
  ProcName: STRING; (* accessed by OPI (read only) *)
  EntryNr: INTEGER; (* accessed by OPI (read only) *)
  allocTd, bassert, findpc: BOOLEAN; 	(* allocTd used only by OM *)
  ParamAdr: ARRAY [0..4] OF  ParamAdrProc;

  PROCEDURE Init(opt: TSET; bpc: LONGINT);
  PROCEDURE AdrAndSize(topScope: OPT.Objects);
  PROCEDURE OPVBegin;
  
IMPLEMENTATION

PROCEDURE Init(opt: TSET; bpc: LONGINT);
CONST
  ass = 6;
  fpc = 7;
BEGIN
  bassert := ass IN opt;
  findpc  := fpc IN opt;
  allocTd := TRUE;			(* allocTd used only by OM *)
  IF findpc THEN
    OPM.breakpc := bpc
  ELSE
    OPM.breakpc := MAX(_LONGINTS);
END;

PROCEDURE IncAdr(VAR adr: LONGINT; s: LONGINT);
BEGIN
  IF (s >= 0) AND (adr <= MaxAdr - s) OR (s < 0) AND (adr >= MinAdr - s) THEN
    INC(adr, s)
  ELSE
    OPM.err(242)
END;

PROCEDURE AlignFld(VAR offset: LONGINT; elemsize: LONGINT);
BEGIN
  IF elemsize >= 4 THEN
    IncAdr(offset,_SYSTEM._MOD(-offset,4))
  ELSE IF elemsize = 2 THEN
    IncAdr(offset,_SYSTEM._MOD(offset,2));
END;

PROCEDURE AllocFld(VAR offset, this: LONGINT; elemsize: LONGINT);
BEGIN
  this := offset;
  IncAdr(offset, elemsize)
END;
	
PROCEDURE AlignRec(VAR size: LONGINT);
BEGIN 
  IncAdr(size,_SYSTEM._MOD(-size,4));
END;
	
PROCEDURE AlignPar(VAR elemsize: LONGINT);
BEGIN
  IncAdr(elemsize, _SYSTEM._MOD(-elemsize,4));
END;
	
PROCEDURE AllocPar (VAR adr, this: LONGINT; elemsize: LONGINT);
BEGIN
  IncAdr(adr, elemsize);
  this := adr
END;

PROCEDURE AlignVar(VAR adr: LONGINT; elemsize: LONGINT);
BEGIN
  IF elemsize >= 4 THEN IncAdr(adr, -_SYSTEM._MOD(adr,4))
  ELSE IF elemsize = 2 THEN IncAdr(adr, -_SYSTEM._MOD(adr,2));
END;
	
PROCEDURE AllocVar(VAR adr, this: LONGINT; elemsize: LONGINT);
BEGIN
  IncAdr(adr, -elemsize);
  this := adr
END;

PROCEDURE AlignBlock(adr: LONGINT; VAR dsize: LONGINT);
BEGIN 
  IncAdr(adr,-_SYSTEM._MOD(adr,4));
  dsize := -adr;
END;

FUNCTION RecursionTest (typ: OPT.Structs; mode: LONGINT): BOOLEAN;
VAR
  p: OPT.Objects;
BEGIN
  IF typ <> NIL THEN BEGIN
    IF typ.rectest < 0 THEN BEGIN
      RESULT:=FALSE;
      EXIT;
    END ELSE IF (typ.rectest < mode) OR (typ.rectest > 5) THEN BEGIN
      (* traverse btyp *)
      IF (typ.form = sfComp) THEN BEGIN
        typ.rectest := -typ.rectest;
        IF NOT RecursionTest (typ.BaseTyp, mode) THEN BEGIN
          RESULT:=FALSE;
          EXIT;
        END;
        IF typ.comp = csRecord THEN BEGIN
          p := typ.link;
          WHILE (p <> NIL) AND (p.mode = omFld) DO BEGIN
            IF NOT RecursionTest (p.typ, 2) THEN BEGIN
              OPM.errpos.CodePos := p.conval^.intval;
              OPM.err(58);
              p.typ := OPT.undftyp
            END;
            p := p.link
          END;
        END;
      END;
    END;
    typ.rectest := mode;
  END;
  RESULT:=TRUE;
END;

PROCEDURE Traverse(obj: Objects; exported: BOOLEAN); FORWARD;
PROCEDURE VisitTProcs(obj: Objects); FORWARD;

PROCEDURE TypSize(typ: OPT.Structs; allocDesc: BOOLEAN);
VAR
  f, c: INTEGER;
  offset, size, dummy: LONGINT;
  fld: OPT.Objects;
  btyp, ftyp: OPT.Structs;
  sizeUndef, doAlloc: BOOLEAN;		(* ftyp, OM only *)
BEGIN
  IF typ = OPT.undftyp THEN
    OPM.err(58)		(* MH 26.1.94; bug fix RC 17.6.93 *)
  ELSE BEGIN
    sizeUndef := typ.size = -1;
    IF NOT RecursionTest (typ, 1) THEN BEGIN
      OPM.errpos := typ.txtpos;
      OPM.err(58);
      typ.BaseTyp := OPT.undftyp
    END;
    doAlloc := allocDesc AND (typ.tdadr = OPM.TDAdrUndef) AND (typ.offset = OPM.TDAdrUndef);
    IF sizeUndef OR doAlloc THEN BEGIN
      IF doAlloc THEN typ.tdadr := -2; (* avoid cycles *)
      f := typ.form;
      c := typ.comp;
      btyp := typ.BaseTyp;
      IF c = csRecord THEN BEGIN
        IF btyp = NIL THEN
          offset := 0
        ELSE BEGIN
          TypSize(btyp, allocDesc);
          offset := btyp.size;
        END;
        fld := typ.link;
        WHILE (fld <> NIL) DO BEGIN
          IF fld.mode = omFld THEN BEGIN
            ftyp := fld.typ;
            TypSize(ftyp, allocDesc);
            IF sizeUndef THEN BEGIN
              size := ftyp.size;
              WHILE (ftyp.comp = csStaticArr) DO ftyp := ftyp.BaseTyp;
              AlignFld(offset, ftyp.size);
              AllocFld(offset, fld.adr, size);
            END;
          END;
          fld := fld.link
        END ;
        IF sizeUndef THEN BEGIN
          AlignRec(offset);
          typ.size := offset
        END;
        IF (doAlloc) AND (typ.mno <= 0) THEN BEGIN
          IF allocTd THEN OPL.AllocTypDesc (typ); (* typ.tdadr = Entry number *)
          Traverse(typ.link, TRUE);
          Traverse(typ.link, FALSE)	(* !!! *)
        END;
        IF btyp = NIL THEN typ.n := 0 ELSE typ.n := btyp.n;
        IF (typ.strobj <> NIL) AND (hasBody IN typ.strobj.conval^.setval) THEN VisitTProcs(typ.strobj);
        VisitTProcs(typ.link);
      END ELSE IF c IN ArraySet THEN BEGIN
        TypSize(btyp, allocDesc);
        IF sizeUndef THEN BEGIN
          CASE c OF
             csStaticArr:	typ.size := typ.n * btyp.size;
             csSDynArr  :	typ.size := OPM.PointerSize;
             csOpenArr  :	typ.size := (typ.n+1)*OPM.LIntSize+OPM.PointerSize;	(*dims + pointer to data (descriptor) *)
             csDynArr   :	typ.size := OPM.PointerSize;
          END
        END;
        IF btyp.comp IN ArraySet THEN
          typ.offset := btyp.offset+OPM.LIntSize
        ELSE
          typ.offset := OPM.LIntSize;
      END ELSE IF f = sfPointer THEN BEGIN
        typ.size := OPM.PointerSize;
        IF doAlloc THEN  TypSize(btyp, allocDesc);
      END ELSE IF f = sfProcTyp THEN BEGIN
        typ.size := OPM.ProcSize;
        IF doAlloc THEN BEGIN
          TypSize(btyp, TRUE);
          size := ExtProcOff;
        END;
        ParamAdr[typ.sysflag](typ.link, size, dummy, FALSE, FALSE, allocDesc)
      END ELSE
        HALT(99);
    END;
  END;
END;

PROCEDURE VarAdr(vvar: Objects; VAR dsize: LONGINT);
VAR
  adr, s: LONGINT;
  list: Objects;	(* used by the inline feature *)
BEGIN
  adr := -dsize;
  list := vvar;
  WHILE vvar <> NIL DO BEGIN
    IF (vvar.link2 = NIL) AND (vvar.vis = internal) THEN BEGIN
      TypSize(vvar.typ, TRUE);
      s := vvar.typ.size;
      AlignVar(adr, s);
      AllocVar(adr, vvar.linkadr, s);
      vvar.adr := 0
    END;
    vvar := vvar.link
  END;
  AlignBlock(adr, dsize);
  WHILE list <> NIL DO BEGIN			(* process inline parameters *)
    IF list.link2 <> NIL THEN
      list.linkadr := list.link2.adr + dsize;
    list := list.link
  END
END;

PROCEDURE ProcSize0(obj: Objects; firstpass: BOOLEAN);
VAR
  psize, entryNr: LONGINT;
  oldPos: TErrorPos;
BEGIN
  ASSERT (obj.prio <> 127);
  oldPos := OPM.errpos;
  IF obj.scope <> NIL THEN
    OPM.errpos.CodePos := obj.scope.adr
  ELSE
    OPM.errpos.CodePos := 0;
  TypSize(obj.typ, TRUE);
  IF ((obj.vis <> internal) = firstpass) OR (obj.mode IN [omTProc, omTyp]) THEN BEGIN
    IF (obj.mode = omLProc) OR ((obj.mode = omXProc) AND (obj.mnolev >= 0) AND (slNeeded IN obj.conval^.setval)) THEN BEGIN
      IF slNeeded IN obj.conval^.setval THEN
        psize := LProcOffSL
      ELSE
        psize := LProcOff;
      obj.adr := 0
    END ELSE
      psize := ExtProcOff;
    IF NOT (obj.mode IN [omLProc, omCProc, omTProc]) THEN BEGIN
      OPL.NewEntry (entryNr);
      INC (obj.adr, entryNr)
    END;
    TypSize(obj.typ, TRUE);
    ParamAdr[obj.sysflag](obj.link, psize, obj.conval^.intval2, obj.mode IN [omTProc], obj.mode = omCProc, TRUE);
    obj.conval^.intval := psize;
    obj.linkadr := OPM.LANotAlloc;
  END ;
  IF NOT firstpass AND (obj.mode <> omTProc) THEN	BEGIN(* mfix *)
    IF NOT (hasBody IN obj.conval^.setval) THEN OPM.err(129);
    VarAdr(obj.scope.scope, obj.conval^.intval2);	(* local variables *)
    Traverse(obj.scope.right, FALSE)
  END;
  OPM.errpos := oldPos
END;

PROCEDURE OPVProcSize(obj: Objects; firstpass: BOOLEAN);
BEGIN
  IF obj.prio = 127 THEN BEGIN
    obj := obj.link2;
    WHILE obj <> NIL DO BEGIN
      ProcSize0 (obj, firstpass);
      obj := obj.left
    END;
  END ELSE
    ProcSize0 (obj, firstpass);
END;

PROCEDURE Traverse(obj: Objects; exported: BOOLEAN);
BEGIN
  IF obj <> NIL THEN BEGIN
    Traverse(obj.left, exported);
    IF (obj.mode = omTyp) AND ((obj.vis <> internal) = exported) THEN BEGIN
      TypSize(obj.typ, TRUE);
      IF (obj.typ.comp = csRecord) AND (obj.typ.tdadr <= -1) THEN OPL.AllocTypDesc(obj.typ);	(* mfix *)
    END ELSE IF obj.mode IN [omLProc, omXProc, omTProc, omCProc, omIProc] THEN
     OPVProcSize(obj, exported);
    Traverse(obj.right, exported);
  END
END;

PROCEDURE FindMth(curmod: SHORTINT; VAR name: STRING; typ: Structs; VAR res: Objects);	(*has a semantic different from OPT.FindField: the viewpoint is not (always) the current module*)
VAR
  obj: Objects;
BEGIN
  res:=NIL;
  WHILE ((typ <> NIL) AND (typ <> OPT.ToBeFixed) AND (typ <> OPT.undftyp)) AND (res=NIL) DO BEGIN
    obj := typ.link;
    WHILE obj <> NIL DO BEGIN
      IF name < obj.name THEN
        obj := obj.left
      ELSE IF name > obj.name THEN
        obj := obj.right
      ELSE BEGIN
        obj.flag:=obj.flag+[OPT.used];
        res := obj;
        obj:=NIL
      END
    END;
    IF (res <> NIL) AND (res.mode = omTProc) AND (res.mnolev <> curmod) AND (res.vis = 0(*internal*)) THEN
      res := NIL;
    IF res=NIL THEN typ := typ.BaseTyp;
  END
END;

PROCEDURE VisitTProcs(obj: Objects);	(* TProcs of base type already visited *)
VAR
  typ, t: Structs;
  redef: Objects;
BEGIN
  IF obj <> NIL THEN BEGIN
    IF obj.mode <> omTyp THEN VisitTProcs(obj.left);
    IF obj.mode IN [omTProc, omTyp] THEN BEGIN
      typ := obj.link.typ;	(* <- type of self *)
      IF typ.form = sfPointer THEN typ := typ.BaseTyp;
      redef := NIL;
      IF obj.mode = omTyp THEN BEGIN	(* Body Method *)
        t := typ.BaseTyp;
        WHILE (t <> NIL) AND NOT (hasBody IN t.strobj.conval^.setval) DO t := t.BaseTyp;
        IF t <> NIL THEN redef := t.strobj;
      END ELSE IF typ.BaseTyp <> NIL THEN	BEGIN(* Normal Method *)
        FindMth(-typ.mno, obj.name, typ.BaseTyp, redef);
        IF (redef <> NIL) AND (redef.mode <> omTProc) THEN redef := NIL;
      END;
      IF redef <> NIL THEN
        obj.adr := $10000*(_SYSTEM._DIV(redef.adr,$10000)) + (_SYSTEM._MOD(obj.adr,$10000))
      ELSE BEGIN
        obj.adr := _SYSTEM._MOD(obj.adr,$10000) + $10000*typ.n;
        INC(typ.n);
      END;
    END;
    IF obj.mode <> omTyp THEN VisitTProcs(obj.right);
  END
END;

//######################################################################

PROCEDURE ParamAdrLR(VAR firstPar: Objects; VAR psize, dsize: LONGINT; TProc, vInline, allocDesc: BOOLEAN);
VAR
  par: Objects;
  typ: Structs;
  adr, s: LONGINT;
BEGIN
  adr := 0;
  IF TProc THEN
    par := firstPar.link
  ELSE
    par := firstPar;		(* guarantee that self is always the last parameter *)

  WHILE par <> NIL DO BEGIN
    typ := par.typ;
    TypSize(typ, allocDesc);
    IF (par.mode = omVarPar) THEN BEGIN
      IF (typ.comp = csRecord) THEN BEGIN
        s:= RecVarParSize;
        IF typ.sysflag = notag THEN DEC(s, 4)
      END ELSE IF (typ.comp = csOpenArr) THEN
        s:= typ.size
      ELSE
        s:= VarParSize;
    END ELSE
      s:= typ.size;
    AlignPar(s);
    AllocPar(adr, par.adr, s);
    par := par.link
  END;
  IF TProc THEN BEGIN
    TypSize(firstPar.typ, TRUE);
    IF firstPar.mode = omVarPar THEN s:= RecVarParSize ELSE s := firstPar.typ.size;
    AlignPar(s);
    AllocPar(adr, firstPar.adr, s)
  END;
  psize := psize + adr;
  par := firstPar;
  IF NOT vInline THEN BEGIN
    WHILE par <> NIL DO BEGIN
      par.adr := psize - par.adr;
      par.linkadr:= par.adr;
      par := par.link;
    END
  END
END;

PROCEDURE ParamAdrRL(VAR firstPar: Objects; VAR psize, dsize: LONGINT; TProc, vInline, allocDesc: BOOLEAN);
VAR
  par, parOld: Objects;
  typ: Structs;
  adr, s: LONGINT;
  stop: BOOLEAN;
BEGIN
  ASSERT( not vInline);
  adr := 0; 
  par := firstPar;
  IF par <> NIL THEN BEGIN
    stop := FALSE;
    WHILE par.link <> NIL DO par := par.link;
    WHILE NOT stop DO BEGIN
      stop := par = firstPar;
      typ := par.typ;
      TypSize(typ, allocDesc (* TRUE *));
      IF typ.comp = csOpenArr THEN BEGIN
        s:= typ.size;
        IF typ.sysflag = notag THEN s := 4 (*DEC(s, 4)*);
      END ELSE IF (par.mode = omVarPar) OR ((typ.size > 4) AND (typ.form <> sfLReal)) THEN BEGIN
          (* ValPar (RECORD), size > 4 are equal to VarPar *)
        IF (typ.comp = csRecord) AND (par.mode = omVarPar) THEN BEGIN
          s:= RecVarParSize;
          IF typ.sysflag = notag THEN DEC(s, 4);
        END ELSE
          s:= VarParSize;
      END ELSE
        s:= typ.size;
      AlignPar(s);
      AllocPar(adr, par.adr, s);
      parOld := par;
      par := firstPar;
      WHILE NOT stop AND (par.link <> parOld) DO par := par.link;
    END
  END ;
  psize := psize + adr;
  par := firstPar;
  IF par <> NIL THEN BEGIN
    stop := FALSE;
    WHILE par.link <> NIL DO par := par.link;
    WHILE NOT stop DO BEGIN
      stop := par = firstPar;
      par.adr := psize - par.adr;
      IF (par.mode = omVar) AND (par.typ.comp IN [csStaticArr, csRecord]) THEN BEGIN
        dsize:= dsize + par.typ.size;
        par.linkadr:= -dsize
      END ELSE
        par.linkadr:= par.adr;
      parOld := par;
      par := firstPar;
      WHILE NOT stop AND (par.link <> parOld) DO par := par.link;
    END
  END
END;

//#######################################################################

PROCEDURE AssignEntries(obj: OPT.Objects; browse: BOOLEAN);	(*assign entry numbers of imported objects*)
VAR
  adr, nofEntries, nofRecs: LONGINT;
  typ: OPT.Structs;
BEGIN
  allocTd := FALSE;
  adr := 0;
  nofEntries := 1;
  nofRecs := -4;
  WHILE obj <> NIL DO BEGIN
    IF obj.mode = omVar THEN BEGIN
      typ := obj.typ;
      TypSize(typ, TRUE);
      IF browse THEN BEGIN
        AlignVar(adr, typ.size);
        AllocVar(adr, obj.linkadr, typ.size);
        obj.adr := adr
      END;
    END ELSE IF obj.mode=omXProc THEN BEGIN
      obj.adr:=nofEntries;
      INC(nofEntries);
      IF obj.conval = NIL THEN
        obj.conval := OPT.NewConst();
      obj.linkadr:=OPM.LANotAlloc;
      obj.conval^.intval:=-1;
      ParamAdr[obj.sysflag](obj.link, obj.conval^.intval, obj.conval^.intval2, FALSE, FALSE, TRUE);
    END ELSE IF obj.mode = omTyp THEN BEGIN
      typ := obj.typ;
      obj.linkadr:=OPM.LANotAlloc;
      TypSize(typ, TRUE);
      IF typ.comp=csRecord THEN BEGIN
        typ.tdadr:=OPM.TDAdrUndef;
        IF browse THEN BEGIN
          typ.tdadr:=nofRecs;
          DEC(nofRecs, 4);
        END;
      END;
    END;
    obj:=obj.nxtobj;
  END;
  allocTd := TRUE
END;

PROCEDURE AssignOwnEntries(obj: OPT.Objects);
VAR
  size: LONGINT;
BEGIN
  allocTd := FALSE;
  WHILE obj <> NIL DO BEGIN
    IF (obj.mode = omVar) THEN BEGIN
      TypSize(obj.typ, TRUE);
      size := obj.typ.size;
      AlignVar(OPOi386.dsize, size);
      AllocVar(OPOi386.dsize, obj.linkadr, size);
      obj.adr := 0
    END ELSE IF obj.mode = omXProc THEN
      OPVProcSize(obj, TRUE)
    ELSE IF (obj.mode = omTyp) AND (obj.typ.comp = csRecord) THEN BEGIN
      IF (obj.typ.tdadr <= 0) THEN BEGIN
        TypSize(obj.typ, TRUE);
        OPL.AllocTypDesc(obj.typ)
      END;
    END;
    obj:=obj.nxtobj
  END;
  allocTd := TRUE
END;

PROCEDURE AllocAddEntries(obj: OPT.Objects);
VAR
  typ: OPT.Structs;
BEGIN
  WHILE obj <> NIL DO BEGIN
    typ := obj.typ;
    IF (obj.mode = omTyp) AND obj.used THEN BEGIN
      IF (typ.comp = csRecord) AND (typ.tdadr <= 0) THEN
        OPL.AllocTypDesc(obj.typ)
      ELSE IF (obj.typ.form = sfPointer) AND (typ.BaseTyp.tdadr <= 0) THEN
        OPL.AllocTypDesc(obj.typ.BaseTyp);
    END;
    obj := obj.nxtobj
  END
END;

PROCEDURE AssignTent (obj: OPT.Objects);	(* mfix *)
VAR
  typ: OPT.Structs;

  PROCEDURE AssignBody(body: OPT.Objects);
  VAR
    entryNr: LONGINT;
  BEGIN
    IF (body <> NIL) AND (body.conval <> NIL) AND (hasBody IN body.conval^.setval)
    AND (_SYSTEM._MOD(body.adr,$10000) = 0) THEN BEGIN
      body.conval^.intval2 := 0;
      body.conval^.intval := LProcOff;
      ParamAdr[body.sysflag](body.link, body.conval^.intval, body.conval^.intval2, TRUE, FALSE, FALSE);
      OPL.NewEntry(entryNr);
      INC(body.adr, entryNr)
    END
  END;

  PROCEDURE EnumTprocs(fld: OPT.Objects);
  VAR
    entryNr: LONGINT;
  BEGIN
    IF fld <> NIL THEN BEGIN
      EnumTprocs(fld.left);
      IF (fld.mode = omTProc) THEN BEGIN
        IF _SYSTEM._MOD(fld.adr,$10000) = 0 THEN BEGIN
          IF NOT (hasBody IN fld.conval^.setval) THEN OPM.err(129);
          VarAdr(fld.scope.scope, fld.conval^.intval2);
          Traverse(fld.scope.right, FALSE);
          OPL.NewEntry(entryNr);
          INC(fld.adr, entryNr)
        END;
      END;
      EnumTprocs(fld.right);
    END
  END;

BEGIN
  IF obj <> NIL THEN BEGIN
    AssignTent (obj.left);
    IF obj.history = 3 THEN obj.mode := 0;
    typ:=obj.typ;
    IF (obj.mode = omVar) THEN BEGIN	(*anonymous type*)
      IF (typ.form=sfPointer) AND (typ.strobj=NIL) AND (typ.BaseTyp.comp=csRecord)
      AND (typ.BaseTyp.ptr=typ) THEN BEGIN
        typ:=typ.BaseTyp;
        AssignBody(typ.strobj);
        EnumTprocs(typ.link)
      END ELSE IF (typ.comp=csRecord) AND (typ.strobj.name='') THEN BEGIN
        AssignBody(typ.strobj);
        EnumTprocs(typ.link);
      END;
    END ELSE IF (obj.mode = omTyp) AND (typ.mno = 0) AND (typ.strobj=obj) THEN BEGIN
      IF (typ.form=sfPointer) AND (typ.BaseTyp.comp=csRecord) AND (typ.BaseTyp.ptr=typ) THEN BEGIN
        typ:=typ.BaseTyp;
        AssignBody(typ.strobj);
        EnumTprocs(typ.link)
      END ELSE IF (typ.comp=csRecord) THEN BEGIN
        AssignBody(typ.strobj);
        EnumTprocs(typ.link)
      END
    END ELSE IF (obj.mode IN [omXProc, omLProc]) THEN
      AssignTent(obj.scope.right);
    AssignTent (obj.right);
  END
END;

PROCEDURE AdrAndSize(topScope: OPT.Objects);
VAR
  i: INTEGER;
  dsize: LONGINT;
BEGIN
  OPM.errpos.CodePos := topScope.adr;
  OPOi386.dsize := 0;
  i := 1;
  WHILE i < OPT.nofmod DO BEGIN
    AssignEntries(OPT.modulesx[i].publics, FALSE);
    INC(i)
  END;
  AssignOwnEntries(OPT.objectsx);
  Traverse(topScope.right, FALSE);
  AssignTent(topScope.right);
  dsize := -OPOi386.dsize;
  VarAdr(topScope.scope, dsize);	(* global variables *)
  OPOi386.dsize := dsize;
  i := 1;
  WHILE i < OPT.nofmod DO BEGIN
    AllocAddEntries(OPT.modulesx[i].publics);
    INC(i);
  END;
END;

PROCEDURE OPVBegin;
BEGIN
  OPT.typSize := TypSize;
  ParamAdr[0] := ParamAdrLR;
	ParamAdr[winapi] := ParamAdrRL;
	ParamAdr[cdecl] := ParamAdrRL;
END;

END.
