UNIT _SYSTEM;

INTERFACE

USES
  SYSUTILS, DIALOGS;
  
CONST 
  MinSInt       = -$80;
  MinInt        = -$8000;
  {
  MinLInt       = $080000000;
  MinRealPat    = $0FF7FFFFF;
  MinLRealPatL  = $0FFFFFFFF;
  MinLRealPatH  = $0FFEFFFFF; 
  }
  MinLInt       =  -$7FFFFFFF-1;
  MinRealPat    = -$800001;   (*0FF7FFFFFH*)
  MinLRealPatL  = -1;       (*0FFFFFFFFH*)
  MinLRealPatH  = -$100001; (*0FFEFFFFFH*)

  MaxSInt       = $7F;
  MaxInt        = $7FFF;
  MaxLInt       = $7FFFFFFF; 
  MaxSet        = 31;
  MaxRealPat    = $7F7FFFFF;
  //MaxLRealPatL  = $0FFFFFFFF;
  MaxLRealPatL  = -1; (*0FFFFFFFFH*)
  MaxLRealPatH  = $7FEFFFFF;

TYPE
  TSET = SET OF 0..31;
  TTypes = (_CHARS,_BOOLEANS,_SETS,_SHORTINTS,_INTEGERS,_LONGINTS,_REALS,_LONGREALS);
  
  FUNCTION _DIV (CONST x, y: INTEGER): INTEGER;
  FUNCTION _MOD (CONST x, y: INTEGER): INTEGER;

  FUNCTION  ROT(x, n: INTEGER): INTEGER;
  FUNCTION  ASH(e, bit : INTEGER): INTEGER;
  FUNCTION  LSH(e, bit : INTEGER): INTEGER;
  FUNCTION  ABS(e : INTEGER): INTEGER;
  FUNCTION  MAX(T:TTYPES): LONGINT;
  FUNCTION  MIN(T:TTYPES): INT64;
  PROCEDURE INCL(VAR S: TSET; P: LONGINT);

  FUNCTION ToSysByte(x,len: INT64): STRING;
  FUNCTION RealToIEEE(CONST r:Single): Integer;
  FUNCTION IEEEToReal(hex: integer): Single;
  FUNCTION DoubleToIEEE(CONST d: DOUBLE): Int64;
  FUNCTION IEEEToDouble(CONST hex: Int64): DOUBLE;

  FUNCTION CAP(c: CHAR): CHAR;

IMPLEMENTATION

FUNCTION ROT (x, n: INTEGER): INTEGER;
BEGIN
  RESULT := (x SHL n) OR (x SHR -n);
END;

FUNCTION _DIV (CONST x, y: INTEGER): INTEGER;
VAR
  Q,R: INTEGER;
BEGIN
  Q:=TRUNC(x/y);
  R:=x MOD y;
  IF R<0 THEN BEGIN
    IF y>0 THEN
      Q:=Q-1
    ELSE
      Q:=Q+1;
  END;
  RESULT := Q;
END;

FUNCTION _MOD (CONST x, y: INTEGER): INTEGER;
BEGIN
  ASSERT(y > 0);
  IF x >= 0 THEN
    RESULT := x MOD y
  ELSE
    RESULT := (y + (x MOD y)) MOD y;
END;

FUNCTION ASH (e, bit : INTEGER): INTEGER;
BEGIN
  IF bit < 0 THEN e := e SHR SYSTEM.ABS(bit) ELSE e := e SHL bit;
  RESULT:=e;
END;

FUNCTION LSH(e, bit : INTEGER): INTEGER;
BEGIN
  IF bit < 0 THEN e := e SHR SYSTEM.ABS(bit) ELSE e := e SHL bit;
  RESULT:=e;
END;

FUNCTION ABS(e : INTEGER): INTEGER;
BEGIN
  RESULT:=SYSTEM.ABS(e);
END;

PROCEDURE INCL(VAR S: TSET; P: LONGINT);
BEGIN
  S:=S+[P];
END;

FUNCTION ToSysByte(x,len: INT64): STRING;
VAR
  i: INTEGER;
BEGIN
  i:=0;
  RESULT:='';
  WHILE i < len DO BEGIN
    RESULT:=RESULT+CHR(LO(x));
    x := x SHR 8;
    INC(i)
  END;
END;

FUNCTION RealToIEEE(CONST r:Single): Integer;
VAR
  Overlay: Integer ABSOLUTE r;
BEGIN
  RESULT := Overlay;
END;

FUNCTION IEEEToReal(hex: integer): Single;
VAR
  d: Single ABSOLUTE hex;
BEGIN
  RESULT := d;
END;

FUNCTION DoubleToIEEE(CONST d: DOUBLE): Int64;
VAR
  Overlay: Int64 ABSOLUTE d;
BEGIN
  RESULT := Overlay;
END;

FUNCTION IEEEToDouble(CONST hex: Int64): DOUBLE;
VAR
  d: DOUBLE ABSOLUTE hex;
BEGIN
  RESULT := d
END;

FUNCTION CAP(c: CHAR): CHAR;
BEGIN
  RESULT:=UPCASE(c);
END;

FUNCTION MAX(T:TTYPES): LONGINT;
BEGIN
  RESULT:= 0;
  CASE T OF
    _SETS     : RESULT:= MaxSet;
    _SHORTINTS: RESULT:= MaxSInt;
    _INTEGERS : RESULT:= MaxInt;
    _LONGINTS : RESULT:= MaxLInt;
    _REALS    : RESULT:= MaxRealPat;
    _LONGREALS: RESULT:= MaxLRealPatH; //MaxLRealPatL = -1;
  END;
END;

FUNCTION MIN(T:TTYPES): INT64;
BEGIN
  RESULT:= 0;
  CASE T OF
    _SETS     : RESULT:= 0;
    _SHORTINTS: RESULT:= MinSInt;
    _INTEGERS : RESULT:= MinInt;
    _LONGINTS : RESULT:= MinLInt;
    _REALS    : RESULT:= MinRealPat;
    _LONGREALS: RESULT:= MinLRealPatH; //MinLRealPatL = -1;
  END;
END;

END.
